package de.uni_stuttgart.fmi.szs.jmoped.eclipse;

import java.io.File;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.debug.core.DebugException;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchConfigurationWorkingCopy;
import org.eclipse.debug.core.model.IProcess;
import org.eclipse.debug.core.model.IStreamsProxy;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.launching.AbstractJavaLaunchConfigurationDelegate;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfo;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfo.HeapOption;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.AnnotatedInstruction;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageListener;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageTester;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageUtils;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.PreStarReachabilityTester;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.ProgressMonitor;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui.LaunchConfigurationControl;

public class MCLaunchConfigurationDelegate 
	extends AbstractJavaLaunchConfigurationDelegate 
{

	// Remy: Remember the last PDS for future pre*
	static PDS lastPDS;
	
	private static String remoplaFileName;
	
	private static Logger logger 
			= Logger.getLogger(MCLaunchConfigurationDelegate.class);
	
	public void launch(ILaunchConfiguration configuration, String mode,
			ILaunch launch, IProgressMonitor monitor) throws CoreException 
	{
		
		if (monitor == null) {
			monitor = new NullProgressMonitor();
		}

		// Gets class path of project for class file lookup
		String[] classPath = getClasspath(configuration);
		
		String classToLaunch = verifyMainTypeName(configuration);
		logger.debug("classToLaunch: " + classToLaunch);
		
		updateRemoplaFileName(configuration, classToLaunch);

		if (configuration.getAttribute(MCPlugin.ATTR_PRE_STAR_METHOD_NAME, 
				(String)null) != null) {
			logger.debug("MCPlugin.ATTR_PRE_STAR_METHOD_NAME: " 
					+ configuration.getAttribute(MCPlugin.ATTR_PRE_STAR_METHOD_NAME, 
					(String)null));
			launchPreStar(configuration, launch, classToLaunch, classPath, monitor);
		}
		else {
			launchPostStar(configuration, launch, classToLaunch, classPath, monitor);
		}
	}
	
	/**
	 * Returns the remopla file's name that was previously analysed.
	 * 
	 * @return the remopla file's name
	 */
	public static String getRemoplaFileName() {
		
		return remoplaFileName;
	}
	
	private void launchPreStar(ILaunchConfiguration configuration, ILaunch launch,
			String classToLaunch, String[] classPath, IProgressMonitor monitor)
		throws CoreException
	{
		String methodToLaunch = configuration.getAttribute(MCPlugin.ATTR_METHOD_NAME, "");
		log(methodToLaunch);
		
		IMethod method = (IMethod)JavaCore.create(methodToLaunch);
		MCProgressView view = MCProgressView.getInstance();
		LaunchConfigurationControl params = (view != null) 
				? view.getConfigControl() : null;
		
		try {
			final PreStarReachabilityTester tester =
				createPreStarTester(classToLaunch, classPath, method);
			
			PDSInfo info = tester.getInfo();
			configuration = configureInfo(info, params, configuration);
			
			int lineOffset = configuration.getAttribute(MCPlugin.ATTR_LINE_OFFSET, -1);
			String label = configuration.getAttribute(MCPlugin.ATTR_PRE_STAR_METHOD_NAME, (String)null);
			String memento = configuration.getAttribute(MCPlugin.ATTR_LAUNCH_CONFIGURATION, (String)null);
			
			CoverageMethod cMethod = CoverageUtils.createMethod(tester.getPDS(), label);
			AnnotatedInstruction[] instructions =
				cMethod.getInstructions(lineOffset).toArray(new AnnotatedInstruction[0]);
			
			// Find label at the corresponding line.
			label = null;
			for (int i = instructions.length - 1; i >= 0; i--) {
				if (!instructions[i].isCovered()) {
					label = instructions[i].getLabel();
					break;
				}
			}
			if (label == null) {
				throw new Exception("No uncovered instruction found");
			}
			final String labelToCover = label;

			// Remy
			//tester.addCoverageListener(new PreStarCoverageListener(method.getResource(), memento));
			TestCoverageListener listener;
			listener = new TestCoverageListener(method.getResource(), memento);
			tester.addCoverageListener(listener);
			
			// Adding these lines will update "Covered Methods" tab.
			CoverageListener l = getCoverageListener();
			if (l != null)
				tester.addCoverageListener(l);
			
			Runnable r = new Runnable() {
				public void run() 
				{
					logger.debug("Running test for label " + labelToCover);
					tester.test(labelToCover, getProgressMonitor());
				}
				
			};
			new Thread(r).start();
			
		}
		catch (Exception e) {
			abort("Analyses threw an exception", e, 0);
		}
	}
	
	/**
	 * Configures <code>info</code>.
	 * If <code>params</code> is null (which is the case when the progress
	 * view has not yet opened), old values from <code>conf</code> are used.
	 * If <code>params</code> is not null, <code>info</code> is configured
	 * accordingly. Moreover, values from <code>params</code> are saved
	 * into a new <code>conf</conf>. The updated <code>conf</code> is returned. 
	 * 
	 * @param info the instance of PDSInfo
	 * @param params the parameters from users
	 * @param conf the launch configuration
	 * @return the updated <code>conf</code>
	 * @throws CoreException
	 */
	private ILaunchConfiguration configureInfo(PDSInfo info, 
			LaunchConfigurationControl params,
			ILaunchConfiguration conf) throws CoreException
	{
		if (params == null) {
			
			info.setBits(conf.getAttribute(MCPlugin.ATTR_BITS, info.getBits()));
			info.setHeapSize(conf.getAttribute(MCPlugin.ATTR_HEAP_SIZE,
					info.getHeapSize()));
			
			HeapOption heapOption = HeapOption.SIMPLE;
			boolean b = conf.getAttribute(MCPlugin.ATTR_ONE_BIT_HEAP, false);
			if (b) heapOption = HeapOption.TWODIMS;
			info.setHeapOption(heapOption);
			
			b = conf.getAttribute(MCPlugin.ATTR_CHECK_FOR_EXCEPTION, true);
			info.setCheckForNullPointerExceptions(b);
			info.setCheckForIndexOutOfBounds(b);
			info.setAbortAfterFailedAssertion(!conf.getAttribute(
					MCPlugin.ATTR_CONTINUE_AFTER_ASSERTION_ERROR, true));
			
			return conf;
		} else {
			
			ExtendedRunnable r;
			
			r = new ExtendedRunnable(params, MCPlugin.ATTR_BITS);
			MCPlugin.getDisplay().syncExec(r);
			info.setBits(r.intValue);
			logger.debug("info.getBits(): " + info.getBits());
			
			r = new ExtendedRunnable(params, MCPlugin.ATTR_HEAP_SIZE);
			MCPlugin.getDisplay().syncExec(r);
			info.setHeapSize(r.intValue);
			logger.debug("info.getHeapSize(): " + info.getHeapSize());
			
			r = new ExtendedRunnable(params, MCPlugin.ATTR_ONE_BIT_HEAP);
			MCPlugin.getDisplay().syncExec(r);
			info.setHeapOption(r.booleanValue ? HeapOption.TWODIMS : HeapOption.SIMPLE);
			
			r = new ExtendedRunnable(params, MCPlugin.ATTR_CHECK_FOR_EXCEPTION);
			MCPlugin.getDisplay().syncExec(r);
			boolean checkExceptions = r.booleanValue;
			info.setCheckForNullPointerExceptions(checkExceptions);
			info.setCheckForIndexOutOfBounds(checkExceptions);
			
			r = new ExtendedRunnable(params, MCPlugin.ATTR_CONTINUE_AFTER_ASSERTION_ERROR);
			MCPlugin.getDisplay().syncExec(r);
			boolean cont = r.booleanValue;
			info.setAbortAfterFailedAssertion(!cont);
			
			ILaunchConfigurationWorkingCopy wc = conf.getWorkingCopy();
			wc.setAttribute(MCPlugin.ATTR_BITS, info.getBits());
			wc.setAttribute(MCPlugin.ATTR_HEAP_SIZE, info.getHeapSize());
			wc.setAttribute(MCPlugin.ATTR_CHECK_FOR_EXCEPTION, checkExceptions);
			wc.setAttribute(MCPlugin.ATTR_CONTINUE_AFTER_ASSERTION_ERROR, cont);
			
			return wc.doSave();
		}
	}
	
	private void launchPostStar(ILaunchConfiguration conf, ILaunch launch,
			String classToLaunch, String[] classPath, IProgressMonitor monitor)
		throws CoreException
	{
		String methodToLaunch = conf.getAttribute(MCPlugin.ATTR_METHOD_NAME, "");
		IMethod method = (IMethod)JavaCore.create(methodToLaunch);
		MCProgressView view = MCProgressView.getInstance();
		LaunchConfigurationControl params = (view != null) 
				? view.getConfigControl() : null;
		
		try {
			// expensive? move it to thread
			boolean initOnly = false;
			/*boolean oldInitOnly = conf.getAttribute(
					MCPlugin.ATTR_CHECK_INITIAL_METHOD_ONLY, false);
			if (params == null) {
				initOnly = oldInitOnly;
			} else {
				ExtendedRunnable r = new ExtendedRunnable(params, 
						MCPlugin.ATTR_CHECK_INITIAL_METHOD_ONLY);
				MCPlugin.getDisplay().syncExec(r);
				initOnly = r.booleanValue;
				if (initOnly != oldInitOnly) {
					ILaunchConfigurationWorkingCopy wc = conf.getWorkingCopy();
					wc.setAttribute(MCPlugin.ATTR_CHECK_INITIAL_METHOD_ONLY, initOnly);
					conf = wc.doSave();
				}
			}*/
			CoverageTester tester = createTester(classToLaunch, 
					classPath, method, initOnly);
			
			lastPDS = tester.getPDS();
			PDSInfo info = tester.getInfo();
			conf = configureInfo(info, params, conf);
			
			tester.addCoverageListener(new TestCoverageListener(method, 
					conf.getMemento()));
			CoverageListener listener = getCoverageListener();
			if (listener != null) {
				tester.addCoverageListener(listener);
			}
			
			new ThreadProcess(launch, tester, null, getProgressMonitor());
		}
		catch (Exception e) {
			abort("Analyses threw an exception", e, 0);
		}
	}
	
	private static class ExtendedRunnable implements Runnable
	{
		LaunchConfigurationControl params;
		public String selection;
		public int intValue;
		public boolean booleanValue;
		
		public ExtendedRunnable(LaunchConfigurationControl params, String selection) {
			
			super();
			this.params = params;
			this.selection = selection;
		}
		
		public void run() {
			if (selection.equals(MCPlugin.ATTR_BITS)) {
				intValue = params.getBits();
			} else if (selection.equals(MCPlugin.ATTR_HEAP_SIZE)) {
				intValue = params.getHeapSize();
			//} else if (selection.equals(MCPlugin.ATTR_CHECK_INITIAL_METHOD_ONLY)) {
			//	booleanValue = params.checkInitialMethodOnly();
			} else if (selection.equals(MCPlugin.ATTR_ONE_BIT_HEAP)) {
				booleanValue = params.useOneBitHeap();
//			} else if (selection.equals(MCPlugin.ATTR_CONTINUE_AFTER_ASSERTION_ERROR)) {
//				booleanValue = params.continueAfterAssertionError();
			} else if (selection.equals(MCPlugin.ATTR_CHECK_FOR_EXCEPTION)) {
				booleanValue = params.checkExceptions();
			}
		}
	}
	
	/**
	 * Updates the field <code>remoplaFileName</code> to
	 * workingDir/remopla/classToLaunch.rem
	 * 
	 * @param conf the launch configuration
	 * @param classToLaunch the class name
	 * @throws CoreException
	 */
	private void updateRemoplaFileName(ILaunchConfiguration conf, 
			String classToLaunch) throws CoreException {
		
		String workingDir = getWorkingDirectory(conf).getAbsolutePath();
		logger.debug("workingDir: " + workingDir);
		
		File remDir = new File(workingDir, "remopla");
		if (!remDir.exists()) remDir.mkdir();
		remoplaFileName = remDir.getAbsolutePath() + File.separator 
				+ classToLaunch + ".rem";
		logger.debug("remoplaFileName: " + remoplaFileName);
	}
	
	private CoverageTester createTester(
			String classToLaunch, String[] classPath,
			IMethod method, boolean filter) throws Exception
	{
		CoverageTester tester;
		if (method == null || method.isMainMethod()) {
			tester = new CoverageTester(classToLaunch, classPath, remoplaFileName); 
		}
		else {
			String elemName = method.isConstructor()? "<init>" : method.getElementName();
			tester = new CoverageTester(classToLaunch, classPath, elemName,
					 method.getSignature(), remoplaFileName);
		}
		if (filter) {
			PDSMethod pdsMethod = tester.getPDS().getRootPDSMethod();
			tester.setFilterMethodName(pdsMethod.getFormattedName());
		}
			
		return tester;
	}
	
	private PreStarReachabilityTester createPreStarTester(String classToLaunch, 
			String[] classPath, IMethod method) throws Exception
	{
		// Remy: When running prestar, we use the same PDS as before.
		if (lastPDS != null)
			return new PreStarReachabilityTester(lastPDS);
		
		if (method == null || method.isMainMethod()) {
			return new PreStarReachabilityTester(classToLaunch, classPath);
		}
		else {
			PDS pds = new PDS(classToLaunch, classPath, method.getElementName(),
					method.getSignature());
			return new PreStarReachabilityTester(pds);
		}
	}
	
	private void log(Object o)
	{
		logger.debug(o.toString());
	}
	
	/**
	 * Returns progress monitor of progress view or 
	 * {@link de.uni_stuttgart.fmi.szs.jmoped.coverage.NullProgressMonitor}.
	 * @return
	 */
	private ProgressMonitor getProgressMonitor()
	{
		//GetProgressViewRunnable r = new GetProgressViewRunnable();
		//MCPlugin.getDisplay().syncExec(r);
		MCProgressView view = MCProgressView.getInstance();
		return view != null ? view.getProgressMonitor() 
				: new de.uni_stuttgart.fmi.szs.jmoped.coverage.NullProgressMonitor();
	}
	
	private CoverageListener getCoverageListener()
	{
		MCProgressView view = MCProgressView.getInstance();
		//GetProgressViewRunnable r = new GetProgressViewRunnable();
		//MCPlugin.getDisplay().syncExec(r);
		return view != null ? view.getCoverageListener() : null;
	}
	
	/*private static class GetProgressViewRunnable implements Runnable
	{
		public MCProgressView view;
		public void run() {
			view = MCPlugin.getDefault().findProgressViewtInActivePage();
		}
		
	}*/
	
	private static class ThreadProcess implements IProcess
	{

		private String label;
		private ILaunch launch;
		private Map map;
		private Thread thread;
		
		public ThreadProcess(ILaunch launch, final CoverageTester tester, 
							 String label, final ProgressMonitor monitor)
		{
			this.launch = launch;
			this.label = label;
			thread = new Thread(new Runnable() {
				public void run() { 
					CoverageMethod[] methods = tester.test(monitor);
				}
			});
			thread.start();
		}
		
		public String getLabel() 
		{
			return label;
		}

		public ILaunch getLaunch() 
		{
			return launch;
		}

		public IStreamsProxy getStreamsProxy() 
		{
			return null;
		}

		public void setAttribute(String key, String value) 
		{
			if (map == null) {
				map = new HashMap();
			}
			map.put(key, value);
		}

		public String getAttribute(String key) 
		{
			return (String)(map == null ? null : map.get(key));
		}

		public int getExitValue() throws DebugException 
		{
			return 0;
		}

		public Object getAdapter(Class adapter) 
		{
			return null;
		}

		public boolean canTerminate() 
		{
			return thread.isAlive();
		}

		public boolean isTerminated() 
		{
			return !thread.isAlive();
		}

		public void terminate() throws DebugException 
		{
			thread.interrupt();
		}
	}
}
