package de.uni_stuttgart.fmi.szs.jmoped.eclipse;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.log4j.Logger;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchConfigurationType;
import org.eclipse.debug.core.ILaunchConfigurationWorkingCopy;
import org.eclipse.debug.core.ILaunchManager;
import org.eclipse.debug.ui.DebugUITools;
import org.eclipse.debug.ui.IDebugModelPresentation;
import org.eclipse.debug.ui.ILaunchShortcut;
import org.eclipse.jdt.core.IClassFile;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IMember;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.Signature;
import org.eclipse.jdt.launching.IJavaLaunchConfigurationConstants;
import org.eclipse.jdt.ui.IWorkingCopyManager;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.dialogs.ElementListSelectionDialog;


public class MCLaunchShortcut implements ILaunchShortcut {

	private Logger logger = Logger.getLogger(MCLaunchShortcut.class);
	
	/**
	 * @see ILaunchShortcut#launch(IEditorPart, String)
	 */
	public void launch(IEditorPart editor, String mode) {
		ISelection selection = editor.getSite().getPage().getSelection();
		IJavaElement element = null;
		if (selection instanceof ITextSelection) {
			element = getJavaElement(editor, (ITextSelection)selection);
		}
		if (selection instanceof IStructuredSelection) {
			element = getJavaElement((IStructuredSelection)selection);
		}
		if (element != null) {
			launchElement(element);
		}
	}

	/**
	 * @param element
	 */
	private void launchElement(IJavaElement element) {
		IType classType = getClassType(element);
		IMethod method = null;
		if (element instanceof IMethod) {
			method = (IMethod)element;
		}
		else {
			method = getMainMethod(classType);
		}
		try {
			if (classType == null || method == null || method.isConstructor()) {
				MCPlugin.log("no way we can execute this");
				return;
			}
		}
		catch (JavaModelException jvme) {
			MCPlugin.log(jvme);
		}
		ILaunchConfiguration config = findConfiguration(method);
		if (config == null) {
			config = createConfiguration(method);
		}
		DebugUITools.launch(config, "run");
	}

	/**
	 * @see ILaunchShortcut#launch(ISelection, String)
	 */
	public void launch(ISelection selection, String mode) {
		if (selection instanceof IStructuredSelection) {
			IJavaElement element = getJavaElement((IStructuredSelection)selection);
			if (element != null) {
				launchElement(element);
			}
			else {
				MCPlugin.log("no element for selection launch found");
			}
		}
		else {
			throw new IllegalArgumentException("unhandled case");
		}
	}
	
	private IJavaElement getJavaElement(IStructuredSelection selection) 
	{
		Object element = selection.getFirstElement();
		if (element instanceof IJavaElement) {
			return (IJavaElement)element;
		}
		return null;
	}
		
	private IJavaElement getJavaElement(IEditorPart editor, ITextSelection selection) 
	{
		int offset = selection.getOffset();
		IEditorInput editorInput = editor.getEditorInput();
		IWorkingCopyManager manager = JavaUI.getWorkingCopyManager();
		ICompilationUnit unit = manager.getWorkingCopy(editorInput);
		if (unit != null) {
			try {
				// returns null if unit itself is found at offset
				IJavaElement element = unit.getElementAt(offset);
				return element != null ? element : unit;
			}
			catch (JavaModelException e) {
				MCPlugin.log(e);
			}
		}
		IClassFile classFile = (IClassFile) editorInput.getAdapter(IClassFile.class);
		if (classFile != null) {
			try {
				IJavaElement element = classFile.getElementAt(offset);
				return element != null ? element : classFile;
			}
			catch (JavaModelException ex) {
				MCPlugin.log(ex);
			}
		}
		IJavaElement je = (IJavaElement)editorInput.getAdapter(IJavaElement.class);
		if (je != null) {
			return je;
		}
		return (IJavaElement)editor.getAdapter(IJavaElement.class);
	}
	
	private IType getClassType(IJavaElement element)
	{
		if (element instanceof ICompilationUnit) {
			return ((ICompilationUnit)element).getType(Signature.getQualifier
					(element.getElementName()));
		}
		else if (element instanceof IClassFile) {
			try {
				return ((IClassFile)element).getType();
			}
			catch (JavaModelException e) {
				MCPlugin.log(e);
			}
		}
		else if (element instanceof IType) {
			return (IType)element;
		}
		else if (element instanceof IMember) {
			return ((IMember)element).getDeclaringType();
		}
		return null;
	}
	
	private IMethod getMainMethod(IType type) 
	{
		if (type != null && type.exists()) {
			try {
				IMethod[] methods = type.getMethods();
				for (int i = 0; i < methods.length; i++) {
					if (methods[i].isMainMethod()) {
						return methods[i];
					}
				}
			}
			catch (JavaModelException jme) {
				MCPlugin.log(jme);
			}
		}
		return null;
	}
	
//		String name= JavaElementLabels.getTextLabel(element,
//				JavaElementLabels.ALL_FULLY_QUALIFIED);
//		
//		ILaunchConfiguration config = createConfiguration(element.getJavaProject(),
//				name, "",	element.getHandleIdentifier(), ""	);
//		
//		log("still alive?");
//		DebugUITools.launch(config, mode);
	

	/*
	 * lauch:
	 *   -> get class and method
	 *   method either from selection or look for main
	 *   -> neither? abort
	 *  find configuration for IType + IMethod
	 *  not found? create configuration for IType + IMethod
	 *  -> call launcher
	 */

	private ILaunchConfiguration createConfiguration(IMethod method) 
	{
		String className = method.getDeclaringType().getFullyQualifiedName();
		String name = className + "." + method.getElementName();
		ILaunchConfiguration config = null;
		try {
			logger.debug("MCLaunchShortcut.createConfiguration");
			ILaunchConfigurationType configType= getMCLaunchConfigType();
			ILaunchConfigurationWorkingCopy wc = configType.newInstance(null,
					DebugPlugin.getDefault().getLaunchManager()
					.generateUniqueLaunchConfigurationNameFrom(name)); 
			wc.setAttribute(IJavaLaunchConfigurationConstants.ATTR_MAIN_TYPE_NAME, 
					className);
			wc.setAttribute(IJavaLaunchConfigurationConstants.ATTR_PROJECT_NAME, 
					method.getJavaProject().getElementName());
			wc.setAttribute(MCPlugin.ATTR_METHOD_NAME, method.getHandleIdentifier());
			config = wc.doSave();		
		} 
		catch (CoreException ce) {
			log(ce);
		}
		return config;
	}
	
	
	private ILaunchConfiguration findConfiguration(IMethod method) {
		
		ILaunchConfigurationType configType = getMCLaunchConfigType();
		String className = method.getDeclaringType().getFullyQualifiedName();
		List candidateConfigs = Collections.EMPTY_LIST;
		
		try {
			logger.debug("findConfiguration()");
			logger.debug("  className: " + className);
			logger.debug("  methodName: " + method.getElementName());
			ILaunchConfiguration[] configs = DebugPlugin.getDefault().getLaunchManager().
				getLaunchConfigurations(configType);
			candidateConfigs = new ArrayList(configs.length);
			
			logger.debug("  configs.length: " + configs.length);
			for (int i = 0; i < configs.length; i++) {
				
				ILaunchConfiguration config = configs[i];
				
				if (config.getAttribute(IJavaLaunchConfigurationConstants.ATTR_MAIN_TYPE_NAME, "")
						.equals(className)) { //$NON-NLS-1$
					if (config.getAttribute(IJavaLaunchConfigurationConstants.ATTR_PROJECT_NAME, "").
							equals(method.getJavaProject().getElementName())) { //$NON-NLS-1$
						if (config.getAttribute(MCPlugin.ATTR_METHOD_NAME, "").
								equals(method.getHandleIdentifier())) {
							
							// TODO: I don't know how config got the value of ATTR_PRE_STAR_METHOD_NAME
							logger.debug("  method.getHandleIdentifier: " + method.getHandleIdentifier());
							logger.debug("  ATTR_PRE_STAR_METHOD_NAME: " + config.getAttribute(MCPlugin.ATTR_PRE_STAR_METHOD_NAME, ""));
							candidateConfigs.add(config);							
						}
					}
				}
			}
		} 
		catch (CoreException e) {
			MCPlugin.log(e);
		}

		int candidateCount = candidateConfigs.size();
		 if (candidateCount == 1) {
			logger.debug("  one candidate");
			return (ILaunchConfiguration) candidateConfigs.get(0);
		} 
		 else if (candidateCount > 1) {
			return chooseConfiguration(candidateConfigs);
		}
		
		return null;
	}
	
	private ILaunchConfiguration chooseConfiguration(List configList) {
		IDebugModelPresentation labelProvider = DebugUITools.newDebugModelPresentation();
		ElementListSelectionDialog dialog= new ElementListSelectionDialog(getShell(), labelProvider);
		dialog.setElements(configList.toArray());
		dialog.setTitle("Choose test run");  //$NON-NLS-1$
		dialog.setMessage("Multiple have been found, yada yada yada");  //$NON-NLS-1$
		dialog.setMultipleSelection(false);
		int result = dialog.open();
		labelProvider.dispose();
		if (result == Window.OK) {
			return (ILaunchConfiguration)dialog.getFirstResult();
		}
		return null;		
	}
	
	private Shell getShell() {
		IWorkbenchWindow window = MCPlugin.getDefault().getWorkbench().
			getActiveWorkbenchWindow();
		return window != null ? window.getShell() : null;
	}
	
	
	
	
	private void log(Object o) {
		logger.debug(o.toString());
	}
	

	private ILaunchConfigurationType getMCLaunchConfigType() {
		
		ILaunchManager lm = DebugPlugin.getDefault().getLaunchManager();
		return lm.getLaunchConfigurationType(MCPlugin.LAUNCH_ID);
	}
	
	
}
