package de.uni_stuttgart.fmi.szs.jmoped.eclipse;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchListener;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IViewPart;
import org.eclipse.ui.IViewReference;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.osgi.framework.BundleContext;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Coverage;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageAdapter;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageListener;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageResult;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.ProgressMonitor;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;

/**
 * The main plugin class to be used in the desktop.
 */
public class MCPlugin extends AbstractUIPlugin {

	//The shared instance.
	private static MCPlugin plugin;
	
	public final static String PLUGIN_ID = "de.uni_stuttgart.fmi.szs.jmoped.eclipse";
	
	public final static String ATTR_METHOD_NAME = PLUGIN_ID + ".METHOD";
	
	public final static String ATTR_BITS = PLUGIN_ID + ".BITS";
	
	public final static String ATTR_HEAP_SIZE = PLUGIN_ID + ".HEAP_SIZE";
	
	public final static String ATTR_ONE_BIT_HEAP = PLUGIN_ID + ".ONE_BIT_HEAP";
	
	public final static String ATTR_CHECK_FOR_EXCEPTION = PLUGIN_ID 
		+ ".CHECK_FOR_EXCEPTIONS";
	
	public final static String ATTR_CONTINUE_AFTER_ASSERTION_ERROR =
		PLUGIN_ID + ".CONTINUE_AFTER_ASSERTION_ERROR";
	
	public final static String ATTR_CHECK_INITIAL_METHOD_ONLY = 
		PLUGIN_ID + ".CHECK_INITIAL_METHOD_ONLY";
	
	public final static String ATTR_LINE_OFFSET = PLUGIN_ID + ".LINE_OFFSET";
	
	public final static String ATTR_LAUNCH_CONFIGURATION = PLUGIN_ID 
		+ ".LAUNCH_CONFIGURATION";
	
	public final static String ATTR_PRE_STAR_METHOD_NAME = PLUGIN_ID
		+ ".PRE_STAR_METHOD";
	
	public final static String ATTR_PC_OFFSET = PLUGIN_ID + ".PC_OFFSET";
	
	public final static String MARKER_COVERED_ID = PLUGIN_ID + ".coveragemarker";
	
	public final static String MARKER_ASSERTION_FAILED_ID = PLUGIN_ID 
		+ ".assertionfailedmarker";
	
	public final static String MARKER_PARTIALLY_COVERED_ID = PLUGIN_ID
		+ ".partiallycoveredmarker";
	
	public final static String MARKER_NOT_COVERED_ID = PLUGIN_ID
		+ ".notcoveredmarker";
	
	public final static String MARKER_HEAP_OVERFLOW_ID = PLUGIN_ID
		+ ".heapoverflowmarker";
	
	public final static String MARKER_NPE_ID = PLUGIN_ID	
		+ ".npemarker";
	
	public final static String MARKER_INDEX_OUT_OF_BOUNDS_ID = PLUGIN_ID
		+ ".ioobmarker";
	
	public static String LAUNCH_ID = PLUGIN_ID + ".launchConfig";
	
	private static URL baseURL;
	
	private MCLaunchListener launchListener;
	
	private Set trackedLaunches = new HashSet();
	
	/**
	 * The current progress monitor.
	 */
	private ProgressMonitor monitor = null;
	
	private MCCoverageListener coverageListener = new MCCoverageListener();
	
 	private CoverageRunnable nextTest;
 	
 	static Logger logger = Logger.getLogger(MCPlugin.class);
	
	/**
	 * The constructor.
	 */
	public MCPlugin() 
	{
		plugin = this;
		try {
			baseURL = new URL(Platform.getBundle(PLUGIN_ID).getEntry("/"), "icons/");
		}
		catch (MalformedURLException mue) {
		}
	}
	
	public static String getUniqueIdentifier() {
		return PLUGIN_ID;
	}
	
	/**
	 * Maps states to their marker id.
	 */
	public static String getMarkerId(State state)
	{
		if (state ==  State.COVERED) {
			return MARKER_COVERED_ID;
		}
		else if (state == State.ASSERTION_FAILED) {
			return MARKER_ASSERTION_FAILED_ID;
		}
		else if (state == State.PARTIALLY_COVERED) {
			return MARKER_PARTIALLY_COVERED_ID;
		}
		else if (state == State.HEAP_OVERFLOW) {
			return MARKER_HEAP_OVERFLOW_ID;
		}
		else if (state == State.NPE) {
			return MARKER_NPE_ID;
		}
		else if (state == State.NOT_COVERED) {
			return MARKER_NOT_COVERED_ID;
		}
		else if (state == State.INDEX_OUT_OF_BOUNDS) {
			return MARKER_INDEX_OUT_OF_BOUNDS_ID;
		}
		else {
			throw new IllegalArgumentException("unexpected state given " + state);
		}
	}
	
	public static ImageDescriptor getImage(State state)
	{
		ImageDescriptor desc;
		if (state == State.COVERED) {
			desc = getImage("covered.png");
		}
		else if (state == State.ASSERTION_FAILED) {
			desc = getImage("assertion.png");
		}
		else if (state == State.PARTIALLY_COVERED) {
			desc = getImage("partiallycovered.png");
		}
		else if (state == State.HEAP_OVERFLOW) {
			desc = getImage("heapoverflow.png");
		}
		else if (state == State.NPE) {
			desc = getImage("npe.png");
		}
		else if (state == State.NOT_COVERED) {
			desc = getImage("notcovered.png");
		}
		else if (state == State.INDEX_OUT_OF_BOUNDS) {
			desc = getImage("ioob.png");
		}
		else {
			throw new IllegalArgumentException("unexpected state given " + state);
		}
		return desc;
	}

	/**
	 * This method is called upon plug-in activation
	 */
	public void start(BundleContext context) throws Exception 
	{
		super.start(context);
		launchListener = new MCLaunchListener();
		DebugPlugin.getDefault().getLaunchManager().addLaunchListener(launchListener);
	}
	
	/**
	 * Runs the coverage test, no other test must currently be running.
	 * @param test
	 */
	public void run(CoverageRunnable test)
	{
		if (coverageListener.coverageRunning()) {
			throw new IllegalStateException("Other test is running");
		}
		
		// add listeners
		Coverage coverage = test.getCoverage(); 
		coverage.addCoverageListener(coverageListener);
		
		monitor = getProgressMonitor();
	}
	
	public void cancelCurrentAndEnqueue(CoverageRunnable test)
	{
		ProgressMonitor mon = monitor;
		if (mon != null) {
			mon.setCanceled(true);
			nextTest = test;
		}
		else {
			run(test);
		}
	}
	
	private void runNext()
	{
		CoverageRunnable next = nextTest;
		nextTest = null;
		if (next!= null) {
			run(next);
		}
	}

	/**
	 * This method is called when the plug-in is stopped
	 */
	public void stop(BundleContext context) throws Exception 
	{
		super.stop(context);
		DebugPlugin.getDefault().getLaunchManager().removeLaunchListener(launchListener);
		launchListener = null;
		plugin = null;
	}

	/**
	 * Returns the shared instance.
	 */
	public static MCPlugin getDefault() 
	{
		return plugin;
	}
	
	public static void log(IStatus status) 
	{
		getDefault().getLog().log(status);
	}
	
	public static void log(CoreException e) 
	{
		log(e.getStatus());
	}
	
	public static void log(Throwable e)
	{
		log(new Status(IStatus.WARNING, getUniqueIdentifier(), 0, e.getMessage(), e));
	}
	
	public static void log(String log, Throwable e)
	{
		log(new Status(IStatus.WARNING, getUniqueIdentifier(), 0, log, e));
	}
	
	public static void log(String log) 
	{
		log(new Status(IStatus.WARNING, getUniqueIdentifier(), 0, log, null));
	}
	
	public static ImageDescriptor getImage(String name)
	{
		return getImage(null, name);
	}
	
	public static ImageDescriptor getEnabledImage(String name)
	{
		return getImage("enabled/", name);
	}
	
	public static ImageDescriptor getDisabledImage(String name)
	{
		return getImage("disabled/", name);
	}
	
	public static ImageDescriptor getHoverImage(String name)
	{
		return getImage("hover/", name);
	}
	
	private static ImageDescriptor getImage(String prefix, String name)
	{
		try {
			return ImageDescriptor.createFromURL(getImageURL(prefix, name));
		}
		catch (MalformedURLException mue) {
			return ImageDescriptor.getMissingImageDescriptor();
		}
	}
	
	private static URL getImageURL(String prefix, String name) 
		throws MalformedURLException
	{
		if (baseURL == null) {
			throw new MalformedURLException();
		}
		return new URL(baseURL, prefix != null ? prefix + name : name);
	}

	/**
	 * Returns an image descriptor for the image file at the given
	 * plug-in relative path.
	 *
	 * @param path the path
	 * @return the image descriptor
	 */
	public static ImageDescriptor getImageDescriptor(String path) 
	{
		return AbstractUIPlugin.imageDescriptorFromPlugin("moped_coverage_plugin", 
				path);
	}
	
	public static IWorkbenchPage getActivePage() 
	{
		IWorkbenchWindow activeWorkbenchWindow= getActiveWorkbenchWindow();
		if (activeWorkbenchWindow == null)
			return null;
		return activeWorkbenchWindow.getActivePage();
	}
	
	public static IWorkbenchWindow getActiveWorkbenchWindow() 
	{
		if (plugin == null) {
			return null;
		}
		IWorkbench workBench= plugin.getWorkbench();
		if (workBench == null) {
			return null;
		}
		return workBench.getActiveWorkbenchWindow();
	}
	
	//TODO incomplete
	public static IProject getActiveProject() {
		
		IWorkbenchPage page = getActivePage();
		if (page == null) {
			logger.debug("page null");
			return null;
		}
		
		IViewReference[] parts = page.getViewReferences();
		return null;
	}
	
	public MCProgressView findProgressViewtInActivePage() 
	{
		IWorkbenchPage page = getActivePage();
		if (page == null) {
			return null;
		}
		return (MCProgressView)page.findView(MCProgressView.NAME);
	}
	
	private void showProgressViewInActivePage()
	{
		IWorkbenchPage page = getActivePage();
		if (page != null) {
			try {
				page.showView(MCProgressView.NAME);
			} 
			catch (PartInitException e) {
				MCPlugin.log(e);
			}
		}
	}
	
	private MCProgressView showProgressViewInActivePage(MCProgressView progressView) 
	{
		IWorkbenchPart activePart= null;
		IWorkbenchPage page= null;
		try {
			if (progressView != null && progressView.isCreated()) {
				return progressView;
			}
			page = getActivePage();
			if (page == null) {
				return null;
			}
			activePart = page.getActivePart();
			//	show the result view if it isn't shown yet
			return (MCProgressView)page.showView(MCProgressView.NAME);
		} 
		catch (PartInitException pie) {
			log(pie);
			return null;
		}
		finally{
			//restore focus stolen by the creation of the result view
			if (page != null && activePart != null)
				page.activate(activePart);
		}
	}
	
	public static Display getDisplay() 
	{
		Display display= Display.getCurrent();
		if (display == null) {
			display= Display.getDefault();
		}
		return display;
	}
	
	/**
	 * Returns progress monitor of progress view or 
	 * {@link de.uni_stuttgart.fmi.szs.jmoped.coverage.NullProgressMonitor}.
	 * @return
	 */
	private ProgressMonitor getProgressMonitor()
	{
		GetProgressViewRunnable r = new GetProgressViewRunnable();
		MCPlugin.getDisplay().syncExec(r);
		return r.view != null ? r.view.getProgressMonitor() 
				: new de.uni_stuttgart.fmi.szs.jmoped.coverage.NullProgressMonitor();
	}
	
	private CoverageListener getCoverageListener()
	{
		GetProgressViewRunnable r = new GetProgressViewRunnable();
		MCPlugin.getDisplay().syncExec(r);
		return r.view != null ? r.view.getCoverageListener() : null;
	}
	
	private static class GetProgressViewRunnable implements Runnable
	{
		public MCProgressView view;
		public void run() {
			view = MCPlugin.getDefault().findProgressViewtInActivePage();
		}
		
	}

	private class MCLaunchListener implements ILaunchListener
	{

		public void launchRemoved(ILaunch launch) 
		{
//			trackedLaunches.remove(launch);
		}

		public void launchAdded(final ILaunch launch) 
		{
//			trackedLaunches.add(launch);
			if (!isMCLaunch(launch)) {
				return;
			}
		
			getDisplay().asyncExec(new Runnable() {
				public void run()
				{
					MCProgressView view = findProgressViewtInActivePage();
					if (view != null) {
						showProgressViewInActivePage();
						view.setLaunch(launch);
					}
				}
			});
		}

		public void launchChanged(final ILaunch launch) 
		{
//			if (!trackedLaunches.contains(launch)) {
//				return;
//			}
			
//			trackedLaunches.remove(launch);
			
			if (!isMCLaunch(launch)) {
				return;
			}
		
			getDisplay().asyncExec(new Runnable() {
				public void run()
				{
					MCProgressView view = findProgressViewtInActivePage();
					if (view != null) {
						showProgressViewInActivePage();
						view.setLaunch(launch);
					}
				}
			});
		}
		
		private boolean isMCLaunch(ILaunch launch)
		{
			ILaunchConfiguration config = launch.getLaunchConfiguration();
			try {
				return config.getAttribute(MCPlugin.ATTR_METHOD_NAME, (String)null) != null;
			}
			catch (CoreException e) {
				MCPlugin.log(e);
			}
			return false;
		}
		
	}
	
	private class MCCoverageListener extends CoverageAdapter
	{
		private boolean running;
		
		public boolean coverageRunning()
		{
			return running;
		}
		
		@Override
		public void coverageStarted(PDS pds) 
		{
			running = true;
		}
		
		@Override
		public void coverageEnded(CoverageResult result) 
		{
			running = false;
			monitor = null;
			runNext();
		}
	}
}
