package de.uni_stuttgart.fmi.szs.jmoped.eclipse;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchConfigurationWorkingCopy;
import org.eclipse.debug.ui.DebugUITools;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.ProgressBar;
import org.eclipse.ui.IActionBars;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IMemento;
import org.eclipse.ui.IViewSite;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchActionConstants;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.FileEditorInput;
import org.eclipse.ui.part.ViewPart;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageAdapter;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageListener;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageResult;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.ProgressMonitor;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.TraceNode;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui.LaunchConfigurationControl;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui.SWTUtils;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui.TabControl;

public class MCProgressView extends ViewPart {

	public static final String NAME = 
		"de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCProgressView";
	
	private Action stopAction;
	
	private ProgressBar progressBar;
	private Label titleLabel;
	private Label progressLabel;
	private Label uncoveredLabel;
	
	private Group paramGroup;
	private Group backwardsGroup;
	private Group junitGroup;
	private Composite actionComposite;
	
	private ViewProgressMonitor monitor;
	
	private CoverageListener listener;
	
	private TabControl tabControl;
	
	private List unreachedLabels = new ArrayList();
	
	private LaunchConfigurationControl configControl;
	
	private Button postStarButton;
	private Button stopButton;
	private Button remoplaButton;
	
	private ILaunch launch;

	private boolean ignoreNextChange = false;
	
	private IMemento memento;
	
	static Logger logger = Logger.getLogger(MCProgressView.class);
	
	/**
	 * This is a callback that will allow us
	 * to create the viewer and initialize it.
	 */
	public void createPartControl(Composite parent)
	{
		makeActions();
		contributeToActionBars();
		parent = new Composite(parent, SWT.NONE);
		parent.setLayout(new GridLayout());
		titleLabel = createLabel(parent);
		progressLabel = createLabel(parent);
		createProgressBar(parent);
		createParamsGroup(parent);
		createActionComposite(parent);
		createTabs(parent);
		
		setEnabledRecursively(paramGroup, true);
		setEnabledRecursively(actionComposite, false);

		monitor = new ViewProgressMonitor();
		listener = new ViewCoverageListener();
	}
	
	@Override
	public void saveState(IMemento memento) 
	{
		if (isCreated()) {
			tabControl.saveState(memento);
			memento.putInteger(MCPlugin.ATTR_BITS, configControl.getBits());
			memento.putInteger(MCPlugin.ATTR_HEAP_SIZE, configControl.getHeapSize());
			//memento.putInteger(MCPlugin.ATTR_CHECK_INITIAL_METHOD_ONLY, 
			//		configControl.checkInitialMethodOnly() ? 1 : 0);
			memento.putInteger(MCPlugin.ATTR_ONE_BIT_HEAP,
					configControl.useOneBitHeap() ? 1 : 0);
//			memento.putInteger(MCPlugin.ATTR_CONTINUE_AFTER_ASSERTION_ERROR,
//					configControl.continueAfterAssertionError() ? 1 : 0);
			memento.putInteger(MCPlugin.ATTR_CHECK_FOR_EXCEPTION,
					configControl.checkExceptions() ? 1 : 0);
		}
		super.saveState(memento);
	}
	
	@Override
	public void init(IViewSite site, IMemento memento) throws PartInitException 
	{
		this.memento = memento;
		super.init(site, memento);
	}
	
	public IProgressMonitor getMonitor()
	{
		// TODO side effect in getter, refactor
		monitor.setCanceled(false);
		stopAction.setEnabled(true);
		stopButton.setEnabled(true);
		return monitor;
	}
	
	public ProgressMonitor getProgressMonitor()
	{
		// TODO side effect in getter, refactor
		monitor.setCanceled(false);
		return monitor;
	}
	
	public CoverageListener getCoverageListener()
	{
		return listener;
	}
	
	public LaunchConfigurationControl getConfigControl() {
		
		return configControl;
	}
	
	/**
	 * Gets the active instance of this class.
	 * 
	 * @return the MCProgressView instance
	 */
	public static MCProgressView getInstance() {
		
		GetProgressViewRunnable r = new GetProgressViewRunnable();
		MCPlugin.getDisplay().syncExec(r);
		
		return r.view;
	}
	
	private static class GetProgressViewRunnable implements Runnable {
		
		public MCProgressView view;
		public void run() {
			view = MCPlugin.getDefault().findProgressViewtInActivePage();
		}
	}
	
	private Label createLabel(Composite parent) 
	{
		Label label = new Label(parent, SWT.HORIZONTAL);
		label.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
		return label;
	}

	private void createProgressBar(Composite parent) 
	{
		progressBar = new ProgressBar(parent, SWT.HORIZONTAL);
		progressBar.setEnabled(false);
		progressBar.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
	}
	
	private void createParamsGroup(Composite parent)
	{
		paramGroup = createGroup(parent, "Parameters");
		paramGroup.setLayout(new GridLayout(2, false));
		
		configControl = new LaunchConfigurationControl(paramGroup, SWT.NONE, memento);
		GridData data = new GridData(SWT.LEFT, SWT.TOP, true, true, 2, 1);
		configControl.setLayoutData(data);
	}
	
	private void createActionComposite(Composite parent) {
		
		actionComposite = new Composite(parent, SWT.NONE);
		actionComposite.setLayout(new GridLayout(3, false));
		
		postStarButton = new Button(actionComposite, SWT.PUSH);
		postStarButton.setText("Rerun");
		postStarButton.setEnabled(false);
		postStarButton.setLayoutData(new GridData(SWT.END, SWT.BEGINNING, false, false));
		postStarButton.addSelectionListener(new SelectionListener() {
			public void widgetSelected(SelectionEvent e) {
				analyseButtonPressed();				
			}
			public void widgetDefaultSelected(SelectionEvent e) {
			}
		});
		
		stopButton = new Button(actionComposite, SWT.PUSH);
		stopButton.setText(stopAction.getText());
		stopAction.setEnabled(stopAction.isEnabled());
		stopButton.addSelectionListener(new SelectionListener() {
			public void widgetSelected(SelectionEvent e) {
				stopAction.run();
			}
			public void widgetDefaultSelected(SelectionEvent e) {
			}
		});
		
		remoplaButton = new Button(actionComposite, SWT.PUSH);
		remoplaButton.setText("Remopla");
		remoplaButton.setEnabled(false);
		remoplaButton.addSelectionListener(new SelectionListener() {
			public void widgetSelected(SelectionEvent e) {
				remoplaButtonPressed();
			}
			public void widgetDefaultSelected(SelectionEvent e) {
			}
		});
	}
	
	/**
	 * Opens an editor and displays the content of the remopla file that
	 * was created in the last run.
	 */
	private void remoplaButtonPressed() {
		
		// Creates IPath from the last remopla file name
		IPath path = new Path(MCLaunchConfigurationDelegate.getRemoplaFileName());
		logger.debug("path.toString(): " + path.toString());
		
		// Finds IFile for IPath inside the workspace
		IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
		IFile file = root.getFileForLocation(path);
		try {
			file.refreshLocal(IResource.DEPTH_INFINITE, null);
		} catch (CoreException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		logger.debug("file.getName(): " + file.getName());
		logger.debug("file.getFullPath(): " + file.getFullPath());
		logger.debug("file.exists(): " + file.exists());
		
		// Creates an editor input
		IEditorInput input = new FileEditorInput(file);
		
		// Opens the editor
		IWorkbench workbench = PlatformUI.getWorkbench();
		IWorkbenchWindow window = workbench.getActiveWorkbenchWindow();
		IWorkbenchPage page = window.getActivePage();
		try {
			page.openEditor(input, "org.eclipse.ui.DefaultTextEditor");
		} catch (PartInitException e) {
			e.printStackTrace();
		}
	}
	
	private void createTabs(Composite parent)
	{
		tabControl = new TabControl(parent, SWT.NONE);
		tabControl.setLayoutData(new GridData(GridData.FILL_BOTH));
		if (memento != null) {
			tabControl.init(memento);
		}
	}
	
	private Group createGroup(Composite parent, String title)
	{
		Group group = new Group(parent, SWT.NONE);
		group.setLayout(new GridLayout());
		group.setText(title);
		group.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
		return group;
	}

	private void contributeToActionBars() {
		IActionBars bars = getViewSite().getActionBars();
		fillLocalPullDown(bars.getMenuManager());
		fillLocalToolBar(bars.getToolBarManager());
	}

	private void fillLocalPullDown(IMenuManager manager) {
		manager.add(stopAction);
		manager.add(new Separator());
	}

	private void fillContextMenu(IMenuManager manager) {
		manager.add(stopAction);
		manager.add(new Separator());
		// Other plug-ins can contribute there actions here
		manager.add(new Separator(IWorkbenchActionConstants.MB_ADDITIONS));
	}
	
	private void fillLocalToolBar(IToolBarManager manager) {
		manager.add(stopAction);
		manager.add(new Separator());
	}

	private void makeActions() 
	{
		stopAction = new StopAction();
	}

	public void setFocus() 
	{

	}
	
	private ILaunchConfiguration getLaunchConfiguration()
	{
		ILaunchConfiguration config = launch.getLaunchConfiguration();
		try {
			ILaunchConfigurationWorkingCopy copy = config.getWorkingCopy();
			configControl.applySettings(copy);
			return copy.doSave();
		}
		catch (CoreException ce) {
			MCPlugin.log(ce);
		}
		return config;
	}
	
	private void analyseButtonPressed()
	{
		if (launch != null) {
			DebugUITools.launch(getLaunchConfiguration(), launch.getLaunchMode());
		}
	}
	
	public void setTraceTree(TraceNode[] roots)
	{
		tabControl.setTraceTree(roots);
	}
	
	private void setEnabledRecursively(Control control, boolean enabled)
	{
		control.setEnabled(enabled);
		if (control instanceof Composite) {
			for (Control c : ((Composite)control).getChildren()) {
				setEnabledRecursively(c, enabled);
			}
		}
	}
	
	private class StopAction extends Action
	{
		public StopAction()
		{
			setText("Stop");
			setToolTipText("Stops the current analysis");
			setImageDescriptor(MCPlugin.getEnabledImage("stop.gif"));
			setDisabledImageDescriptor(MCPlugin.getDisabledImage("stop.gif"));
			setHoverImageDescriptor(MCPlugin.getHoverImage("stop.gif"));
		}
		
		@Override
		public void run() 
		{
			monitor.setCanceled(true);
		}
	}
	
	private void dispatch(final Runnable r)
	{
		if (!progressBar.isDisposed()) {
			progressBar.getDisplay().asyncExec(new Runnable() {
				public void run() {
					if (!progressBar.isDisposed()) {
						r.run();
					}
				}
			});
		}
	}
	
	private class ViewProgressMonitor implements IProgressMonitor, ProgressMonitor
	{

		private volatile boolean canceled = false;
		private long cancelTime = 0;
		private long startTime;
		private String mainTask = null;
		
		public void beginTask(final String name, final int totalWork) 
		{
			dispatch(new Runnable() {
				public void run() {
					progressLabel.setText(name);
					mainTask = name;
					progressBar.setMaximum(totalWork);
					progressBar.setSelection(0);
					stopAction.setEnabled(true);
					stopButton.setEnabled(true);
					startTime = System.currentTimeMillis();
				}
			});
		}

		public void done() 
		{
			dispatch(new Runnable() {
				public void run() {
					float secs = (float) ((System.currentTimeMillis() - startTime) / 1000.0);
					if (isCanceled()) {
						progressLabel.setText("Canceled. " + secs + "s");
					}
					else {
						progressLabel.setText("Finished. " + secs + "s");
						progressBar.setSelection(progressBar.getMaximum());
					}
					stopAction.setEnabled(false);
					stopButton.setEnabled(false);
				}
			});
		}

		public void internalWorked(double work) 
		{

		}

		public boolean isCanceled() 
		{
			return canceled;
		}

		public void setCanceled(boolean value) 
		{
			canceled = value;
		}

		public void setTaskName(final String name) 
		{
			dispatch(new Runnable() {
				public void run() {
					progressLabel.setText(name);
				}
			});
		}

		public void subTask(final String name) 
		{
			dispatch(new Runnable() {
				public void run() {
					String label = mainTask + " - " + name;
					progressLabel.setText(label);
				}
			});	
		}

		public void worked(final int work) 
		{
			dispatch(new Runnable() {
				public void run() {
					progressBar.setSelection(progressBar.getSelection() + work);
				}
			});
		}
		
	}

	public boolean isCreated() 
	{
		return progressBar != null;
	}

	public void setLaunch(ILaunch launch) 
	{
		ILaunchConfiguration config = launch.getLaunchConfiguration();
		//configControl.setFrom(config);
		//setEnabledRecursively(paramGroup, true);
		setEnabledRecursively(actionComposite, true);
		
		titleLabel.setText(config.getName());
		this.launch = launch;
	}
	
	private class ViewCoverageListener extends CoverageAdapter
	{

		@Override
		public void coverageStarted(PDS pds) 
		{
			dispatch(new Runnable() {
				public void run() {
					tabControl.clearTraceTree();
					tabControl.setResult(null);
				}
			});
		}

		@Override
		public void coverageEnded(final CoverageResult result)
		{
			dispatch(new Runnable() {
				public void run() {
					tabControl.setResult(result);
				}
			});
		}

		@Override
		public void coverageFailed(String message, Throwable t)
		{
			// TODO Auto-generated method stub
			
		}

	}

}
