package de.uni_stuttgart.fmi.szs.jmoped.eclipse;

import org.eclipse.core.filebuffers.FileBuffers;
import org.eclipse.core.filebuffers.ITextFileBuffer;
import org.eclipse.core.filebuffers.ITextFileBufferManager;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.WorkingCopyOwner;
import org.eclipse.jdt.core.search.IJavaSearchConstants;
import org.eclipse.jdt.core.search.SearchEngine;
import org.eclipse.jdt.core.search.SearchPattern;
import org.eclipse.jdt.core.search.TypeNameRequestor;
import org.eclipse.jdt.internal.corext.util.TypeInfo;
import org.eclipse.jdt.internal.corext.util.TypeInfoFactory;
import org.eclipse.jface.text.IDocument;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod;

public class SearchUtils 
{

	// TODO caching
		
	public interface IDocumentRequestor
	{
		void acceptDocument(IDocument document);
		
		void documentNotFound();
	}
	
	public static IResource searchResource(String className)
	{
		try {
			 ResourceRequestor requestor = new ResourceRequestor();
			 searchMethod(className, requestor);
			 return requestor.resource;
		}
		catch (JavaModelException e) {
		}
		return null;
	}
	
	public static IDocument searchDocument(String className) 
	{
		try {
			 ClassNameRequestor requestor = new ClassNameRequestor();
			 searchMethod(className, requestor);
			 return requestor.document;
		}
		catch (JavaModelException e) {
		}
		return null;
	}
	
	public static void searchMethod(String className,
			TypeNameRequestor requestor) throws JavaModelException
	{
		SearchEngine searchEngine = new SearchEngine((WorkingCopyOwner)null);
		searchEngine.searchAllTypeNames(getPackageBytes(className),
					getClassNameBytes(className),
					SearchPattern.R_EXACT_MATCH | SearchPattern.R_CASE_SENSITIVE,
					IJavaSearchConstants.CLASS, SearchEngine.createWorkspaceScope(),
					requestor, 0, null);
	}
	
	private static char[] getClassNameBytes(String className) 
	{
		int index = className.lastIndexOf('/');
		if (index >= 0) {
			return className.substring(index + 1).toCharArray();
		}
		else {
			return className.toCharArray();
		}
	}

	private static char[] getPackageBytes(String className) 
	{
		int index = className.lastIndexOf('/');
		if (index >= 0) {
			return className.substring(0, index).replace('/', '.').toCharArray();
		}
		return new char[0];
	}
	
	public static IDocument getDocument(IPath path)
	{
		ITextFileBufferManager manager = FileBuffers.getTextFileBufferManager();
		ITextFileBuffer buffer = manager.getTextFileBuffer(path);
		if (buffer != null) {
			return buffer.getDocument();
		}
		return null;
	}
	
	private static class ClassNameRequestor extends TypeNameRequestor
	{
		public IDocument document;
		
		@Override
		public void acceptType(int modifiers, char[] packageName, 
				char[] simpleTypeName, char[][] enclosingTypeNames, String path) 
		{
			document = getDocument(new Path(path));
		}
	}
	
	private static class ResourceRequestor extends TypeNameRequestor
	{
		public IResource resource;
		
		@Override
		public void acceptType(int modifiers, char[] packageName, char[] simpleTypeName, 
				char[][] enclosingTypeNames, String path) 
		{
			TypeInfoFactory factory = new TypeInfoFactory();
			TypeInfo typeInfo = factory.create(packageName, simpleTypeName, 
					enclosingTypeNames, modifiers, path);
			try {
				IType type = typeInfo.resolveType(SearchEngine.createWorkspaceScope());
				resource = (IResource)type.getAdapter(IResource.class);
			}
			catch (JavaModelException jme) {
			}
		}
	}
	
	
	
}
