package de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jdt.core.IBuffer;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.ISourceRange;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.formatter.CodeFormatter;
import org.eclipse.jdt.internal.corext.codemanipulation.StubUtility;
import org.eclipse.jdt.internal.corext.util.CodeFormatterUtil;
import org.eclipse.jdt.internal.corext.util.Strings;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.TraceNode;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;

public class AddTestCaseCreationWizardPage extends NewTestCaseWizardPage
{

	private IType type; 
	
	public AddTestCaseCreationWizardPage(TraceNode[] traceTree,
			IJavaElement element, String methodName, IType type) 
	{
		super(traceTree, element, methodName);
		this.type = type;
	}
	
	@Override
	public IType getCreatedType() 
	{
		return type;
	}
	
	@Override
	public void createControl(Composite parent) 
	{
		initializeDialogUnits(parent);
		
		Composite composite = new Composite(parent, SWT.NONE);
		composite.setFont(parent.getFont());
		
		int nColumns = 4;
		GridLayout layout = new GridLayout();
		layout.numColumns= nColumns;		
		composite.setLayout(layout);
		
		// pick & choose the wanted UI components
		
		createMethodNameControls(composite, nColumns);
		
		setControl(composite);
			
		Dialog.applyDialogFont(composite);
	}
	
	/**
	 * Overriden to work around missing ImportsManager.
	 */
	@Override
	protected String getSourceType(String typeName, ImportsManager imports) 
	{
		typeName = typeName.replace('/', '.');
		IPackageFragment pack = 
			(IPackageFragment) type.getAncestor(IJavaElement.PACKAGE_FRAGMENT);
		if (pack != null && !pack.isDefaultPackage()){
			String packageName = pack.getElementName();
			if (typeName.startsWith(packageName)) {
				typeName = typeName.substring(packageName.length() + 1);
			}
		}
		return typeName;
	}
	
	@Override
	protected boolean methodExists(String methodName) 
	{
		try {
			for (IMethod method : type.getMethods()) {
				if (method.getElementName().equals(methodName)) {
					return true;
				}
			}
		}
		catch (JavaModelException jve) {
			MCPlugin.log(jve);
		}
		return false;
	}
	
	@Override
	public void createType(IProgressMonitor monitor) 
		throws CoreException, InterruptedException 
	{
		if (monitor == null) {
			monitor = new NullProgressMonitor();
		}
		
		monitor.beginTask("Add Test Case", 10);
		
		ICompilationUnit cu = type.getCompilationUnit();
		createTypeMembers(type, (ImportsManager)null, new SubProgressMonitor(monitor, 5));
		
		ISourceRange range= type.getSourceRange();
		
		IBuffer buf = cu.getBuffer();
		String originalContent= buf.getText(range.getOffset(), range.getLength());
		int indent = StubUtility.getIndentUsed(type) + 1;
		String lineDelimiter= StubUtility.getLineDelimiterUsed(type.getJavaProject());
		
		String formattedContent = 
			CodeFormatterUtil.format(CodeFormatter.K_CLASS_BODY_DECLARATIONS, 
					originalContent, indent, null, lineDelimiter, type.getJavaProject());
		formattedContent= Strings.trimLeadingTabsAndSpaces(formattedContent);
		buf.replace(range.getOffset(), range.getLength(), formattedContent);
		
		monitor.done();
	}

}
