package de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.Viewer;
import org.gjt.jclasslib.bytecode.AbstractInstruction;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;

import de.uni_stuttgart.fmi.szs.jmoped.BytecodeUtils;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.AnnotatedInstruction;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Line;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Transition;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;

public class CoverageMethodTableContentProvider implements IStructuredContentProvider
{

	public Object[] getElements(Object inputElement) 
	{
		CoverageMethod[] methods = (CoverageMethod[])inputElement;
		ArrayList<Object> elements = new ArrayList<Object>(methods.length * 4);
		for (CoverageMethod method : methods) {
			if (method.clinitIsAssertionOnly()) {
				continue;
			}
			elements.add(method);
			for (Line line : method.getLines()) {
				for (State state : State.values()) {
					if (state.isException()) {
						List<Transition> transitions = line.getTransitions(state);
						for (Transition t : transitions) {
							elements.add(new TransitionState(t, state, line));
						}
					}
					else {
						List<Transition> transitions = line.getTransitions(state);
						if (!transitions.isEmpty()) {
							elements.add(new TransitionState(transitions.get(0), state, line));
						}
					}
				}
				// collect uncovered instructions
				for (Iterator<AnnotatedInstruction> iter =  line.getInstructions().iterator();
					iter.hasNext();) {
					AnnotatedInstruction instruction = iter.next();
					if (instruction.getTransitions(State.COVERED).isEmpty()) {
						if (!instruction.getTransitions(State.ASSERTION_FAILED).isEmpty()) {
							if (iter.hasNext()) {
								iter.next();
							}
						}
						else {
							elements.add(new InstructionState(instruction, State.NOT_COVERED, 
									line));
						}
					}
				}
			}
		}
		return elements.toArray(new Object[elements.size()]);
	}

	private boolean isAssertionErrorConstructor(ClassFile classFile, AbstractInstruction instruction)
	{
		try {
			return BytecodeUtils.isAssertionErrorConstructor(classFile, instruction);
		} 
		catch (InvalidByteCodeException e) {
			return false;
		}
	}
	
	public void dispose() {
		// TODO Auto-generated method stub
		
	}

	public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
		// TODO Auto-generated method stub
		
	}

}
