package de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.ToolBarManager;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.jface.viewers.ColumnLayoutData;
import org.eclipse.jface.viewers.ColumnPixelData;
import org.eclipse.jface.viewers.ColumnWeightData;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.TableLayout;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ViewForm;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.ToolBar;
import org.eclipse.ui.IMemento;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageResult;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Line;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;

public class CoverageMethodTableControl extends Composite 
{

	private TableViewer coverageViewer;
	
	private ToolBar toolBar;
	
	private ViewForm form;
	
	private CoverageMethodTreeLabelProvider labelProvider;
	
	/**
	 * Filter for markers
	 */
	private StateFilterAction[] actions;
	
	private ColumnLayoutData[] defaultColumns = new ColumnLayoutData[] {
			new ColumnWeightData(3, true),
			new ColumnWeightData(1, true),
			new ColumnPixelData(16, false, false),
	};
	
	private static String COLUMN_WIDTH = MCPlugin.PLUGIN_ID 
		+ ".coverageTableColumnWidth";
	
	public CoverageMethodTableControl(Composite parent, int style)
	{
		super(parent, style);
		setLayout(new FillLayout());
		
		form = new ViewForm(this, SWT.NONE);
		
		toolBar = new ToolBar(form, SWT.FLAT | SWT.HORIZONTAL);
		form.setTopCenter(toolBar);
		
		IToolBarManager manager = new ToolBarManager(toolBar);
		
		State[] states = State.values();
		actions = new StateFilterAction[states.length];
		FilterActionPropertyListener listener = new FilterActionPropertyListener();
				
		for (int i = 0; i < states.length; i++) {
			StateFilterAction action = new StateFilterAction(states[i]);
			action.addPropertyChangeListener(listener);
			manager.add(action);
			actions[i] = action;
		}
		manager.update(true);
	
		coverageViewer = new TableViewer(form, SWT.SINGLE);
		
		Table table = coverageViewer.getTable();
		
		new TableColumn(table, SWT.LEFT).setText("Name");
		new TableColumn(table, SWT.RIGHT).setText("Line");
		TableColumn column = new TableColumn(table, SWT.CENTER);
		
		table.setHeaderVisible(true);
		
		table.addMouseListener(new DoubleClickListener());
		
		form.setContent(table);
		coverageViewer.setContentProvider(new CoverageMethodTableContentProvider());
		labelProvider = new CoverageMethodTreeLabelProvider();
		coverageViewer.setLabelProvider(labelProvider);
		
		createContextMenu();
	}
	
	private void createContextMenu()
	{
		MenuManager menuMgr = new MenuManager("#PopupMenu");
        menuMgr.setRemoveAllWhenShown(true);
        Menu menu = menuMgr.createContextMenu(coverageViewer.getControl());
        coverageViewer.getControl().setMenu(menu);
        
        menuMgr.addMenuListener(new IMenuListener() {

			public void menuAboutToShow(IMenuManager manager) 
			{
				fillContextMenu(manager);
			}
        	
        });
	}

	public void setResult(CoverageResult result)
	{
		if (result != null) {
			labelProvider.setResult(result);
			coverageViewer.setInput(result.getMethods());
		}
		else {
			labelProvider.setResult(null);
			coverageViewer.setInput(null);
		}
	}
	

	public void saveState(IMemento memento) 
	{
		Table table = coverageViewer.getTable();
		if (table == null || table.isDisposed() || table.getBounds().width == 0) {
			return;
		}
		TableColumn[] columns = table.getColumns();
		for (int i = 0; i < columns.length; i++) {
			if (defaultColumns[i].resizable && columns[i].getWidth() > 0) {
				memento.putInteger(COLUMN_WIDTH + i, columns[i].getWidth());
			}
		}
	}

	public void init(IMemento memento) 
	{
		Table table = coverageViewer.getTable();
		TableLayout layout = new TableLayout();
		table.setLayout(layout);
		
		if (table == null || table.isDisposed()) {
			return;
		}
		for (int i = 0; i < defaultColumns.length; i++) {
			if (defaultColumns[i].resizable) {
				Integer width = memento.getInteger(COLUMN_WIDTH + i);
				if (width != null) {
					layout.addColumnData(new ColumnPixelData(width, true, false));
					continue;
				}
			}
			layout.addColumnData(defaultColumns[i]);
		}
	}

	
	private void fillContextMenu(IMenuManager manager)
	{
		IStructuredSelection sel = (IStructuredSelection)coverageViewer.getSelection();
		if (!sel.isEmpty()) {
			Object first = sel.getFirstElement();
			if (first instanceof CoverageMethod) {
				CoverageMethod method = (CoverageMethod)first;
				manager.add(new OpenMethodInEditorAction(method.getClassName(), 
						method.getLines()[0].getLineNumber()));
			}
			else if (first instanceof Line) {
				Line line = (Line)first;
				manager.add(new OpenMethodInEditorAction(line.getMethod().getClassName(), 
						line.getLineNumber()));;
			}
			else if (first instanceof TransitionState) {
				TransitionState trs = (TransitionState)first;
				manager.add(new OpenMethodInEditorAction(trs.getLine().getMethod().getClassName(), 
						trs.getLine().getLineNumber()));
			}
			else if (first instanceof InstructionState) {
				InstructionState trs = (InstructionState)first;
				manager.add(new OpenMethodInEditorAction(trs.getLine().getMethod().getClassName(), 
						trs.getLine().getLineNumber()));
			}
		}
	}
	
	private class StateFilterAction extends Action
	{
		private CoverageMethodStateFilter filter;
		
		public StateFilterAction(State state)
		{
			super("Filter for " + state, IAction.AS_CHECK_BOX);
			setToolTipText(getText());
			setImageDescriptor(MCPlugin.getImage(state));
			filter = new CoverageMethodStateFilter(state);
		}
		
		@Override
		public void setChecked(boolean checked) 
		{
			if (checked) {
				coverageViewer.addFilter(filter);
			}
			else {
				coverageViewer.removeFilter(filter);
			}
			super.setChecked(checked);
		}
	}
	
	private class FilterActionPropertyListener implements IPropertyChangeListener
	{

		public void propertyChange(PropertyChangeEvent event) 
		{
			// pointer comparison should be fine
			if (event.getProperty() == IAction.CHECKED 
					&& event.getNewValue() == Boolean.TRUE) {
				Object source = event.getSource();
				for (StateFilterAction action : actions) {
					if (action != source && action.isChecked()) {
						action.setChecked(false);
					}
				}
			}
		}
		
	}
	
	private class DoubleClickListener extends MouseAdapter
	{
		@Override
		public void mouseDoubleClick(MouseEvent e) 
		{
			IStructuredSelection sel = (IStructuredSelection)coverageViewer.getSelection();
			if (!sel.isEmpty()) {
				Object first = sel.getFirstElement();
				if (first instanceof CoverageMethod) {
					CoverageMethod method = (CoverageMethod)first;
					new OpenMethodInEditorAction(method.getClassName(), 
							method.getLines()[0].getLineNumber()).run();
				}
				else if (first instanceof Line) {
					Line line = (Line)first;
					new OpenMethodInEditorAction(line.getMethod().getClassName(), 
							line.getLineNumber()).run();
				}
				else if (first instanceof TransitionState) {
					TransitionState trs = (TransitionState)first;
					new OpenMethodInEditorAction(trs.getLine().getMethod().getClassName(), 
							trs.getLine().getLineNumber()).run();
				}
				else if (first instanceof InstructionState) {
					InstructionState trs = (InstructionState)first;
					new OpenMethodInEditorAction(trs.getLine().getMethod().getClassName(), 
							trs.getLine().getLineNumber()).run();
				}
			}
		}
	}
	
}
