package de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.Viewer;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.AnnotatedInstruction;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Line;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Transition;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;

public class CoverageMethodTreeContentProvider implements ITreeContentProvider
{

	public Object[] getChildren(Object parentElement) 
	{
		if (parentElement instanceof CoverageMethod) {
			CoverageMethod method = (CoverageMethod)parentElement;
			return method.getLines();
		}
		else if (parentElement instanceof Line) {
			Line line = (Line)parentElement;
			ArrayList<TransitionState> list = new ArrayList<TransitionState>();
			for (State state : State.values()) {
				if (state.isException()) {
					List<Transition> transitions = line.getTransitions(state);
					for (Transition t : transitions) {
						list.add(new TransitionState(t, state, line));
					}
				}
			}
			return list.toArray(new Object[list.size()]);
		}
		return null;
	}

	public Object getParent(Object element) 
	{
		if (element instanceof Line) {
			return ((Line)element).getMethod();	
		}
		else if (element instanceof AnnotatedInstruction) {
			AnnotatedInstruction instruction = (AnnotatedInstruction)element;
			CoverageMethod method = instruction.getMethod();
			return method.getLines()[method.getLineOffset(instruction.getOffset())];
		}
		return null;
	}

	public boolean hasChildren(Object element) 
	{
		return element instanceof CoverageMethod || element instanceof Line;
	}

	public Object[] getElements(Object inputElement) 
	{
		return (CoverageMethod[])inputElement;
	}

	public void dispose() 
	{
		
	}

	public void inputChanged(Viewer viewer, Object oldInput, Object newInput) 
	{
		
	}

}
