package de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.ToolBarManager;
import org.eclipse.jface.viewers.ColumnPixelData;
import org.eclipse.jface.viewers.ColumnWeightData;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.TableLayout;
import org.eclipse.jface.viewers.TableTreeViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ViewForm;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.ToolBar;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageResult;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Line;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;

public class CoverageMethodTreeControl extends Composite 
{

	private TableTreeViewer coverageViewer;
	
	private ToolBar toolBar;
	
	private ViewForm form;
	
	private CoverageMethodTreeLabelProvider labelProvider;
	
	public CoverageMethodTreeControl(Composite parent, int style)
	{
		super(parent, style);
		setLayout(new FillLayout());
		
		form = new ViewForm(this, SWT.NONE);
		
		toolBar = new ToolBar(form, SWT.FLAT | SWT.HORIZONTAL);
		form.setTopCenter(toolBar);
		
		IToolBarManager manager = new ToolBarManager(toolBar);
		for (State s : State.values()) {
			StateFilterAction action = new StateFilterAction(s);
			manager.add(action);
		}
		manager.update(true);
	
		coverageViewer = new TableTreeViewer(form, SWT.VIRTUAL | SWT.SINGLE);
		
		Table table = coverageViewer.getTableTree().getTable();
		TableLayout layout = new TableLayout();
		table.setLayout(layout);
		
		new TableColumn(table, SWT.LEFT).setText("Name");
		layout.addColumnData(new ColumnWeightData(3, true));
		new TableColumn(table, SWT.RIGHT).setText("Line");
		layout.addColumnData(new ColumnWeightData(1, true));
		TableColumn column = new TableColumn(table, SWT.CENTER);
		layout.addColumnData(new ColumnPixelData(16, false, false));
		
		table.setHeaderVisible(true);
		
		coverageViewer.getTableTree().getTable().addMouseListener(new DoubleClickListener());
		
		form.setContent(coverageViewer.getTableTree());
		coverageViewer.setContentProvider(new CoverageMethodTreeContentProvider());
		labelProvider = new CoverageMethodTreeLabelProvider();
		coverageViewer.setLabelProvider(labelProvider);
	}

	public void setResult(CoverageResult result)
	{
		if (result != null) {
			labelProvider.setResult(result);
			coverageViewer.setInput(result.getMethods());
		}
		else {
			labelProvider.setResult(null);
			coverageViewer.setInput(null);
		}
	}
	
	private class StateFilterAction extends Action
	{
		private CoverageMethodStateFilter filter;
		
		public StateFilterAction(State state)
		{
			super("Filter " + state, IAction.AS_CHECK_BOX);
			setToolTipText(getText());
			setImageDescriptor(MCPlugin.getImage(state));
			filter = new CoverageMethodStateFilter(state);
		}
		
		@Override
		public void run() 
		{
			if (isChecked()) {
				coverageViewer.addFilter(filter);
			}
			else {
				coverageViewer.removeFilter(filter);
			}
		}
	}
	
	private class DoubleClickListener extends MouseAdapter
	{
		@Override
		public void mouseDoubleClick(MouseEvent e) 
		{
			IStructuredSelection sel = (IStructuredSelection)coverageViewer.getSelection();
			if (!sel.isEmpty()) {
				Object first = sel.getFirstElement();
				if (first instanceof CoverageMethod) {
					CoverageMethod method = (CoverageMethod)first;
					new OpenMethodInEditorAction(method.getClassName(), 
							method.getLines()[0].getLineNumber()).run();
				}
				else if (first instanceof Line) {
					Line line = (Line)first;
					new OpenMethodInEditorAction(line.getMethod().getClassName(), 
							line.getLineNumber()).run();
				}
				else if (first instanceof TransitionState) {
					TransitionState trs = (TransitionState)first;
					
				}
			}
		}
	}
	
}
