package de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui;

import java.util.ArrayList;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IAction;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;

public class HeapOverFlowControl extends Composite
{

	private List heapOverflowList;
	
	private IAction jumpToSourceAction = new JumpToSourceAction();
	
	private ArrayList<MethodOffsetPair> model = new ArrayList<MethodOffsetPair>();
	
	public HeapOverFlowControl(Composite parent, int style) 
	{
		super(parent, style);

		setLayout(new FillLayout());
		
		heapOverflowList = new List(this, SWT.SINGLE);
		
		DoubleClickListener listener = new DoubleClickListener();
		heapOverflowList.addMouseListener(listener);
		
		Menu contextMenu = new Menu(heapOverflowList);
		MenuItem item = new MenuItem(contextMenu, SWT.PUSH);
		item.setText("Jump to Source");
		item.addSelectionListener(listener);
		
		heapOverflowList.setMenu(contextMenu);
	}
	
	public void clear()
	{
		heapOverflowList.removeAll();
		model.clear();
	}
	
	private void jumpToSource() 
	{
		int index = heapOverflowList.getSelectionIndex();
		if (index != -1) {
			MethodOffsetPair pair = model.get(index);
			CoverageMethod method = pair.method;
			new OpenMethodInEditorAction(method.getClassName(), 
					method.getLineNumber(pair.lineOffset)).run();
		}
	}
	
	public void add(CoverageMethod method, int lineOffset)
	{
		heapOverflowList.add(method.getHumanReadableName() + ":" 
				+ method.getLineNumber(lineOffset));
		model.add(heapOverflowList.getItemCount() - 1, 
				new MethodOffsetPair(method, lineOffset));
	}
	
	private static class MethodOffsetPair
	{
		CoverageMethod method;
		int lineOffset;
		
		public MethodOffsetPair(CoverageMethod method, int offset) 
		{
			this.method = method;
			lineOffset = offset;
		}
	}
	
	private class DoubleClickListener extends MouseAdapter
		implements SelectionListener
	{
		@Override
		public void mouseDoubleClick(MouseEvent e) 
		{
			jumpToSource();
		}
		
		

		public void widgetSelected(SelectionEvent e) 
		{
			jumpToSource();		
		}

		public void widgetDefaultSelected(SelectionEvent e) 
		{
					
		}
	}
	
	private class JumpToSourceAction extends Action
	{
			public JumpToSourceAction() 
			{
				super("Jump to Source", MCPlugin.getEnabledImage("stop.gif"));
			}
			
			@Override
			public void run() 
			{
				jumpToSource();
			}
	}
	
}
