package de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui;

import org.apache.log4j.Logger;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchConfigurationWorkingCopy;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Spinner;
import org.eclipse.ui.IMemento;

import de.uni_stuttgart.fmi.szs.jmoped.PDSDefault;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;

public class LaunchConfigurationControl extends Composite {
	
	private Spinner bitsSpinner;
	private Spinner heapSizeSpinner;
	
	public static final int ONE_BIT_HEAP_CHECKBOX = 0;
	
	private Button[] checkboxes;
	private static enum CheckboxType { 
		ONE_BIT_HEAP, CONTINUE_AFTER_ERROR, CHECK_EXCEPION }
	private Button oneBitHeapCheckbox;
	//private Button checkInitialMethodCheckBox;
	//private Button continueAfterAssertionErrorCheckBox;
	private Button checkExceptionsCheckBox;
	
	private static Logger logger = Logger.getLogger(LaunchConfigurationControl.class);
	
	public LaunchConfigurationControl(Composite parent, int style) {
		
		this(parent, style, null);
	}
	
	public LaunchConfigurationControl(Composite parent, int style, 
			IMemento memento) 
	{
		super(parent, style);
		setLayout(new GridLayout(2, false));
		
		// Adds bits spinner
		new Label(this, SWT.HORIZONTAL).setText("Number of Bits: ");
		bitsSpinner = SWTUtils.createBitsSpinner(this, memento);

		// Adds heap size spinner
		new Label(this, SWT.HORIZONTAL).setText("Heap Size: ");
		heapSizeSpinner = SWTUtils.createHeapSizeSpiner(this, memento);
		
		// Adds listeners to spinners
		SWTUtils.installSpinnerSyncListener(bitsSpinner, heapSizeSpinner);
		
		// Adds checkboxes
		//checkInitialMethodCheckBox = SWTUtils.createCheckbox(this, memento, 
		//		"Limit coverage to initial method",
		//		MCPlugin.ATTR_CHECK_INITIAL_METHOD_ONLY, false);
		oneBitHeapCheckbox = SWTUtils.createCheckbox(this, memento,
				"Use one-bit heap", MCPlugin.ATTR_ONE_BIT_HEAP, false);
//		continueAfterAssertionErrorCheckBox = SWTUtils.createCheckbox(this, 
//				memento, "Continue after assertion error",
//				MCPlugin.ATTR_CONTINUE_AFTER_ASSERTION_ERROR, true);
		checkExceptionsCheckBox = SWTUtils.createCheckbox(this, memento,
				"Check for exceptions", MCPlugin.ATTR_CHECK_FOR_EXCEPTION, true);
		
		// Adds listeners to checkboxes
		//SWTUtils.installCheckBoxSyncListener(oneBitHeapCheckbox,
		//		continueAfterAssertionErrorCheckBox, checkExceptionsCheckBox);
	}
	
	public int getBits() {
		
		return bitsSpinner.getSelection();
	}
	
	public void setBits(int bits) {
		
		bitsSpinner.setSelection(bits);
	}
	
	public int getHeapSize() {
		
		return heapSizeSpinner.getSelection();
	}
	
	public boolean useOneBitHeap() {
		
		return oneBitHeapCheckbox.getSelection();
	}
	
	//public boolean checkInitialMethodOnly() {
		
	//	return checkInitialMethodCheckBox.getSelection();
	//}
	
//	public boolean continueAfterAssertionError() {
//		
//		return continueAfterAssertionErrorCheckBox.getSelection();
//	}
	
	public boolean checkExceptions() {
		
		return checkExceptionsCheckBox.getSelection();
	}
	
	public void setFrom(ILaunchConfiguration conf)
	{
		try {
			int bits = conf.getAttribute(MCPlugin.ATTR_BITS, PDSDefault.BITS);
			setBits(bits);
			
			int heapSize = conf.getAttribute(MCPlugin.ATTR_HEAP_SIZE,
					PDSDefault.HEAP_SIZE);
			heapSizeSpinner.setSelection(heapSize);
			
			boolean selected;
			selected = conf.getAttribute(MCPlugin.ATTR_ONE_BIT_HEAP, false);
			oneBitHeapCheckbox.setSelection(selected);
			//selected =
			//	configuration.getAttribute(MCPlugin.ATTR_CHECK_INITIAL_METHOD_ONLY, false);
			//checkInitialMethodCheckBox.setSelection(selected);
			
//			selected = 
//				conf.getAttribute(MCPlugin.ATTR_CONTINUE_AFTER_ASSERTION_ERROR, true);
//			continueAfterAssertionErrorCheckBox.setSelection(selected);
			
			selected = 
				conf.getAttribute(MCPlugin.ATTR_CHECK_FOR_EXCEPTION, true);
			checkExceptionsCheckBox.setSelection(selected);
		}
		catch (CoreException e) {
			MCPlugin.log(e);
		}
	}
	
	public void applySettings(ILaunchConfigurationWorkingCopy conf)
	{
		conf.setAttribute(MCPlugin.ATTR_BITS, bitsSpinner.getSelection());
		conf.setAttribute(MCPlugin.ATTR_HEAP_SIZE, heapSizeSpinner.getSelection());
		conf.setAttribute(MCPlugin.ATTR_ONE_BIT_HEAP, 
				oneBitHeapCheckbox.getSelection());
		//configuration.setAttribute(MCPlugin.ATTR_CHECK_INITIAL_METHOD_ONLY, 
		//		checkInitialMethodCheckBox.getSelection());
//		conf.setAttribute(MCPlugin.ATTR_CONTINUE_AFTER_ASSERTION_ERROR,
//				continueAfterAssertionErrorCheckBox.getSelection());
		conf.setAttribute(MCPlugin.ATTR_CHECK_FOR_EXCEPTION, 
				checkExceptionsCheckBox.getSelection());
	}
	
	public void addSettingsChangedListener(final SettingsChangedListener listener)
	{
		ModifyListener modifyListener = new ModifyListener() {
			public void modifyText(org.eclipse.swt.events.ModifyEvent e) 
			{
				listener.settingChanged();
			}
		};
		bitsSpinner.addModifyListener(modifyListener);
		heapSizeSpinner.addModifyListener(modifyListener);
		
		SelectionListener selectionListener = new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) 
			{
				listener.settingChanged();
			}
		};
		oneBitHeapCheckbox.addSelectionListener(selectionListener);
		//checkInitialMethodCheckBox.addSelectionListener(selectionListener);
		//continueAfterAssertionErrorCheckBox.addSelectionListener(selectionListener);
		checkExceptionsCheckBox.addSelectionListener(selectionListener);
	}
	
	public static interface SettingsChangedListener
	{
		void settingChanged();
	}
}
