package de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui;

import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.gjt.jclasslib.structures.attributes.CodeAttribute;
import org.gjt.jclasslib.structures.attributes.LineNumberTableAttribute;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.MethodCall;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.TraceNode;

public class MethodCallTraceTreeControl extends Composite 
{
	
	private TreeViewer treeViewer;

	public MethodCallTraceTreeControl(Composite parent, int style)
	{
		super(parent, style);
		setLayout(new FillLayout());
		
		treeViewer = new TreeViewer(this, SWT.VIRTUAL | SWT.SINGLE);
		treeViewer.setContentProvider(new MethodCallTraceTreeContentProvider());
		treeViewer.setLabelProvider(new MethodCallTraceTreeLabelProvider());
		treeViewer.getTree().addMouseListener(new DoubleClickListener());
	}

	public void setTraceTree(TraceNode[] roots) 
	{
		treeViewer.setInput(roots);
	}
	
	private class DoubleClickListener extends MouseAdapter
	{
		@Override
		public void mouseDoubleClick(MouseEvent e) 
		{
			IStructuredSelection sel = (IStructuredSelection)treeViewer.getSelection();
			if (!sel.isEmpty()) {
				Object first = sel.getFirstElement();
				if (first instanceof TraceNode) {
					TraceNode node = (TraceNode)first;
					MethodCall call = node.getMethodCall();
					CodeAttribute code = 
						(CodeAttribute)call.getMethod().getMethodInfo().findAttribute(CodeAttribute.class);
					LineNumberTableAttribute table = (LineNumberTableAttribute)
						code.findAttribute(LineNumberTableAttribute.class);
					int line = table.getLineNumberTable()[0].getLineNumber();
					new OpenMethodInEditorAction(
							call.getMethod().getPDSClass().getName(), line)
							.run();
				}
			}
		}
	}


}
