package de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.internal.ui.wizards.NewElementWizard;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.TraceNode;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;

public class NewTestCaseCreationWizard extends NewElementWizard 
{

	private NewTestCaseWizardPage fPage;
	
	private TraceNode[] traceTree;
	
	protected IJavaElement element;
	
	private String methodName;
	
	
	
	public NewTestCaseCreationWizard(TraceNode[] traceTree, IJavaElement element,
			String methodName) 
	{
		this.traceTree = traceTree;
		this.element = element;
		this.methodName = methodName;
//		setDefaultPageImageDescriptor(JavaPluginImages.DESC_WIZBAN_NEWCLASS);
//		setDialogSettings(JavaPlugin.getDefault().getDialogSettings());
		
	}

	/*
	 * @see Wizard#createPages
	 */	
	public void addPages() 
	{
		super.addPages();
		try {
			IType type = TestCreationUtils.getCorrespondingTestType(element);
			if (type != null) {
				fPage = new AddTestCaseCreationWizardPage(traceTree, element, methodName, type);
				setWindowTitle("Add Test Case");
			}
		}
		catch (CoreException ce) {
			MCPlugin.log(ce);
		}
		if (fPage == null) {
			fPage = new NewTestCaseWizardPage(traceTree, element, methodName);
			setWindowTitle("New Test Case");
		}
		addPage(fPage);
		fPage.init(getSelection());
	}
	
	/*(non-Javadoc)
	 * @see org.eclipse.jdt.internal.ui.wizards.NewElementWizard#canRunForked()
	 */
	protected boolean canRunForked() 
	{
		return !fPage.isEnclosingTypeSelected();
	}

	/* (non-Javadoc)
	 * @see org.eclipse.jdt.internal.ui.wizards.NewElementWizard#finishPage(org.eclipse.core.runtime.IProgressMonitor)
	 */
	protected void finishPage(IProgressMonitor monitor)
		throws InterruptedException, CoreException 
	{
		fPage.createType(monitor); // use the full progress monitor
	}
		
	/* (non-Javadoc)
	 * @see org.eclipse.jface.wizard.IWizard#performFinish()
	 */
	public boolean performFinish() 
	{
		warnAboutTypeCommentDeprecation();
		boolean res= super.performFinish();
		if (res) {
			IResource resource= fPage.getModifiedResource();
			if (resource != null) {
				selectAndReveal(resource);
				openResource((IFile) resource);
			}	
		}
		return res;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.jdt.internal.ui.wizards.NewElementWizard#getCreatedElement()
	 */
	public IJavaElement getCreatedElement() 
	{
		return fPage.getCreatedType();
	}

}
