package de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IPackageFragmentRoot;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaConventions;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.internal.ui.dialogs.StatusInfo;
import org.eclipse.jdt.internal.ui.wizards.dialogfields.DialogField;
import org.eclipse.jdt.internal.ui.wizards.dialogfields.IDialogFieldListener;
import org.eclipse.jdt.internal.ui.wizards.dialogfields.LayoutUtil;
import org.eclipse.jdt.internal.ui.wizards.dialogfields.StringDialogField;
import org.eclipse.jdt.ui.wizards.NewTypeWizardPage;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.MethodArgument;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.MethodCall;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.TraceNode;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;

public class NewTestCaseWizardPage extends NewTypeWizardPage 
{
	
	private TraceNode[] traceTree;
	protected IJavaElement element;
	private String methodName;
	
	private StringDialogField fMethodNameDialogField;
	
	protected IStatus methodNameStatus = null;
	
	public NewTestCaseWizardPage(TraceNode[] traceTree, IJavaElement element,
			String methodName) 
	{
		super(true, "NewTestCaseWizardPage");
		this.traceTree = traceTree;
		this.element = element;
		this.methodName = methodName;
		
		fMethodNameDialogField = new StringDialogField();
		fMethodNameDialogField.setLabelText("Method Name:");
		fMethodNameDialogField.setDialogFieldListener(new MethodNameListener());
	}
	
	public void createControl(Composite parent) 
	{
		initializeDialogUnits(parent);
		
		Composite composite = new Composite(parent, SWT.NONE);
		composite.setFont(parent.getFont());
		
		int nColumns = 4;
		
		GridLayout layout = new GridLayout();
		layout.numColumns= nColumns;		
		composite.setLayout(layout);
		
		// pick & choose the wanted UI components
		
		createContainerControls(composite, nColumns);	
		createPackageControls(composite, nColumns);	
				
		createSeparator(composite, nColumns);
		
		createTypeNameControls(composite, nColumns);
		createModifierControls(composite, nColumns);
		
		createMethodNameControls(composite, nColumns);
			
		createSuperClassControls(composite, nColumns);
		
		setControl(composite);
			
		Dialog.applyDialogFont(composite);
	}
	
	protected void createMethodNameControls(Composite composite, int nColumns) 
	{
		
		fMethodNameDialogField.doFillIntoGrid(composite, nColumns - 1);
		DialogField.createEmptySpace(composite);
		LayoutUtil.setWidthHint(fMethodNameDialogField.getTextControl(null),
				getMaxFieldWidth());
	}
	
	public void init(IStructuredSelection selection)
	{
		initContainerPage(element);
		initTypePage(element);
		initTypeName(element);
		initMethodName(methodName);
		setSuperClass("junit.framework.TestCase", true);
		
		doStatusUpdate();
		
		IProject project = element.getJavaProject().getProject();
		
		try {
			String testDir = project.getPersistentProperty(TestCreationUtils.TEST_SOURCE_FOLDER);
			if (testDir != null) {
				IPackageFragmentRoot root = (IPackageFragmentRoot)JavaCore.create(testDir);
				setPackageFragmentRoot(root, true);
			}
		} 
		catch (CoreException e) {
			MCPlugin.log(e);
		}
	}
	
	protected void initTypeName(IJavaElement element)
	{
		ICompilationUnit unit = (ICompilationUnit) element.getAncestor(IJavaElement.COMPILATION_UNIT);
		IType type = null;
		if (unit != null) {
			type = unit.findPrimaryType();
		}
		if (type != null) { 
			String name = type.getElementName();
			setTypeName(name + TestCreationUtils.TEST_CLASS_NAME_POSTFIX, true);
		}
		else {
			MCPlugin.log("Could not init type name for element " + element);
		}
	}

	protected void initMethodName(String methodName)
	{
		fMethodNameDialogField.setText(methodName);
	}
	
	@Override
	public void setVisible(boolean visible)
	{
		super.setVisible(visible);
		if (visible) {
			setFocus();
		}
	}
	
	@Override
	protected void setFocus() 
	{
		fMethodNameDialogField.setFocus();
	}
	
	@Override
	protected void createTypeMembers(IType type, ImportsManager imports, 
			IProgressMonitor monitor) throws CoreException 
	{
		
		StringBuilder buf = new StringBuilder();
		
		final String lineDelim= "\n"; // OK, since content is formatted afterwards //$NON-NLS-1$
		
		
		buf.append("public void "); //$NON-NLS-1$
		buf.append(getNormalizedMethodName());
		buf.append("() {");
		buf.append(lineDelim);
	
		addTestMethodContent(buf, imports);
		
		buf.append("}"); //$NON-NLS-1$
		type.createMethod(buf.toString(), null, false, null);
			
		// store project properties
		 
		IPackageFragmentRoot root = getPackageFragmentRoot();
		TestCreationUtils.setTestSourceFolder(root);
		
		if (monitor != null) {
			monitor.done();
		}	
	}
	
	private String getNormalizedMethodName()
	{
		String methodName = getMethodName();
		if (methodName.startsWith(TestCreationUtils.TEST_METHOD_PREFIX)) {
			return methodName;
		}
		else {
			return TestCreationUtils.TEST_METHOD_PREFIX + methodName;
		}
	}

	private void addTestMethodContent(StringBuilder buf, ImportsManager imports) 
	{
		TraceNode current = traceTree[traceTree.length - 1];
		MethodCall call = current.getMethodCall();
		MethodArgument instance = call.getInstance();
		if (instance != null) {
			MethodCall constructor = instance.getConstructorCall();
			String type = getSourceType(instance.getFormalParameter().getType(), imports);
			buf.append(type).append(" instance = new ").append(type).append("(");
			if (constructor != null) addArguments(buf, constructor.getArguments());
			buf.append(");\n");
			buf.append("instance.").append(call.getMethod().getName()).append("(");
			addArguments(buf, call.getArguments());
			buf.append(");\n");
		}
		else {
			String type = getSourceType(
					call.getMethod().getPDSClass().getName(), imports);
			buf.append(type).append(".").append(call.getMethod().getName()).append("(");
			addArguments(buf, call.getArguments());
			buf.append(");\n");
		}
	}
	
	private void addArguments(StringBuilder buf, MethodArgument[] args)
	{
		for (int i = 0; i < args.length; i++) {
			MethodArgument arg = args[i];
			if (arg.isValue()) {
				buf.append(arg.getVariable().intValue());
			}
			else if (arg.isArray()) {
				int[] array = arg.getArray();
				if (array.length == 0) {
					buf.append("new int[0]");
				}
				else {
					buf.append("new int[] { ");
					for (int j = 0; j < array.length; j++) {
						buf.append(array[j]);
						if (j < array.length - 1) {
							buf.append(',');
						}
					}
					buf.append(" }");
				}
			}
			if (i < args.length  - 1) {
				buf.append(",");
			}
		}
	}
	
	protected String getSourceType(String type, ImportsManager imports)
	{
		type = type.replace('/', '.');
		return imports.addImport(type); 
	}
	
	protected IStatus methodNameChanged()
	{
		StatusInfo status = new StatusInfo();
		String methodName = getMethodName();
		if (methodName.length() == 0) {
			status.setError("Name of method must not be empty");
			return status;
		}
		IStatus val = JavaConventions.validateMethodName(methodName);
		if (val.getSeverity() == IStatus.ERROR) {
			status.setError(val.getMessage());
			return status;
		}
		if (methodExists(getNormalizedMethodName())) {
			status.setError("A method with the name " + getNormalizedMethodName() 
					+ " already exsists");
			return status;
		}
		if (!methodName.startsWith(TestCreationUtils.TEST_METHOD_PREFIX)) {
			status.setWarning("Junit test methods have to start with \"test\", it will be prepended");
		}
		
		return status;
	}
	
	protected void doStatusUpdate() 
	{
		updateStatus(methodNameStatus);
	}

	protected boolean methodExists(String methodName)
	{
		return false;
	}
	
	protected String getMethodName()
	{
		return fMethodNameDialogField.getText();
	}
	
	private class MethodNameListener implements IDialogFieldListener
	{

		public void dialogFieldChanged(DialogField field) 
		{
			methodNameStatus = methodNameChanged();
			doStatusUpdate();
		}
		
	}
	
}
