package de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui;

import org.apache.log4j.Logger;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.events.TypedEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Spinner;
import org.eclipse.ui.IMemento;

import de.uni_stuttgart.fmi.szs.jmoped.PDSDefault;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;

public class SWTUtils {

	private static int bits;
	private static int heapSize;
	/*private static boolean initMethodOnly = false;
	private static boolean contAfterFail = true;
	private static boolean checkExcep = true;*/
	
	private static Logger logger = Logger.getLogger(SWTUtils.class);
	
	public static Spinner createBitsSpinner(Composite parent, IMemento memento)
	{
		Spinner bitsSpinner = new Spinner(parent, SWT.BORDER);
		bitsSpinner.setMinimum(1);
		bitsSpinner.setMaximum(6);
		bitsSpinner.setIncrement(1);
		bitsSpinner.setPageIncrement(2);
		if (memento != null) {
			Integer b = memento.getInteger(MCPlugin.ATTR_BITS);
			bits = (b != null) ? b.intValue() : 3; 
		} else {
			bits = 3;
		}
		bitsSpinner.setSelection(bits);
		//bitsSpinner.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
		bitsSpinner.setLayoutData(new GridData(20, SWT.DEFAULT));
		
		return bitsSpinner;
	}
	
	public static Spinner createHeapSizeSpiner(Composite parent, IMemento memento)
	{
		Spinner heapSizeSpinner = new Spinner(parent, SWT.BORDER);
		heapSizeSpinner.setMinimum(2);
		heapSizeSpinner.setMaximum(63);
		heapSizeSpinner.setIncrement(1);
		heapSizeSpinner.setPageIncrement(2);
		if (memento != null) {
			Integer h = memento.getInteger(MCPlugin.ATTR_HEAP_SIZE);
			heapSize = (h != null) ? h.intValue() : 7;
		} else {
			heapSize = 7;
		}
		heapSizeSpinner.setSelection(heapSize);
		heapSizeSpinner.setLayoutData(new GridData(20, SWT.DEFAULT));
		
		return heapSizeSpinner;
	}
	
	public static void installSpinnerSyncListener(Spinner bitsSpinner, Spinner heapSizeSpinner)
	{
		new SpinnerSyncListener(bitsSpinner, heapSizeSpinner);
	}
	
	public static Button createCheckbox(Composite parent, IMemento memento,
			String label, String tag, boolean value) {
		
		Button checkbox = new Button(parent, SWT.CHECK);
		checkbox.setText(label);
		checkbox.setLayoutData(new GridData(SWT.LEFT, SWT.TOP, true, false, 2, 1));
		if (memento != null) {
			Integer i = memento.getInteger(tag);
			if (i != null) value = (i.intValue() == 1) ? true : false;
		}
		checkbox.setSelection(value);
		
		return checkbox;
	}
	
	/*public static void installCheckBoxSyncListener(Button init, Button cont,
			Button excep) {
		
		new CheckBoxSyncListener(init, cont, excep);
	}*/
	
	public static int getBits() {
		
		return bits;
	}
	
	public static void setBits(int bits) {
		
		SWTUtils.bits = bits;
	}
	
	public static int getHeapSize() {
		
		return heapSize;
	}
	
	/*public static boolean checkInitialMethodOnly() {
		
		return initMethodOnly;
	}
	
	public static boolean continueAfterAssertionError() {
		
		return contAfterFail;
	}
	
	public static boolean checkExceptions() {
		
		return checkExcep;
	}*/
	
	/*private static class CheckBoxSyncListener implements 
			SelectionListener, DisposeListener {
		
		private Button init;
		private Button cont;
		private Button excep;
		
		public CheckBoxSyncListener(Button init, Button cont, Button excep) {
			
			this.init = init;
			this.cont = cont;
			this.excep = excep;
			
			init.addSelectionListener(this);
			init.addDisposeListener(this);
			
			cont.addSelectionListener(this);
			cont.addDisposeListener(this);
			
			excep.addSelectionListener(this);
			excep.addDisposeListener(this);
		}
		
		private void selected(SelectionEvent e) {
			
			if (e.getSource() == init) {
				
				initMethodOnly = init.getSelection();
				logger.debug("intMethodOnly: " + initMethodOnly);
			} else if (e.getSource() == cont) {
				
				contAfterFail = cont.getSelection();
				logger.debug("contAfterFail: " + contAfterFail);
			} else {
				
				checkExcep = excep.getSelection();
				logger.debug("checkExcep: " + checkExcep);
			}
		}

		public void widgetSelected(SelectionEvent e) {
			
			selected(e);
		}

		public void widgetDefaultSelected(SelectionEvent e) {
			
			selected(e);
		}

		public void widgetDisposed(DisposeEvent e) {

			if (init != null) {
				
				init.removeSelectionListener(this);
				init.removeDisposeListener(this);
				init = null;
			}
			
			if (cont != null) {
				
				cont.removeSelectionListener(this);
				cont.removeDisposeListener(this);
				cont = null;
			}
			
			if (excep != null) {
				
				excep.removeSelectionListener(this);
				excep.removeDisposeListener(this);
				excep = null;
			}
		}
	}*/
		
	private static class SpinnerSyncListener implements ModifyListener, 
		SelectionListener, DisposeListener
	{

		private Spinner bitsSpinner;
		private Spinner heapSizeSpinner;
		
		public SpinnerSyncListener(Spinner bitsSpinner, Spinner heapSizeSpinner)
		{
			this.bitsSpinner = bitsSpinner;
			this.heapSizeSpinner = heapSizeSpinner;
			
			bitsSpinner.addModifyListener(this);
			bitsSpinner.addSelectionListener(this);
			bitsSpinner.addDisposeListener(this);
			
			heapSizeSpinner.addModifyListener(this);
			heapSizeSpinner.addSelectionListener(this);
			heapSizeSpinner.addDisposeListener(this);
		}
		
		private void modified(TypedEvent event) throws CoreException
		{
			if (event.getSource() == bitsSpinner) {
				bits = bitsSpinner.getSelection();
				logger.debug("bits: " + bits);
				
				int newHeapSize = PDSDefault.getMaxIntForBits(bits);
				if (newHeapSize < heapSizeSpinner.getSelection()) {
					heapSizeSpinner.setSelection(newHeapSize);
					heapSize = newHeapSize;
					logger.debug("heapSize: " + heapSize);
				}
			}
			else {
				heapSize = heapSizeSpinner.getSelection();
				logger.debug("heapSize: " + heapSize);
				
				int newBits = PDSDefault.getMinimumBitsForHeapSize(heapSize);
				if (newBits > bitsSpinner.getSelection()) {
					bitsSpinner.setSelection(newBits);
					bits = newBits;
					logger.debug("bits: " + bits);
				}
			}
		}
		
		public void modifyText(ModifyEvent e) 
		{
			try {
				modified(e);
			} catch (CoreException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
		}

		public void widgetSelected(SelectionEvent e) 
		{
			try {
				modified(e);
			} catch (CoreException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
		}

		public void widgetDefaultSelected(SelectionEvent e) 
		{
		}

		public void widgetDisposed(DisposeEvent e) 
		{
			if (bitsSpinner != null && heapSizeSpinner != null) {
				bitsSpinner.removeModifyListener(this);
				bitsSpinner.removeSelectionListener(this);
				bitsSpinner.removeDisposeListener(this);
				
				heapSizeSpinner.removeModifyListener(this);
				heapSizeSpinner.removeSelectionListener(this);
				heapSizeSpinner.removeDisposeListener(this);
				
				bitsSpinner = heapSizeSpinner = null;
			}
		}
		
	}
	
}
