package de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.QualifiedName;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IPackageFragmentRoot;
import org.eclipse.jdt.core.IType;

import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;

public class TestCreationUtils 
{

	public static final String TEST_METHOD_PREFIX = "test";
	
	public static final String TEST_CLASS_NAME_POSTFIX = "Test";
	
	public static final QualifiedName TEST_SOURCE_FOLDER = 
		new QualifiedName(MCPlugin.PLUGIN_ID, "TEST_SOURCE_FOLDER");

	/**
	 * Returns true if element is a compilation unit and a corresponding test
	 * class exsists in the project specific test directory.
	 * 
	 * @param element
	 * @return if project specific test directory is not set yet, if no test class exists
	 */
	public static boolean testTypeExists(IJavaElement element)
	{
		try {
			return getCorrespondingTestType(element) != null;
		}
		catch (CoreException ce) {
			MCPlugin.log(ce);
		}
		return false;
	}
	
	public static IType getCorrespondingTestType(IJavaElement element) throws CoreException
	{
		IJavaProject javaProject = element.getJavaProject();
		IPackageFragment pack = 
			(IPackageFragment) element.getAncestor(IJavaElement.PACKAGE_FRAGMENT);
		ICompilationUnit unit = 
			(ICompilationUnit) element.getAncestor(IJavaElement.COMPILATION_UNIT);
		IType type = null;
		if (unit != null) {
			type = unit.findPrimaryType();
		}
		if (pack != null && type != null) {
			String testTypeName = type.getElementName() + TEST_CLASS_NAME_POSTFIX;
			String fullyQualifiedName = (!pack.isDefaultPackage()) ? 
					pack.getElementName() + "." + testTypeName : testTypeName;
			return  javaProject.findType(fullyQualifiedName);
		}
		return null;
	}
	
	public static String getTestSourceFolder(IJavaElement element)
	{
		IProject project = element.getJavaProject().getProject();
		try {
			return project.getPersistentProperty(TEST_SOURCE_FOLDER);
		}
		catch (CoreException e) {
			MCPlugin.log(e);
		}
		return null;
	}
	
	public static void setTestSourceFolder(IPackageFragmentRoot root)
	{
		IProject project = root.getJavaProject().getProject();
		try {
			project.setPersistentProperty(TestCreationUtils.TEST_SOURCE_FOLDER, 
					root.getHandleIdentifier());
		}
		catch (CoreException e) {
			MCPlugin.log(e);
		}
	}
	
}
