package de.uni_stuttgart.fmi.szs.jmoped.eclipse.marker;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.internal.ui.text.correction.IJavaMarkerResolutionExtension;
import org.eclipse.swt.graphics.Image;
import org.gjt.jclasslib.structures.InvalidByteCodeException;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageResult;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.IntConfig;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.JavaAssertionExpressionResolver;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.JavaMethodArgumentsResolver;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.TraceGenerator;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.TraceNode;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Transition;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui.OpenTestCaseCreationWizardAction;

public class CreateTestCaseResolution implements IJavaMarkerResolutionExtension
{

	private Transition transition;
	
	private State state;
	
	private String memento;
	
	public CreateTestCaseResolution(Transition transition, State state, String memento)
	{
		this.transition = transition;
		this.state = state;
		this.memento = memento;
	}
	
	public String getLabel() 
	{
		return "Create Test Case for " + state;
	}
	
	public String getDescription() 
	{
		if (state == State.ASSERTION_FAILED) {
			CoverageResult res = MarkerManager.getInstance().getCoverageResult(memento);
			if (res != null) {
				try {
					String failedExpression = 
						JavaAssertionExpressionResolver.getFailedAssertionExpression
						(res, transition);
					if (failedExpression != null) {
						return "Creates test case for failed assertion: " + escapeHTML(failedExpression);
					}
				}
				catch (InvalidByteCodeException e) {
					MCPlugin.log(e);
				}
			}
		}
		return "Creates a test case that covers the path to the " + state;
	}

	public int getRelevance() 
	{
		return Integer.MAX_VALUE;
	}

	public Image getImage() 
	{
		return null;
	}
	
	public void run(IMarker marker) 
	{
		CoverageResult result = MarkerUtils.getResult(marker);
			
		if (result != null && transition != null) {
			IntConfig startConfig = TraceGenerator.getTraceConfig(result, transition);
			JavaMethodArgumentsResolver resolver = new JavaMethodArgumentsResolver();
			TraceNode[] roots =
				resolver.resolveMethodCallTraceTree(startConfig, result.getPDS());
			
			IResource resource = marker.getResource();
			IJavaElement elem = (IJavaElement)resource.getAdapter(IJavaElement.class);
			
			new OpenTestCaseCreationWizardAction(roots, elem, 
					"testFor" + state.toString().replaceAll(" ", "") + "In"  
					+ getMethodName(roots)).run();
		}
	}
	
	private String getMethodName(TraceNode[] roots)
	{
		TraceNode node = roots[roots.length - 1];
		String classAndMethod = 
			node.getMethodCall().getMethod().getHumanReadableName();
		return classAndMethod.substring(classAndMethod.lastIndexOf('.') + 1);
	}
	
	private static String escapeHTML(String value)
	{
		value = value.replaceAll("&","&amp;");
        value = value.replaceAll("<","&lt;");
        return value.replaceAll(">","&gt;");
	}
}
