package de.uni_stuttgart.fmi.szs.jmoped.eclipse.marker;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchConfigurationWorkingCopy;
import org.eclipse.jdt.internal.ui.text.correction.IJavaMarkerResolutionExtension;
import org.eclipse.swt.graphics.Image;

import de.uni_stuttgart.fmi.szs.jmoped.PDSDefault;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;

public class HeapOverflowMarkerResolution implements
		IJavaMarkerResolutionExtension
{

	private String label;
	private String desc;
	
	private int missing;
	
	public HeapOverflowMarkerResolution(int missing)
	{
		this.missing = missing;
		if (missing > 0) {
			label = "Increase heap by " + missing + " and run again";
			desc = "Increases heap by the number of missing fields (" + missing + ") "
			+ "and runs the analysis again";
		}
		else {
			label = "Run with increased heap size";
			desc = "Increases heap size without increasing the bit size if possible or otherwise " +
			"increases the bitsize by one."; 
		}
	}
	
	public String getDescription() 
	{
		return desc; 
	}

	public Image getImage() 
	{
		return null;
	}

	public String getLabel() 
	{
		return label;
	}

	public void run(IMarker marker) 
	{
		try {
			ILaunchConfiguration configuration = MarkerUtils.getLaunchConfiguration(marker);
			ILaunchConfigurationWorkingCopy wCopy = configuration.getWorkingCopy();
			int bits = configuration.getAttribute(MCPlugin.ATTR_BITS, PDSDefault.BITS);
			int heapSize = configuration.getAttribute(MCPlugin.ATTR_HEAP_SIZE,
					PDSDefault.HEAP_SIZE);
			
			if (missing > 0) {
				heapSize = heapSize + missing;
				bits = PDSDefault.getMinimumBitsForHeapSize(heapSize);
			}
			else {
				if (heapSize == PDSDefault.getMaxIntForBits(bits)) {
					bits++;
					heapSize = PDSDefault.getMaxIntForBits(bits);
				}
				else {
					heapSize = PDSDefault.getMaxIntForBits(bits);
				}
			}
			
			wCopy.setAttribute(MCPlugin.ATTR_BITS, bits);
			wCopy.setAttribute(MCPlugin.ATTR_HEAP_SIZE, heapSize);
			wCopy.doSave().launch("run", null);
		}
		catch (CoreException e) {
			MCPlugin.log(e);
		}
	}
	
	public int getRelevance() 
	{
		return Integer.MAX_VALUE;
	}


}
