package de.uni_stuttgart.fmi.szs.jmoped.eclipse.marker;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IMarker;
import org.eclipse.ui.IMarkerResolution;
import org.eclipse.ui.IMarkerResolution2;
import org.eclipse.ui.IMarkerResolutionGenerator2;
import org.gjt.jclasslib.structures.InvalidByteCodeException;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageResult;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.JavaHeapOverflowResolver;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Line;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Transition;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;


public class HeapOverflowMarkerResolutionGenerator 
	implements IMarkerResolutionGenerator2
{

	public IMarkerResolution[] getResolutions(IMarker marker) 
	{
		CoverageResult result = MarkerUtils.getResult(marker);
		Line line = MarkerUtils.getLine(marker, result);
		List<Transition> transitions = line.getTransitions(State.HEAP_OVERFLOW);
		int maxIncrement = getMaxHeapIncrement(transitions, result);
		
		return new IMarkerResolution2[] {
				new HeapOverflowMarkerResolution(maxIncrement)
		};
	}
	
	private int getMaxHeapIncrement(List<Transition> transitions, CoverageResult result)
	{
		int missing = 0;
		try {
			for (Transition transition : transitions) {
				missing = Math.max(missing,
					JavaHeapOverflowResolver.getMissingHeapFields(result, transition));
			}
		}
		catch (InvalidByteCodeException ive) {
			MCPlugin.log(ive);
		}
		return missing;
	}

	public boolean hasResolutions(IMarker marker) 
	{
		return MarkerUtils.getResult(marker) != null;
	}

}
