package de.uni_stuttgart.fmi.szs.jmoped.eclipse.marker;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRunnable;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchManager;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageResult;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Line;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;

public class MarkerUtils 
{
	
	public static void deleteAllCoverageMarkers(final IResource resource) throws CoreException
	{
		IWorkspaceRunnable runnable = new IWorkspaceRunnable() {
			public void run(IProgressMonitor monitor) throws CoreException 
			{
				for (State state : State.values()) {
					resource.getProject().deleteMarkers(MCPlugin.getMarkerId(state), false, 
							IResource.DEPTH_INFINITE);
				}
			}
		};
		resource.getWorkspace().run(runnable, null);
	}
	
	public static ILaunchConfiguration getLaunchConfiguration(IMarker marker) 
		throws CoreException
	{
		ILaunchManager manager = DebugPlugin.getDefault().getLaunchManager();
		return manager.getLaunchConfiguration(getLaunchConfigurationMemento(marker));
	}
	
	public static String getLaunchConfigurationMemento(IMarker marker)
		throws CoreException
	{
		return (String)marker.getAttribute(MCPlugin.ATTR_LAUNCH_CONFIGURATION);
	}
	
	public static CoverageMethod getMethod(IMarker marker, CoverageResult result)
	{
		String methodName = marker.getAttribute(MCPlugin.ATTR_METHOD_NAME, null);
		return getMethod(result.getMethods(), methodName);
	}
	
	/**
	 * Get Line at the marker.
	 * 
	 * @param marker
	 * @param result
	 * @return
	 */
	public static Line getLine(IMarker marker, CoverageResult result)
	{
		CoverageMethod method = getMethod(marker, result);
		// Remy: I add this check to prevent NPE
		if (method == null)
			return null;
		int lineOffset = marker.getAttribute(MCPlugin.ATTR_LINE_OFFSET, -1);
		return method.getLines()[lineOffset];
	}
	
	public static CoverageMethod getMethod(CoverageMethod[] methods, String methodName)
	{
		for (CoverageMethod method : methods) {
			if (method.getMethod().getFormattedName().equals(methodName)) {
				return method;
			}
		}
		return null;
	}
	
	public static CoverageResult getResult(IMarker marker)
	{
		String launchConfigMememento = 
			marker.getAttribute(MCPlugin.ATTR_LAUNCH_CONFIGURATION, null);
		return launchConfigMememento != null ? 
				MarkerManager.getInstance().getCoverageResult(launchConfigMememento)
				: null;
	}
	
}
