package de.uni_stuttgart.fmi.szs.jmoped.eclipse.marker;

import org.eclipse.core.resources.IMarker;
import org.eclipse.jdt.internal.ui.text.correction.IJavaMarkerResolutionExtension;
import org.eclipse.swt.graphics.Image;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageResult;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.IntConfig;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.JavaMethodArgumentsResolver;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.TraceGenerator;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.TraceNode;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Transition;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCProgressView;

public class PostTraceResolution implements IJavaMarkerResolutionExtension
{

	private Transition transition;
	
	private State state;
	
	public PostTraceResolution(Transition transition, State state)
	{
		this.transition = transition;
		this.state = state;
	}
	
	public String getLabel() 
	{
		return "Get Trace for State " + state;
	}
	
	public String getDescription() 
	{
		return "Computes the method call trace and input values for " + state;
	}

	public int getRelevance() 
	{
		return Integer.MAX_VALUE;
	}

	public Image getImage() 
	{
		return null;
	}
	
	public void run(IMarker marker) 
	{
		CoverageResult result = MarkerUtils.getResult(marker);
		
		if (result == null) {
			return;
		}
		
		IntConfig startConfig = TraceGenerator.getTraceConfig(result, transition);
		JavaMethodArgumentsResolver resolver =	new JavaMethodArgumentsResolver();
		TraceNode[] roots = 
			resolver.resolveMethodCallTraceTree(startConfig, result.getPDS());
		MCProgressView viewer =
			MCPlugin.getDefault().findProgressViewtInActivePage();
		viewer.setTraceTree(roots);
	}
}
