package de.uni_stuttgart.fmi.szs.jmoped.eclipse.marker;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IMarker;
import org.eclipse.ui.IMarkerResolution;
import org.eclipse.ui.IMarkerResolutionGenerator2;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageResult;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Line;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Transition;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;

public class PostTraceResolutionGenerator implements IMarkerResolutionGenerator2
{

	public IMarkerResolution[] getResolutions(IMarker marker) 
	{
		CoverageResult result = MarkerUtils.getResult(marker);
		if (!result.isPostTrace()) return new IMarkerResolution[0];
		
		String memento =
			marker.getAttribute(MCPlugin.ATTR_LAUNCH_CONFIGURATION, null);
		boolean createTest = !result.getPDS().getRootPDSMethod().isMainMethod(); 
		
		Line line = MarkerUtils.getLine(marker, result);
		ArrayList<IMarkerResolution> resolutions = new ArrayList<IMarkerResolution>();
		for (State state : State.values()) {
			// Remy: I add this check to prevent NPE
			if (line == null)
				continue;
			if (state.isException()) {
				List<Transition> list = line.getTransitions(state);
				for (Transition trans : list) {
					resolutions.add(new PostTraceResolution(trans, state));
					if (createTest) {
						resolutions.add(new CreateTestCaseResolution(trans, state, memento));
					}
				}
			}
			else if (state == State.COVERED) {				
				List<Transition> list = line.getTransitions(state);
				if (!list.isEmpty()) {
					resolutions.add(new PostTraceResolution(list.get(0), state));
					if (createTest) {
						resolutions.add(new CreateTestCaseResolution(list.get(0), state, memento));
					}
				}
			}
		}
		return resolutions.toArray(new IMarkerResolution[resolutions.size()]);
	}

	/**
	 * Return true if the previous analysis was post*, and the line at the 
	 * marker is not NOT_COVERED (e.g. COVERED, PARTIALLY_COVERED, ...).
	 */
	public boolean hasResolutions(IMarker marker) 
	{
		CoverageResult result = MarkerUtils.getResult(marker);
		if (result == null) return false;
		
		if (!result.isPostTrace()) return false;
		
		Line line = MarkerUtils.getLine(marker, result);
		if (line == null) return false;
		
		return line.getState() != State.NOT_COVERED;
	}
}
