package de.uni_stuttgart.fmi.szs.jmoped.eclipse.marker;

import java.text.MessageFormat;

import org.apache.log4j.Logger;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchConfigurationWorkingCopy;
import org.eclipse.jdt.internal.ui.text.correction.IJavaMarkerResolutionExtension;
import org.eclipse.swt.graphics.Image;

import de.uni_stuttgart.fmi.szs.jmoped.PDSDefault;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCProgressView;

public class PreStarMarkerResolution implements IJavaMarkerResolutionExtension
{
	
	private int bitsIncrement;
	private static Logger logger = Logger.getLogger(PreStarMarkerResolution.class);
	
	public PreStarMarkerResolution(int bitsIncrement)
	{
		this.bitsIncrement = bitsIncrement;
	}

	public String getDescription() 
	{
		if (bitsIncrement == 0) {
			return "Starts a backwards analysis for this line with same number of bits";
		}
		else {
			return bitsIncrement == 1 ? "Starts a backwards analysis for this line with 1 more bit"
					: MessageFormat.format("Starts a backwards analysis for this line with {0} more bits",
							new Object[] { bitsIncrement });
		}
	}

	public Image getImage() 
	{
		return MCPlugin.getImage("up.png").createImage();
	}

	public String getLabel() 
	{
		if (bitsIncrement == 0) {
			return "Run backwards analysis with same number of bits";
		}
		else {
			return bitsIncrement == 1 ? "Run backwards analysis with 1 more bit"
					: MessageFormat.format("Run backwards analysis with {0} more bits",
							new Object[] { bitsIncrement });
		}
	}

	public void run(IMarker marker) 
	{
		try {
			logger.debug("PreStarMarkerResolution.run");
			ILaunchConfiguration configuration = MarkerUtils.getLaunchConfiguration(marker);
			ILaunchConfigurationWorkingCopy wCopy = configuration.getWorkingCopy();
			wCopy.setAttribute(MCPlugin.ATTR_PRE_STAR_METHOD_NAME, 
					(String)marker.getAttribute(MCPlugin.ATTR_METHOD_NAME));
			wCopy.setAttribute(MCPlugin.ATTR_LINE_OFFSET, 
					marker.getAttribute(MCPlugin.ATTR_LINE_OFFSET, 0));
			
			int bits = configuration.getAttribute(MCPlugin.ATTR_BITS, PDSDefault.BITS);
			MCProgressView view = MCProgressView.getInstance();
			view.getConfigControl().setBits(bits + bitsIncrement);
			wCopy.setAttribute(MCPlugin.ATTR_BITS, bits + bitsIncrement);
			
			wCopy.setAttribute(MCPlugin.ATTR_LAUNCH_CONFIGURATION, 
					MarkerUtils.getLaunchConfigurationMemento(marker));
			
			wCopy.launch("run", null);
			
			/* Remy: We need to delete this attribute, otherwise  we'll always
			 * do pre* analysis after the button "Rerun Analysis" is clicked.
			 */
			wCopy.setAttribute(MCPlugin.ATTR_PRE_STAR_METHOD_NAME, (String) null);
			wCopy.doSave();
		}
		catch (CoreException e) {
			MCPlugin.log(e);
		}
	}

	public int getRelevance() 
	{
		return Integer.MAX_VALUE - bitsIncrement;
	}
}
