package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.gjt.jclasslib.bytecode.AbstractInstruction;
import org.gjt.jclasslib.bytecode.BranchInstruction;
import org.gjt.jclasslib.bytecode.Opcodes;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.PDSClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInst;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;
import de.uni_stuttgart.fmi.szs.jmoped.BytecodeUtils;

public class BranchCoverageTester extends PostStarReachabilityTester
{

	public BranchCoverageTester(Class clazz) throws Exception 
	{
		super(clazz);
	}
	
	public BranchCoverageTester(Class clazz, String methodName, String descriptor)
		throws Exception
	{
		super(clazz.getName(), new String[0], methodName, descriptor);
	}
	
	public void test()
	{
		test(new NullProgressMonitor());
	}
	
	public void test(ProgressMonitor monitor) 
	{
		String[] labels = getBranchLabels(getPDS());
		test(labels, monitor);
	}
	
	/**
	 * Returns an array of String containing all target labels of branch 
	 * instructions.
	 * 
	 * @param pds the pushdown system
	 * @return the array of String
	 */
	public static String[] getBranchLabels(PDS pds) {
		
		Set<String> labels = new HashSet<String>();
		
		for (PDSClass clazz : pds.getPDSClasses()) {
			
			for (PDSMethod method : clazz.getMethodList()) {
				
				AbstractInstruction inst = null;
				Iterator<AbstractInstruction> iter 
						= method.getInstructionList().iterator();
				while (iter.hasNext()) {
					
					inst = inst == null ? iter.next() : inst;
					if (BytecodeUtils.isComparison(inst.getOpcode())) {
						
						labels.add(BytecodeUtils.branchTarget(method.getFormattedName(), inst));
						if (iter.hasNext()) {
							inst = iter.next();
							labels.add(method.getFormattedName() + inst.getOffset());
						}
					}
					else if (inst.getOpcode() == Opcodes.OPCODE_GOTO) {
						
						labels.add(BytecodeUtils.branchTarget(method.getFormattedName(), inst));
						inst = null;
					}
					else {
						inst = null;
					}
				}
			}
		}
		return labels.toArray(new String[labels.size()]);
	}
}
