package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;

/**
 * Defines the requirements for class that want to listen to progress and changes
 * of a coverage analysis. 
 * <p>
 * They are notified, when the coverage starts, fails or ends and when a byte code 
 *  instruction of a method was covered.
 */
public interface CoverageListener 
{
	/**
	 * Called when the coverage was started.
	 *
	 * @param pds the push down system that is analysed.
	 */
	void coverageStarted(PDS pds);
	/**
	 * Called when 
	 * 
	 * @param method
	 * @param lineOffset
	 * @param pcOffset
	 * @param state
	 * @param transition
	 */
	void coveredMethodOffset(CoverageMethod method, int lineOffset, int pcOffset,
			State state, Transition transition);
	/**
	 * Called when the coverage ended successfully and there is a coverage result
	 * 
	 * @param result the result of the analysis which can be used to retrieve further
	 * details
	 */
	void coverageEnded(CoverageResult result);
	/**
	 * Called when the coverage failed due to an exception
	 * @param message an error message
	 * @param t the exception
	 */
	void coverageFailed(String message, Throwable t);
	/**
	 * Called when a label in the pushdown system was reached.
	 * <p>
	 * This is only called when 
	 * {@link Coverage#setEnabledNotifyofReachedLabels(boolean)} is set to true.
	 * @param label
	 */
	void labelReached(String label);
}
