package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.Arrays;

/**
 * Maps to the native struct <code>intVariable</code> but does not hold any
 * handles to the native data structures.
 * <p>
 * This class is instantiated in the JNI code.
 */
public class IntVariable
{
	
	private final String name;

	private final int type;
	
	private final int dimension;
	
	private final int value;
	
	private final int[] arrayValue;
	
	public enum Type {
		BOOLEAN,
		INTEGER,
		ARRAY;
		
		@Override
		public String toString() 
		{
			switch (ordinal()) {
			case 0:
				return "boolean";
			case 1:
				return "int";
			case 2:
				return "array";
			}
			return "";
		}
	}
	
	IntVariable(String name, int type, int dimension, int value)
	{
		this.name = name;
		this.type = type;
		this.dimension = dimension;
		this.value = value;
		this.arrayValue = new int[0];
	}
	
	IntVariable(String name, int type, int dimension, int[] arrayValue)
	{
		this.name = name;
		this.type = type;
		this.dimension = dimension;
		this.value = 0;
		this.arrayValue = arrayValue;
	}
	
	@Override
	public boolean equals(Object obj) 
	{
		if (obj instanceof IntVariable) {
			IntVariable other = (IntVariable)obj;
			return type == other.type && name.equals(other.name) 
			&& dimension == other.dimension && value == other.value 
			&& Arrays.equals(arrayValue, other.arrayValue);
		}
		return false;
	}
	
	public String getName()
	{
		return name;
	}
	
	public boolean booleanValue()
	{
		return value != 0;
	}
	
	public int intValue()
	{
		return value;
	}
	
	public int[] arrayValue()
	{
		return arrayValue;
	}
	
	public int getIntegerBits()
	{
		return type;
	}
	
	public Type getType() 
	{
		if (type == 0) {
			return Type.BOOLEAN;
		}
		else if (dimension == 0) {
			return Type.INTEGER;
		}
		else {
			return Type.ARRAY;
		}
	}
	
	@Override
	public String toString() 
	{
		Type type = getType();
		StringBuilder builder = new StringBuilder();
		builder.append(type).append(" ");
		builder.append(getName()).append(" = ");
		if (type == Type.BOOLEAN) {
			builder.append(booleanValue());
		}
		else if (type == Type.INTEGER) {
			builder.append(intValue());
		}
		else {
			builder.append(Arrays.toString(arrayValue()));
		}
		return builder.toString();
	}
}
