package de.uni_stuttgart.fmi.szs.jmoped.coverage;

/**
 * An argument of a method call. 
 */
public interface MethodArgument 
{
	/**
	 * Returns whether the argument is given by value. This holds for primitive
	 * types.
	 */
	boolean isValue();
	/**
	 * Returns the int variable. In case the method argument is a reference, the 
	 * variable is pointer into the heap.
	 */
	IntVariable getVariable();
	/**
	 * Returns true if the argument is an array.
	 * @return
	 */
	boolean isArray();
	/**
	 * Returns the resolved array.
	 * <p>
	 * Arrays are also objects in the heap but resolved.
	 * 
	 * @return null if the method argument is not an array
	 */
	int[] getArray();
	/**
	 * Returns true if the method argument denotes a reference.
	 * @return
	 */
	boolean isReference();
	/**
	 * If the method argument is a reference this method returns the constructor
	 * call that created the object this refernce points to. 
	 * 
	 * @return null if constructor call could not be resolved or the method argument
	 * is not a reference
	 */
	MethodCall getConstructorCall();
	/**
	 * The formal parameter of this method argument.
	 * @return
	 */
	Parameter getFormalParameter();
}
