package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.io.File;
import java.io.IOException;

import org.gjt.jclasslib.structures.InvalidByteCodeException;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;

public class PreStarReachabilityTester extends Coverage {

	/**
	 * Convenience construtor for {@link #CoverageTester(String, String[], String)
	 * CoverageTester(clazz.getName(), searchPaths, methodName)}.
	 */
	public PreStarReachabilityTester(Class clazz, String[] searchPaths)
			throws Exception {
		this(clazz.getName(), searchPaths);
	}

	/**
	 * Constructs a CoverageTester for class <code>className</code>.
	 * <p>
	 * 
	 * @param className should be the fully qualified name of the class as it
	 * is returned by {@link Class#getName()}.
	 * @param searchPaths array of names of directories and jar files that are
	 * searched for the class file of <code>className</code>
	 * @param methodName the method that should be tested for coverage. The format
	 * should match {@link PDSMethod#getFormattedName()}, can be <code>null</code>
	 * @throws Exception
	 */
	public PreStarReachabilityTester(String className, String[] searchPaths)
			throws Exception 
	{
		this(new PDS(className, searchPaths));
	}

	public PreStarReachabilityTester(PDS pds) 
	{
		super(pds);
	}

	public CoverageMethod[] test(String label) 
	{
		return test(new String[] { label });
	}

	public CoverageMethod[] test(String label, ProgressMonitor monitor)
	{
		return test(new String[] { label }, monitor);
	}
	
	public CoverageMethod[] test(String[] labels) 
	{
		return test(labels, new NullProgressMonitor());
	}

	public CoverageMethod[] test(String[] labels, ProgressMonitor monitor) 
	{
		clearMethodsMap();
		CoverageListener listenerSupport = getListenerSupport();

		listenerSupport.coverageStarted(getPDS());
		monitor.beginTask("PreStar Analysis", 10000);
		
		// Write PDS to a temp file.
		File tmpFile = null;
		try {
			tmpFile = CoverageUtils.writeRemopla(getPDS(), labels[0]);
		}
		catch (InvalidByteCodeException ibce) {
			listenerSupport.coverageFailed("Invalid bytecode", ibce);
			monitor.done();
			return null;
		}
		catch (IOException ie) {
			listenerSupport.coverageFailed("Error writing to temporary file "
					+ tmpFile, ie);
			monitor.done();
			return null;
		}
		
		// Create CoverageResult.
		CoverageResult result = null;
		synchronized (super.getClass()) {
			if (monitor.isCanceled()) {
				monitor.done();
				return null;
			}
			result = analyseReachability(tmpFile.getAbsolutePath(), labels, 
					new SubProgressMonitor(10000, monitor));
		}
		if (result == null) {
			listenerSupport.coverageFailed("Prestar Coverage was cancelled",
					new InterruptedException());
		}
		else {
			result.setMethods(getMethods(new String[0]));
			result.setPDS(getPDS());
			listenerSupport.coverageEnded(result);
		}
		monitor.done();
		return getMethods(new String[0]);
	}
	
	private void unreachedLabels(String[] labels)
	{
		getMethods(labels == null ? new String[0] : labels);
	}

	private native CoverageResult analyseReachability(String fileName, String[] labels,
			ProgressMonitor monitor);

}
