package de.uni_stuttgart.fmi.szs.jmoped.coverage;

/**
 * This interface is closely modelled after the eclipse ProgressMonitor interface to ease the
 * implementation of both interfaces by just one class.  
 */
public interface ProgressMonitor 
{
	/**
	 * Must only be called once, when the task is started.
	 * @param description
	 * @param workAmount the expected amount that is accumulated through calls
	 * to {@link #worked(int)}.
	 */
	void beginTask(String description, int workAmount);
	/**
	 * Called when the task is done.
	 */
	void done();
	/**
	 * The task can query the caller if he wants to cancel the task. Long tasks should
	 * query this frequently to ensure liveliness.
	 */
	boolean isCanceled();
	/**
	 * After set to  true, should not be reset.
	 */
	void setCanceled(boolean cancel);
	/**
	 * Subtasks can notify the progress monitor of what they are doing.
	 * @param description
	 */
	void subTask(String description);
	/**
	 * The amount of work that was just finished. Should accumulate to the 
	 * <code>workAmount</code> given in {@link #beginTask(String, int)}.
	 */
	void worked(int amount);
}
