package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.Arrays;

import org.gjt.jclasslib.structures.InvalidByteCodeException;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;

/**
 * Implements the {@link de.uni_stuttgart.fmi.szs.jmoped.coverage.MethodArgument}
 * interface for references.
 */
class ReferenceMethodArgument implements MethodArgument
{
	private Parameter param;
	
	private IntVariable var;
	
	private IntConfig config; 
	
	private PDS pds;
	
	private MethodCall constructorCall;
	
	public ReferenceMethodArgument(Parameter param, IntVariable var, 
			IntConfig config, PDS pds)
	{
		this.param = param;
		this.var = var;
		this.config = config;
		this.pds = pds;
	}

	public boolean isValue() 
	{
		return false;
	}

	/**
	 * Returns the variable that points to the heap where the object is.
	 */
	public IntVariable getVariable() 
	{
		return var;
	}

	public boolean isArray() 
	{
		return false;
	}

	public int[] getArray() 
	{
		return null;
	}

	public boolean isReference() 
	{
		return true;
	}

	public Parameter getFormalParameter() 
	{
		return param;
	}
	
	@Override
	public String toString() 
	{
		StringBuilder builder = new StringBuilder("Heap: ");
		IntVariable[] globalVars = config.getGlobalVariables();
		IntVariable heap = JavaMethodArgumentsResolver.findVariable(globalVars,
				pds.getInfo().getHeapName());
		builder.append(Arrays.toString(heap.arrayValue()));
		builder.append(", Index: ");
		builder.append(var.intValue());
		builder.append(" : ");
		builder.append(param);
		return builder.toString();
	}
	
	public MethodCall getConstructorCall()
	{
		if (constructorCall == null) {
			try {
			constructorCall = resolveConstructorCall();
			} catch (InvalidByteCodeException e) {
				e.printStackTrace();
			}
		}
		return constructorCall;
	}

	private MethodCall resolveConstructorCall() throws InvalidByteCodeException
	{
		String className = param.getType();
		String constructorPrefix = 
			PDSMethod.formatMethodName(new String[] { className, "<init>", "" });
		int heapIndex = var.intValue();
		int offset = 1 + pds.getInfo().getClassSize(className);
		String heapPtrName = pds.getInfo().getHeapPtrName();
		IntConfig previous = config.getPreviousConfig();
		PDSMethod constructor = null;
		
		// go backwards through configs and find matching constructor all
		while (previous != null) {
			if (previous.getStackSymbol().startsWith(constructorPrefix)) {
				constructor = pds.getPDSMethod(previous.getStackSymbol());
				if (constructor != null) {
					IntVariable[] globalVars = previous.getGlobalVariables();
					IntVariable heapPtr = 
						JavaMethodArgumentsResolver.findVariable(globalVars, heapPtrName);
					if (heapPtr != null && heapPtr.intValue() - offset == heapIndex) {
						break;
					}
				}
			}
			previous = previous.getPreviousConfig();
		}
		
		if (previous != null) {
			JavaMethodArgumentsResolver resolver = new JavaMethodArgumentsResolver();
			return resolver.resolveMethodCalls(previous, constructor, pds)[0];
		}
		return null;
	}
	
	private int getObjectIndex() throws InvalidByteCodeException
	{
		return var.intValue() - pds.getInfo().getClassSize(param.getType()) - 1;
	}
	
	
}