package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;


public class TraceMinimizer 
{

	private CoverageResult result;
	private CoverageMethod method;
	
	private List<Transition> transitions = new ArrayList<Transition>();
	private Set<Integer> coveredOffsets = new HashSet<Integer>();
	
	private AnnotatedInstruction[] instructions;
	
	private int index;
	
	public TraceMinimizer(CoverageResult result, CoverageMethod method)
	{
		this.result = result;
		this.method = method;
		if (result == null || method == null) {
			throw new NullPointerException();
		}
		instructions = method.getAnnotatedInstructions().toArray
			(new AnnotatedInstruction[0]);
		index = instructions.length - 1;
		
		for (Transition t = findNextUnmarkedTransition(); 
			 t != null && !isCovered(); t = findNextUnmarkedTransition()) {
			markTransition(t);
		}
	}
	
	public Transition[] getTransitions()
	{
		return transitions.toArray(new Transition[transitions.size()]);
	}
	
	public boolean isCovered()
	{
		return coveredOffsets.size() == instructions.length;
	}
	
	public int getNumerOfCoveredInstructions()
	{
		return coveredOffsets.size();
	}
	
	private void markTransition(Transition transition)
	{
		IntConfig config
			= JavaAssertionExpressionResolver.getLastConfig(result, transition);
		String prefix = method.getMethod().getFormattedName();
		for (; config != null && config.getStackSymbol().startsWith(prefix); 
			 config = config.getPreviousConfig()) {
			int offset = CoverageUtils.extractOffset(config.getStackSymbol());
			coveredOffsets.add(offset);
		}
		transitions.add(transition);
	}
	
	private Transition findNextUnmarkedTransition()
	{
		for (; index >= 0; index--) {
			AnnotatedInstruction inst = instructions[index];
			if (coveredOffsets.contains(inst.getOffset())) {
				continue;
			}
			List<Transition> transitions = inst.getTransitions(State.COVERED);
			if (!transitions.isEmpty()) {
				return transitions.get(0);
			}
		}
		return null;
	}
	
	
	
}
