package de.uni_stuttgart.fmi.szs.jmoped;

import java.io.IOException;
import java.util.Arrays;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageTester;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.NullProgressMonitor;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageTesterTest.QuickSort;

public class OutputGeneratorTest extends PDSTestCase 
{
	
	public void testOutput() throws Exception
	{
		PDS pds = getPDS(FinallyClass.class);
		pds.translate();
		System.out.println(pds.toRemopla());
	}
	
	public static class FinallyClass 
	{
		public static void main(String[] args) throws IOException 
		{
			int i;
			try {
				main(args);
			}
			catch (RuntimeException re) {
				i = 2 * args.length;
			}
			finally {
				i = 2;
			}
			test(5);
			throw new IOException();
		}
		
		public static void test(int i)
		{
			assert(i != 5);
			assert(i != i);
			assert(i > i);
			int k = i * 2;
			assert(k < i * 2);
		}
		
		public static void blah()
		{
			int a = 0;
			int b = a;
			int c = b;
			if (a == 1 || b == 2 || c == 3) {
				a = 1;
			}
			else {
				a = 2;
			}
		}
		
	}
	
	public static class TestOutput
	{
		
		private int i;
		private int j = 1;
		
		public TestOutput(int i, int j)
		{
			this.i = i;
			this.j = j;
		}
		
		public TestOutput()
		{
			int[][] array = new int[1][2];
			int[] t = new int[2];
			int i = array[0][1];	
			int[] part = array[0];
		}
		
		/*public static void main(String[] args)
		{
			TestOutput out = new TestOutput();
			int[] array = new int[3];
			for (int i = 0; i < array.length; i++) {
				array[i] = 3;
			}
			out.test(array);
//			test(new int[0], 5, 3, 3, 4);
		}
	*/

		public static void main(String[] args)
		{
			int[] onedim = new int[13];
			int array = 1;
			// set size for 1 dim
			onedim[array] = 2;
			int offset = 0;
			int newoffset = 0;
			int size = 1;
			int[] dimensions = new int[] {1,  2, 3};
			
			for (int dim = 1; dim < dimensions.length && dimensions[dim] != 0; dim++) {
				int s1 = dimensions[dim - 1];
				int s2 = dimensions[dim];
				offset = newoffset; 
				newoffset += (s1 + 1);
				size *= s1;
				for (int i = 0, j = 0; i < size; i++, j++) {
					// set sizes for current
					onedim[array + newoffset + i * (s2 + 1)] = s2;
					// set pointer in first dim to second dim array
					if (j != 0 && j % (s1 + 1) == 0) {
						j++;
					}
					onedim[1 + array + offset + j] =  array + newoffset + i * (s2 + 1);
				}
			}
			System.out.println(Arrays.toString(onedim));
		}
	}
	
	public void testQuickSortPerformance() throws Exception
	{
		QuickSort sort = new QuickSort();
		for (int i = 0; i < 5; i++) {
			int[] array = new int[i];
			int total = (int)Math.pow(8, i);
			sort.sort(array);
			for (int j = 1; j < total; j++) {
				increment(array, array.length - 1, 7);
//				System.out.println(Arrays.toString(array));
				int[] copy = new int[i];
				System.arraycopy(array, 0, copy, 0, i);
				sort.sort(copy);
			}
		}
	}
	
	public void testQuickSortCoverage() throws Exception
	{
		CoverageTester tester = getTester(QuickSort.class, "sort", "([I)V");
		tester.test(new ConstructionMonitor());
	}
	
	public void testReversePerformance()
	{
		for (int i = 0; i < 5; i++) {
			int[] array = new int[i];
			int total = (int)Math.pow(5, i);
			System.out.println("total " + total);
			verifyReversal(array);
			for (int j = 1; j < total; j++) {
				increment(array, array.length - 1, 4);
//				System.out.println(Arrays.toString(array));
				verifyReversal(array);
				verifyReversal(array);
			}
		}
	}
	
	private void increment(int[] array, int index, int base)
	{
		int newValue = array[index] + 1;
		if (newValue <= base) {
			array[index] = newValue;
		}
		else {
			array[index] = 0;
			increment(array, --index, base);
		}
	}
	
	public static void verifyReversal(int[] array)
	{
		int[] copy = new int[array.length];
		for (int i = 0; i < copy.length; i++) {
			copy[i] = array[i];
		}
		reverse(array);
		for (int i = 0; i < copy.length; i++) {
			assert(copy[i] == array[copy.length - 1 - i]);
		}
	}
	
	public static void reverse(int[] array) {
        if (array == null) {
            return;
        }
        int i = 0;
        int j = array.length - 1;
        int tmp;
        while (j > i) {
            tmp = array[j];
            array[j] = array[i];
            array[i] = tmp;
            j--;
            i++;
        }
    }
	
	private static class ConstructionMonitor extends NullProgressMonitor
	{
		@Override
		public void beginTask(String description, int workAmount) {
			System.out.println(description);
		}
		
		@Override
		public void subTask(String description) {
			System.out.println(description);
		}
	}
	
}
