/*
    Copyright (C) 2005 Felix Berger <bflat1@gmx.net>

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package de.uni_stuttgart.fmi.szs.jmoped;


public class PDSClassTest extends PDSTestCase 
{
	
	public void testGetMethodMain() throws Exception
	{
		PDSClass clazz = getPDSClass(HasMainMethodClass.class);
		assertNotNull(clazz.getMethodMain());
		
		clazz = getPDSClass(HasNoMethodClass.class);
		assertNull(clazz.getMethodMain());
		
		clazz = getPDSClass(HasMethodsSimilarToMainClass.class);
		assertNull(clazz.getMethodMain());
	}
	
	public static class HasMainMethodClass
	{
		public static void main(String[] args)
		{
			
		}
	}
	
	public static class HasNoMethodClass
	{
		
	}
	
	public static class HasMethodsSimilarToMainClass
	{
		public static void main()
		{
			
		}
		
		public static int main(String[] args)
		{
			return 0;
		}
		
		public void main(String args)
		{
			
		}
	}
	
	public void testgetMethodClinit() throws Exception
	{
		PDSClass clazz = getPDSClass(NoStaticInitializerClass.class);

		if (!isRunUnderEmma()) {
			assertNull(clazz.getMethodClinit());
		}
		else {
			// emma adds a static initializer
			assertNotNull(clazz.getMethodClinit());
		}
		
		clazz = getPDSClass(StaticInitializerClass.class);
		assertNotNull(clazz.getMethodClinit());
	}
	
	public static class NoStaticInitializerClass
	{
		static int i;
	}
	
	public static class StaticInitializerClass
	{
		static int i = 1;
	}
	
	public void testSize() throws Exception
	{
		PDSClass clazz = getPDSClass(NoInstanceFieldsClass.class);
		assertEquals(0, clazz.size());
		
		clazz = getPDSClass(OnlyStaticFieldsClass.class);
		assertEquals(0, clazz.size());
		
		clazz = getPDSClass(InstanceFieldsClass.class);
		assertEquals(2, clazz.size());
	}
	
	public static class NoInstanceFieldsClass
	{
		
	}
	
	public static class OnlyStaticFieldsClass
	{
		public static int FIELD1;
		private static Object FIELD2;
	}
	
	public static class InstanceFieldsClass
	{
		private int i1;
		public Object o1;
	}
	
	public static class ParameterCountClass
	{
		public ParameterCountClass()
		{
		}
		
		public ParameterCountClass(String i1, int i2, int i3)
		{
			
		}
		
		public void test()
		{
			
		}
		
		public int test1()
		{
			return 0;
		}
		
		protected void test2(int i1, String i2, Object o, Class c)
		{
			
		}
		
		private Object test3(int k)
		{
			return null;
		}
	}
}
