/*
    Copyright (C) 2005 Felix Berger <bflat1@gmx.net>

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package de.uni_stuttgart.fmi.szs.jmoped;

import java.util.Iterator;

import de.uni_stuttgart.fmi.szs.jmoped.annotation.PDSFieldBits;

public class PDSFieldTest extends PDSTestCase 
{
	
	public void testIsStatic() throws Exception
	{
		PDSClass clazz = getPDSClass(StaticFieldClass.class);
		
		for (Iterator i = clazz.getStaticFieldList().iterator(); i.hasNext();) {
			PDSField field = (PDSField) i.next();
			assertTrue(PDSField.isStatic(field.getFieldInfo().getAccessFlags()));
		}
		
		for (Iterator i = clazz.getNonStaticFieldList().iterator(); i.hasNext();) {
			PDSField field = (PDSField) i.next();
			assertFalse(PDSField.isStatic(field.getFieldInfo().getAccessFlags()));
		}
	}
	
	public static class StaticFieldClass
	{
		public final static int STATIC_FIELD = 1;
		
		public static StaticFieldClass INSTANCE;
		
		private static int PRIVATE_STATIC_FIELD;
		
		private static Object PRIVATE_STATIC_OBJECT_FIELD;
		
		private final int instanceField = 0;
		
		public int instanceField2;
		
		public StaticFieldClass neighbour;
		
		private Object obj;
	}
	
	public void testIsBoolean() throws Exception 
	{
		PDSClass clazz = getPDSClass(MixedBooleanFieldClass.class);
		Iterator i = clazz.getStaticFieldList().iterator();
		assertTrue(((PDSField)i.next()).isBoolean());
		assertFalse(((PDSField)i.next()).isBoolean());
		
		i = clazz.getNonStaticFieldList().iterator();
		assertFalse(((PDSField)i.next()).isBoolean());
		assertTrue(((PDSField)i.next()).isBoolean());
	}
	
	public static class MixedBooleanFieldClass
	{
		public static boolean STATIC_BOOLEAN;
		
		private static Object STATIC_OBJECT;
		
		private Object instanceObject;
		
		private boolean instanceBoolean;
	}
	
	public void testIdempotence()
	{
		String formatted = PDSField.formatFieldName("][bla.s?//\\dlk$$fjf..");
		assertEquals(formatted, PDSField.formatFieldName(formatted));
	}
	
	public void testIsReference() throws Exception
	{
		PDSClass clazz = getPDSClass(OnlyReferenceFieldsClass.class);
		
		for (PDSField field : clazz.getNonStaticFieldList()) {
			assertTrue(PDSField.isReference(field.getFieldInfo().getDescriptor()));
		}
		
		clazz = getPDSClass(OnlyPrimitiveFieldsClass.class);
		
		for (PDSField field : clazz.getNonStaticFieldList()) {
			assertFalse(PDSField.isReference(field.getFieldInfo().getDescriptor()));
		}
	}
	
	public static class OnlyReferenceFieldsClass
	{
		private int[] array;
		
		public Object o;
		
		Comparable comp;
		
		private String s;
	}
	
	public static class OnlyPrimitiveFieldsClass
	{
		private int i;
		
		long l;
		
		boolean z;
	}
	
	public void testToRemopla() throws Exception
	{
		PDSClass clazz = getPDSClass(FieldAnnotationsClass.class);
		for (PDSField field : clazz.getStaticFieldList()) {
			if (field.getName().equals("field1")) {
				assertEquals("int de_uni_stuttgart_fmi_szs_jmoped_PDSFieldTest_FieldAnnotationsClass_field1(2);\n",
						field.toRemopla());
			}
			else if (field.getName().equals("field2")) {
				assertEquals("int de_uni_stuttgart_fmi_szs_jmoped_PDSFieldTest_FieldAnnotationsClass_field2(1);\n",
						field.toRemopla());
			}
			else if (field.getName().equals("field3")) {
				assertEquals("int de_uni_stuttgart_fmi_szs_jmoped_PDSFieldTest_FieldAnnotationsClass_field3;\n",
						field.toRemopla());
			}
			else if (field.getName().equals("field4")) {
				assertEquals("int de_uni_stuttgart_fmi_szs_jmoped_PDSFieldTest_FieldAnnotationsClass_field4;\n",
						field.toRemopla());
			}
		}
	}
	
	public static class FieldAnnotationsClass
	{
		@PDSFieldBits(2)
		public static int field1;
		
		public static boolean field2;
		
		public static int field3;
		
		@PDSFieldBits(0)
		public static int field4;
	}
}
