/*
    Copyright (C) 2005 Felix Berger <bflat1@gmx.net>

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package de.uni_stuttgart.fmi.szs.jmoped;

import java.awt.event.ComponentAdapter;
import java.util.ArrayList;
import java.util.List;

import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.InterfaceImplementerClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.SubSubTestClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.SubTestClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.TestClass;
import de.uni_stuttgart.fmi.szs.jmoped.data.Bresenham;

/**
 * See also {@link de.uni_stuttgart.fmi.szs.jmoped.coverage.PDSInstCoverageTest}
 * for test that need the coverage package.
 */
public class PDSInstTest extends PDSTestCase 
{

	@SuppressWarnings("unchecked")
	public void testCommaList()
	{
		List list = new ArrayList();
		list.add("1");
		list.add("2");
		list.add("3");
		
		assertEquals("1, 2, 3", PDSStmt.commaList(list));
//		assertEquals("1, 2, 3", PDSStmt.commaList(list, list.size()));
//		assertEquals("1", PDSStmt.commaList(list, 1));
//		
//		try {
//			PDSStmt.commaList(list, list.size() + 1);
//			fail("IndexOutOfBoundsException expected");
//		}
//		catch (IndexOutOfBoundsException iobe) {
//		}
	}
	
	public void testIastore() throws Exception 
	{
		PDS pds = getPDS(ArrayExceptionsClass.class);
		pds.getInfo().setCheckForNullPointerExceptions(true);
// 		System.out.println(pds.toRemopla());
	}
	
	public static class ArrayExceptionsClass
	{
		public static void main(String[] args)
		{
			int[] nullArray = null;
			int[] emptyArray = new int[0];
			nullArray[0] = 1;
			emptyArray[0] = 1;
		}
	}
	
	/**
	 * Tests if goto Exception code is generated.
	 */
	public void testGeneralException() throws Exception
	{
		PDS pds = getPDS(ThrowsExceptionClass.class);
		pds.translate();
		String remopla = pds.toRemopla();
		
		int count = 0;
		int index = remopla.indexOf("goto " + pds.getInfo().getLabelGeneralException());
		System.out.println(remopla);
		while (index != -1) {
			++count;
			index = remopla.indexOf("goto " + pds.getInfo().getLabelGeneralException(), 
					index + 1);
		}
		// 1 for loop and 1 for throw
		assertEquals(2, count);
	}
	
	public static class ThrowsExceptionClass
	{
		public static void main(String[] args)
		{
			// assert also creates code that throws an exception but is handled differently
			assert(args.length != 0);
			throw new RuntimeException("exception");
		}
	}
	
	/**
	 * Test to document null pointer exception in {@link PDSInst#getfieldInst()}
	 * when run on 
	 * {@link de.uni_stuttgart.fmi.szs.jmoped.data.Bresenham#drawLine(int, int, int, int)}.
	 * 
	 * Fix: Added Applet, Graphics and Canvas to excluded classes in PDSDefault.
	 */
	public void testGetFieldNullPointerException() throws Exception
	{
		PDS pds = new PDS(Bresenham.class, "drawLine", "(IIII)V");
		assertNotNull(pds);
	}
	
	// TODO make this a real test
	public void testInvokevirtualInst() throws Exception
	{
		PDS pds = getPDS(InvokeVirtualInstClass.class);
//		System.out.println(pds.toRemopla());
	}
	
	public static class InvokeVirtualInstClass
	{
		public static void main(String[] args) 
		{
			TestClass instance = new SubSubTestClass();
			instance.me();
			instance.onlyInTestClass();
		}
	}
}
