/*
    Copyright (C) 2005 Felix Berger <bflat1@gmx.net>

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package de.uni_stuttgart.fmi.szs.jmoped;

import java.util.Iterator;

import org.gjt.jclasslib.bytecode.AbstractInstruction;
import org.gjt.jclasslib.structures.ClassFile;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethodTest.AssertionLinesCoveredClass;


public class PDSMethodTest extends PDSTestCase 
{
	
	public void testIsArray()
	{
		assertFalse(PDSMethod.isArray("()V", 0));
		assertFalse(PDSMethod.isArray("()V", 1));
		assertFalse(PDSMethod.isArray("()V", 2));
		
		assertFalse(PDSMethod.isArray("(I)V", 0));
		assertFalse(PDSMethod.isArray("(II)V", 1));
		assertFalse(PDSMethod.isArray("(Ljava/lang/Object;I)V", 2));
		
		assertTrue(PDSMethod.isArray("([I)I", 0));
		assertFalse(PDSMethod.isArray("([II)I", 1));
		assertTrue(PDSMethod.isArray("([I[I)I", 1));
		assertTrue(PDSMethod.isArray("([[I[I)I", 1));
		assertTrue(PDSMethod.isArray("([Ljava/lang/Object;[I)V", 0));
		assertTrue(PDSMethod.isArray("([Ljava/lang/Object;[I)V", 1));
		assertTrue(PDSMethod.isArray("([[Ljava/lang/Object;[I)V", 0));
		assertTrue(PDSMethod.isArray("([[Ljava/lang/Object;[[I)V", 1));
	}
	
	public void testIsStatic() throws Exception
	{
		PDSClass clazz = getPDSClass(NoStaticMethodsClass.class);
		
		if (!isRunUnderEmma()) {
			for (Iterator i = clazz.getMethodList().iterator(); i.hasNext();) {
				assertFalse(((PDSMethod)i.next()).isStatic());
			}
		}
		else {
			// emma adds two static methods
		}
		
		clazz = getPDSClass(StaticMethodsClass.class);
		Iterator i = clazz.getMethodList().iterator();
		assertTrue(i.hasNext());
		// skip constructor
		i.next();
		while (i.hasNext()) {
			assertTrue(((PDSMethod)i.next()).isStatic());
		}
	}
	
	public void testReturnsReference()
	{
		assertFalse(PDSMethod.returnsReference("()V"));
		assertFalse(PDSMethod.returnsReference("([I)I"));
	
		assertTrue(PDSMethod.returnsReference("()Ljava/lang/String;"));
		assertTrue(PDSMethod.returnsReference("(I[I)Ljava/lang/Integer;"));
		assertTrue(PDSMethod.returnsReference("(I[I)[Ljava/lang/Integer;"));
		assertTrue(PDSMethod.returnsReference("(I[I)[I"));
	}
	
	public static class NoStaticMethodsClass
	{
		public void me()
		{
			
		}
		
		private int test(int i)
		{
			return 0;
		}
	}
	
	public static class StaticMethodsClass
	{
		public static void me()
		{
			
		}
		
		private static Object test(int i)
		{
			return null;
		}
	}
	

	public void testGetHumanReadableName() throws Exception
	{
		PDSClass clazz = getPDSClass(AssertionLinesCoveredClass.class);
		PDSMethod method = clazz.getMethodMain();
		assertEquals("CoverageMethodTest.AssertionLinesCoveredClass.main", 
				method.getHumanReadableName());
		
		clazz = getPDSClass(Number.class);
		for (PDSMethod pdsMethod : clazz.getMethodList()) {
			// ignore native methods
			if (pdsMethod.getInstructionList().size() > 0) {
				if (pdsMethod.getName().equals("<init>")) {
					assertEquals("Number.Number", pdsMethod.getHumanReadableName());
				}
				else {
					assertEquals("Number." + pdsMethod.getName(),
							pdsMethod.getHumanReadableName());
				}
			}
		}
	}
	
	public void testIsAssertionErrorConstructor() throws Exception
	{
		PDSClass clazz = getPDSClass(AssertionErrorConstructorClass.class);
		PDSMethod mainMethod = clazz.getMethodMain();
		assertNotNull(mainMethod);
		ClassFile classFile = mainMethod.getMethodInfo().getClassFile();
		
		int assertionConstructorCount = 0;
		for (AbstractInstruction inst : mainMethod.getInstructionList()) {
			if (BytecodeUtils.isAssertionErrorConstructor(classFile, inst)) {
				++assertionConstructorCount;
			}
		}
		assertEquals(2, assertionConstructorCount);
	}
	
	public static class AssertionErrorConstructorClass
	{
		public static void main(String[] args)
		{
			assert(args.length == 0);
			throw new AssertionError();
		}
	}
	
}
