package de.uni_stuttgart.fmi.szs.jmoped;

import de.uni_stuttgart.fmi.szs.jmoped.PDSInfo.HeapOption;
import de.uni_stuttgart.fmi.szs.jmoped.annotation.PDSBits;
import de.uni_stuttgart.fmi.szs.jmoped.annotation.PDSFieldBits;

public class PDSMethodWrapperTest extends PDSTestCase {
	
	private PDSMethodWrapper wrap(Class c, String methodName, String methodDesc,
			HeapOption heapOption) throws Exception {
		
		PDS pds = new PDS(c, methodName, methodDesc);
		pds.setHeapOption(heapOption);
		PDSClass initClass = pds.getRootPDSClass();
		PDSMethod initMethod = initClass.getMethod(methodName, methodDesc);
		PDSMethodWrapper wrapper = new PDSMethodWrapper(initClass, initMethod);
		pds.getInfo().setup(pds.getPDSClasses());
		wrapper.wrap();
		
		return wrapper;
	}
	
	public void testStaticMethod() throws Exception {
		
		PDSMethodWrapper wrapper = wrap(A.class, "staticMethod", "()V", 
				HeapOption.SIMPLE);
		String expected = 
			"module void wrapperFor_de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_staticMethod_V()\n" +
			"{\n" +
			"de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_staticMethod_V();\n" +
			"}\n";
		assertEquals(expected, wrapper.toRemopla());
	}
	
	public void testStaticIntMethod() throws Exception {
		
		PDSMethodWrapper wrapper = wrap(A.class, "staticMethod", "(I)V", 
				HeapOption.SIMPLE);
		String expected = 
			"module void wrapperFor_de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_staticMethod_I_V()\n" +
			"{\n" +
			"int a(3);\n" +
			"a=undef;\n" +
			"de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_staticMethod_I_V(a);\n" +
			"}\n";
		assertEquals(expected, wrapper.toRemopla());
	}
	
	public void testStaticArrayMethod() throws Exception {
		
		PDSMethodWrapper wrapper = wrap(A.class, "staticMethod", "([I)V", 
				HeapOption.SIMPLE);
		String expected = 
			"module void wrapperFor_de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_staticMethod_AI_V()\n" +
			"{\n" +
			"int a(3);\n" +
			"int counter(3);\n" +
			"skip (a'<=7 && counter'==0);\n" +
			"do\n" +
			"\t:: counter<a -> skip (heap'[heap_ptr+counter+1]<=7 && counter'==counter+1);\n" +
			"\t:: else -> break;\n" +
			"od;\n" +
			"heap[heap_ptr]=a, heap_ptr=heap_ptr+a+1, a=heap_ptr;\n" +
			"de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_staticMethod_AI_V(a);\n" +
			"}\n";
		assertEquals(expected, wrapper.toRemopla());
	}
	
	public void testStaticAnnotatedArrayMethod() throws Exception {
		
		PDSMethodWrapper wrapper = wrap(A.class, "staticAnnotatedMethod", "([I)V", 
				HeapOption.SIMPLE);
		String expected = 
			"module void wrapperFor_de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_staticAnnotatedMethod_AI_V()\n" +
			"{\n" +
			"int a(3);\n" +
			"int counter(2);\n" +
			"skip (a'<=3 && counter'==0);\n" +
			"do\n" +
			"\t:: counter<a -> skip (heap'[heap_ptr+counter+1]<=1 && counter'==counter+1);\n" +
			"\t:: else -> break;\n" +
			"od;\n" +
			"heap[heap_ptr]=a, heap_ptr=heap_ptr+a+1, a=heap_ptr;\n" +
			"de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_staticAnnotatedMethod_AI_V(a);\n" +
			"}\n";
		assertEquals(expected, wrapper.toRemopla());
	}
	
	public void testStaticAnnotatedMethod() throws Exception {
		
		PDSMethodWrapper wrapper = wrap(A.class, "staticAnnotatedMethod", 
				"(I[ILjava/lang/Object;Ljava/lang/String;[I)V",
				HeapOption.SIMPLE);
		String expected = "module void wrapperFor_de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_staticAnnotatedMethod_IAILjava_lang_Object_Ljava_lang_String_AI_V()\n" +
			"{\n" +
			"int a(2);\n" +
			"int b(3);\n" +
			"int counter(3);\n" +
			"int c(3);\n" +
			"int d(3);\n" +
			"int e(3);\n" +
			"a=undef;\n" +
			"skip (b'<=3 && counter'==0);\n" +
			"do\n" +
			"\t:: counter<b -> skip (heap'[heap_ptr+counter+1]<=3 && counter'==counter+1);\n" +
			"\t:: else -> break;\n" +
			"od;\n" +
			"heap[heap_ptr]=b, heap_ptr=heap_ptr+b+1, b=heap_ptr;\n" +
			"c=0;\n" +
			"d=0;\n" +
			"skip (e'<=7 && counter'==0);\n" +
			"do\n" +
			"\t:: counter<e -> skip (heap'[heap_ptr+counter+1]<=7 && counter'==counter+1);\n" +
			"\t:: else -> break;\n" +
			"od;\n" +
			"heap[heap_ptr]=e, heap_ptr=heap_ptr+e+1, e=heap_ptr;\n" +
			"de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_staticAnnotatedMethod_IAILjava_lang_Object_Ljava_lang_String_AI_V(a, b, c, d, e);\n" +
			"}\n";
		assertEquals(expected, wrapper.toRemopla());
	}
	
	public void testStaticAnnotatedMethodTwodims() throws Exception {
		
		PDSMethodWrapper wrapper = wrap(A.class, "staticAnnotatedMethod", 
				"(I[ILjava/lang/Object;Ljava/lang/String;[I)V"
				, HeapOption.TWODIMS);
		String expected = 
			"module void wrapperFor_de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_staticAnnotatedMethod_IAILjava_lang_Object_Ljava_lang_String_AI_V()\n" +
			"{\n" +
			"int a(2);\n" +
			"int b(3);\n" +
			"int counter(3);\n" +
			"int c(3);\n" +
			"int d(3);\n" +
			"int e(3);\n" +
			"a=undef;\n" +
			"skip (b'<=3 && counter'==0);\n" +
			"do\n" +
			"\t:: counter<b -> skip (heap'[3*(heap_ptr+2) + 2*counter + 0]<=3 && heap'[3*(heap_ptr+2) + 2*counter + 1]<=3 && counter'==counter+1);\n" +
			"\t:: else -> break;\n" +
			"od;\n" +
			"if\n" +
			"\t:: (b*2)-(3*((b*2)/3))==0 && heap_ptr+(b*2)/3+2>7 -> goto HeapOverflow;\n" +
			"\t:: (b*2)-(3*((b*2)/3))==0 && heap_ptr+(b*2)/3+2<=7 -> b=heap_ptr, heap[3*heap_ptr+0]=(b>>2)&1, heap[3*heap_ptr+1]=(b>>1)&1, heap[3*heap_ptr+2]=(b>>0)&1, heap[3*(heap_ptr+1)+0]=0, heap[3*(heap_ptr+1)+1]=1, heap[3*(heap_ptr+1)+2]=0, heap_ptr=heap_ptr+(b*2)/3+2;\n" +
			"\t:: (b*2)-(3*((b*2)/3))!=0 && heap_ptr+(b*2)/3+2+1>7 -> goto HeapOverflow;\n" +
			"\t:: (b*2)-(3*((b*2)/3))!=0 && heap_ptr+(b*2)/3+2+1<=7 -> b=heap_ptr, heap[3*heap_ptr+0]=(b>>2)&1, heap[3*heap_ptr+1]=(b>>1)&1, heap[3*heap_ptr+2]=(b>>0)&1, heap[3*(heap_ptr+1)+0]=0, heap[3*(heap_ptr+1)+1]=1, heap[3*(heap_ptr+1)+2]=0, heap_ptr=heap_ptr+(b*2)/3+2+1;\n" +
			"fi;\n" +
			"c=0;\n" +
			"d=0;\n" +
			"skip (e'<=7 && counter'==0);\n" +
			"do\n" +
			"\t:: counter<e -> skip (heap'[3*(heap_ptr+2) + 3*counter + 0]<=7 && heap'[3*(heap_ptr+2) + 3*counter + 1]<=7 && heap'[3*(heap_ptr+2) + 3*counter + 2]<=7 && counter'==counter+1);\n" +
			"\t:: else -> break;\n" +
			"od;\n" +
			"if\n" +
			"\t:: (e*3)-(3*((e*3)/3))==0 && heap_ptr+(e*3)/3+2>7 -> goto HeapOverflow;\n" +
			"\t:: (e*3)-(3*((e*3)/3))==0 && heap_ptr+(e*3)/3+2<=7 -> e=heap_ptr, heap[3*heap_ptr+0]=(e>>2)&1, heap[3*heap_ptr+1]=(e>>1)&1, heap[3*heap_ptr+2]=(e>>0)&1, heap[3*(heap_ptr+1)+0]=0, heap[3*(heap_ptr+1)+1]=1, heap[3*(heap_ptr+1)+2]=1, heap_ptr=heap_ptr+(e*3)/3+2;\n" +
			"\t:: (e*3)-(3*((e*3)/3))!=0 && heap_ptr+(e*3)/3+2+1>7 -> goto HeapOverflow;\n" +
			"\t:: (e*3)-(3*((e*3)/3))!=0 && heap_ptr+(e*3)/3+2+1<=7 -> e=heap_ptr, heap[3*heap_ptr+0]=(e>>2)&1, heap[3*heap_ptr+1]=(e>>1)&1, heap[3*heap_ptr+2]=(e>>0)&1, heap[3*(heap_ptr+1)+0]=0, heap[3*(heap_ptr+1)+1]=1, heap[3*(heap_ptr+1)+2]=1, heap_ptr=heap_ptr+(e*3)/3+2+1;\n" +
			"fi;\n" +
			"de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_staticAnnotatedMethod_IAILjava_lang_Object_Ljava_lang_String_AI_V(a, b, c, d, e);\n" +
			"}\n";
		assertEquals(expected, wrapper.toRemopla());
	}
	
	public void testVirtualMethod() throws Exception {
		
		PDSMethodWrapper wrapper = wrap(A.class, "virtualMethod", "()V", 
				HeapOption.SIMPLE);
		String expected = 
			"module void wrapperFor_de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_virtualMethod_V()\n" +
			"{\n" +
			"int this(3);\n" +
			"skip (this'==heap_ptr && heap_ptr'==heap_ptr+0+1 && heap'[heap_ptr]==1);\n" +
			"de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_virtualMethod_V(this);\n" +
			"}\n";
		assertEquals(expected, wrapper.toRemopla());
	}
	
	public void testVirtualMethodTwodims() throws Exception {
		
		PDSMethodWrapper wrapper = wrap(A.class, "virtualMethod", "()V", 
				HeapOption.TWODIMS);
		String expected = 
			"module void wrapperFor_de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_virtualMethod_V()\n" +
			"{\n" +
			"int this(3);\n" +
			"this=heap_ptr, heap_ptr=heap_ptr+0+1, heap[3*heap_ptr+0]=0, heap[3*heap_ptr+1]=0, heap[3*heap_ptr+2]=1;\n" +
			"de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_virtualMethod_V(this);\n" +
			"}\n";
		assertEquals(expected, wrapper.toRemopla());
	}
	
	public void testVirtualIntMethod() throws Exception {
		
		PDSMethodWrapper wrapper = wrap(A.class, "virtualMethod", "(I)V", 
				HeapOption.SIMPLE);
		String expected = 
			"module void wrapperFor_de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_virtualMethod_I_V()\n" +
			"{\n" +
			"int this(3);\n" +
			"int a(3);\n" +
			"skip (this'==heap_ptr && heap_ptr'==heap_ptr+0+1 && heap'[heap_ptr]==1);\n" +
			"a=undef;\n" +
			"de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_virtualMethod_I_V(this, a);\n" +
			"}\n";
		assertEquals(expected, wrapper.toRemopla());
	}
	
	public void testVirtualIntMethodTwodims() throws Exception {
		
		PDSMethodWrapper wrapper = wrap(A.class, "virtualMethod", "(I)V", 
				HeapOption.TWODIMS);
		String expected = 
			"module void wrapperFor_de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_virtualMethod_I_V()\n" +
			"{\n" +
			"int this(3);\n" +
			"int a(3);\n" +
			"this=heap_ptr, heap_ptr=heap_ptr+0+1, heap[3*heap_ptr+0]=0, heap[3*heap_ptr+1]=0, heap[3*heap_ptr+2]=1;\n" +
			"a=undef;\n" +
			"de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_A_virtualMethod_I_V(this, a);\n" +
			"}\n";
		assertEquals(expected, wrapper.toRemopla());
	}
	
	private static class A {
		
		static void staticMethod() {
			
		}
		
		static void staticMethod(int a) {
			
		}
		
		static void staticMethod(int[] a) {
			
		}
		
		@PDSBits({"a=2", "a[] := 1"})
		static void staticAnnotatedMethod(int[] a) {
			
		}
		
		@PDSBits({"2", "b=2", "b[] = 2"})
		static void staticAnnotatedMethod(int a, int[] b, Object c, String d,
				int[] e) {
			
		}
		
		void virtualMethod() {
			
		}
		
		void virtualMethod(int a) {
			
		}
	}
	
	public void testField() throws Exception {
		
		PDSMethodWrapper wrapper = wrap(B.class, "virtualMethod", "(I)V", 
				HeapOption.SIMPLE);
		String expected = 
			"module void wrapperFor_de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_B_virtualMethod_I_V()\n" +
			"{\n" +
			"int this(3);\n" +
			"int a(3);\n" +
			"skip (de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_B_a'<=1 && de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_B_b'<=3 && de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_B_c'<=7 && de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_B_d'==0);\n" + 
			"skip (this'==heap_ptr && heap_ptr'==heap_ptr+4+1 && heap'[heap_ptr]==1 && heap'[heap_ptr+0+1]<=3 && heap'[heap_ptr+1+1]<=1 && heap'[heap_ptr+2+1]<=7 && heap'[heap_ptr+3+1]==0);\n" +
			"a=undef;\n" +
			"de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_B_virtualMethod_I_V(this, a);\n" +
			"}\n";
		assertEquals(expected, wrapper.toRemopla());
	}
	
	public void testFieldTwodims() throws Exception {
		
		PDSMethodWrapper wrapper = wrap(B.class, "virtualMethod", "(I)V", 
				HeapOption.TWODIMS);
		String expected = 
			"module void wrapperFor_de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_B_virtualMethod_I_V()\n" +
			"{\n" +
			"int this(3);\n" +
			"int a(3);\n" +
			"skip (de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_B_a'<=1 && de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_B_b'<=3 && de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_B_c'<=7 && de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_B_d'==0);\n" + 
			"this=heap_ptr, heap_ptr=heap_ptr+3+1, heap[3*heap_ptr+0]=0, heap[3*heap_ptr+1]=0, heap[3*heap_ptr+2]=1, heap[3*(heap_ptr+1)+0+0]=undef, heap[3*(heap_ptr+1)+0+1]=undef, heap[3*(heap_ptr+1)+2+0]=undef, heap[3*(heap_ptr+1)+3+0]=undef, heap[3*(heap_ptr+1)+3+1]=undef, heap[3*(heap_ptr+1)+3+2]=undef, heap[3*(heap_ptr+1)+6+0]=0, heap[3*(heap_ptr+1)+6+1]=0, heap[3*(heap_ptr+1)+6+2]=0;\n" +
			"a=undef;\n" +
			"de_uni_stuttgart_fmi_szs_jmoped_PDSMethodWrapperTest_B_virtualMethod_I_V(this, a);\n" +
			"}\n";
		assertEquals(expected, wrapper.toRemopla());
	}
	
	private static class B {
		
		@PDSFieldBits(2)
		int i;
		@PDSFieldBits(1)
		int j;
		int k;
		String s;
		@PDSFieldBits(1)
		static int a;
		@PDSFieldBits(2)
		static int b;
		static int c;
		static Object d;
		
		void virtualMethod(int a) {
			
		}
	}
}
