/*
    Copyright (C) 2005 Felix Berger <bflat1@gmx.net>

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package de.uni_stuttgart.fmi.szs.jmoped;

import java.io.File;

import org.gjt.jclasslib.structures.ClassFile;

import de.uni_stuttgart.fmi.szs.jmoped.PDSClassTest.HasNoMethodClass;

public class PDSTest extends PDSTestCase 
{
	
	public void testConstructor() throws Exception
	{
		File classFile = findClassFile(TestClass.class);
		assertNotNull(classFile);
		PDS structure = new PDS(classFile.getAbsolutePath());
		assertEquals(1, structure.getPDSClasses().size());
		structure.translate();
		assertTrue(structure.toRemopla().length() > 0);
	}
	
	public void testPDSInfoConstructor() throws Exception
	{
		ClassFile classFile = BytecodeUtils.findClassFile(TestClass.class.getName().replace('.', '/'));
		PDSInfo info = new PDSInfo(classFile);
		PDS newPds = new PDS(info);
		newPds.translate();
		
		PDS oldPds = getPDS(TestClass.class);
		oldPds.translate();
		assertTrue(newPds.toRemopla().length() > 0);
		assertEquals(oldPds.toRemopla(), newPds.toRemopla());
		assertEquals(oldPds.getPDSClasses(), newPds.getPDSClasses());
	}
	
	public static class TestClass
	{
		public void test()
		{
		}
	}
	
	public void testGetInitialMethod() throws Exception
	{
		PDS pds = getPDS(HasMainAndStaticMethod.class);
		assertNotNull(pds.getRootPDSMethod());
		
		pds = new PDS(HasMainAndStaticMethod.class, "main", "([Ljava/lang/String;)V");
		PDSClass clazz = pds.getRootPDSClass();
		assertEquals(clazz.getMethodMain(), pds.getRootPDSMethod());
		
		pds = new PDS(HasMainAndStaticMethod.class, "other", "()V");
		clazz = pds.getRootPDSClass();
		PDSMethod method = clazz.getMethod("other", "()V");
		assertNotNull(method);
		assertEquals(method, pds.getRootPDSMethod());
		
		pds = getPDS(HasNoMethodClass.class);
		assertNull(pds.getRootPDSMethod());
	}
	
	public static class HasMainAndStaticMethod
	{
		public static void main(String[] args)
		{
		}
		
		public static void other()
		{
		}
	}
	
	public void testClinitDependencies() throws Exception
	{
		PDS pds = getPDS(ClinitBClass.class);
	}
	
	public static class ClinitAClass
	{
		public static int a = 1;
		
		public static int getA()
		{
			return a;
		}
	}
	
	public static class ClinitBClass
	{
		public static int b = ClinitAClass.a;
	}

}
