/*
    Copyright (C) 2005 Felix Berger <bflat1@gmx.net>

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package de.uni_stuttgart.fmi.szs.jmoped;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.StringTokenizer;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageTester;

import junit.framework.TestCase;

public abstract class PDSTestCase extends TestCase 
{
	
	/**
	 * Finds a class file in the class path, a file with in jars is currently not
	 * found.
	 * @param clazz
	 * @return
	 * @deprecated use {@link PDSInfo#findClassFile(String)} eventually
	 */
	protected File findClassFile(Class clazz) throws Exception
	{
		String subFilename = clazz.getName().replace('.', File.separatorChar) + ".class";
		String classPath = System.getProperty("java.class.path",".");

		StringTokenizer st = new StringTokenizer(classPath, 
				System.getProperty("path.separator"));
		while (st.hasMoreTokens()) {
			File f = new File(st.nextToken(), subFilename);
			if (f.exists()) {
				return f;
			}
		}
		throw new FileNotFoundException("file for class " + clazz.getName() 
				+ " could not be found");
	}
	
	/**
	 * Creates a PDS object for the given class.
	 * @param clazz
	 * @return
	 * @throws Exception
	 */
	protected PDS getPDS(Class clazz) throws Exception
	{
		return new PDS(clazz);
	}

	/**
	 * Looks for the class file in the class path and creates the corresponding
	 * PDSClass.
	 * @param clazz
	 * @return
	 * @throws Exception
	 */
	protected PDSClass getPDSClass(Class clazz) throws Exception
	{
		PDS pds = getPDS(clazz);
		return pds.getRootPDSClass();
	}
	
	/**
	 * Searches for the method with name and descriptor in the array of methods
	 * and returns it.
	 * @param methods
	 * @param name
	 * @param descriptor can be null
	 * @return
	 */
	protected CoverageMethod getMethod(CoverageMethod[] methods, String name,
			String descriptor)
	{
		for (CoverageMethod method : methods) {
			PDSMethod m = method.getMethod();
			if (m.isName(name) && (descriptor == null 
					|| m.isNameAndDescriptor(name, descriptor))) {
				return method;
			}
		}
		return null;
	}

	protected CoverageMethod getMethod(CoverageMethod[] methods, String name)
	{
		return getMethod(methods, name, null);
	}
	
	protected void print(CoverageMethod method)
	{
		System.out.println(method);
		for (int i = 0; i < method.getLineTable().length; i++) {
			System.out.println(i + ": " + method.getLineNumber(i) + ": " 
					+ method.getLineState(i));
		}
	}
	
	/**
	 * Returns <code>true</code> if emma's RT class can be loaded.
	 * <p>
	 * This a good enough criterion to assume the bytecode has been
	 * instrumented.
	 */
	protected boolean isRunUnderEmma()
	{
		try {
			Class emmaClass = Class.forName("com.vladium.emma.rt.RT");
			return true;
		}
		catch (Exception e) {
			return false;
		}
	}

	/**
	 * Creates a CoverageTester and sets the initial method, 
	 * asserting that it is not null, etc.
	 * @param clazz
	 * @param name
	 * @param descriptor
	 */
	protected CoverageTester getTester(Class clazz, String name, String descriptor) 
		throws Exception 
	{
		CoverageTester tester = new CoverageTester(clazz, name, descriptor);
		PDS pds = tester.getPDS();
		PDSMethod initialMethod = pds.getRootPDSClass().getMethod(name,
				descriptor);
		
		assertNotNull(initialMethod);
		assertEquals(initialMethod, pds.getRootPDSMethod());
		
		return tester;
	}
	
}
