package de.uni_stuttgart.fmi.szs.jmoped.annotation;

import java.util.List;

import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.FieldInfo;
import org.gjt.jclasslib.structures.MethodInfo;
import org.gjt.jclasslib.structures.elementvalues.ElementValue;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.PDSClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSField;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;
import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;

public class PDSAnnotationUtilsTest extends PDSTestCase {
	
	PDS pds;
	PDSClass clazz;
	MethodInfo methodInfo;
	ClassFile classFile;
	ElementValue[] bits;
	
	protected void setUp() throws Exception {
		
		pds = new PDS(ParameterBitsTestClass.class, "test", "(I[II)V");
		clazz = pds.getRootPDSClass();
		methodInfo = clazz.getMethod("test", "(I[II)V").getMethodInfo();
		classFile = clazz.getClassFile();
		bits = PDSAnnotationUtils.getBits(methodInfo);
		assertNotNull(bits);
	}

	public void testGetBit() throws Exception {
		
		assertEquals(2, PDSAnnotationUtils.getBit(bits, classFile, null, 0));
		assertEquals(2, PDSAnnotationUtils.getBit(bits, classFile, "j", -1));
		assertEquals(3, PDSAnnotationUtils.getBit(bits, classFile, "k", -1));
		assertEquals(1, PDSAnnotationUtils.getBit(bits, classFile, "l", -1));
		assertEquals(3, PDSAnnotationUtils.getBit(bits, classFile, "m", -1));
		assertEquals(0, PDSAnnotationUtils.getBit(bits, classFile, "o", -1));
	}
	
	public void testGetArrayBits() throws Exception {
		
		assertEquals(1, PDSAnnotationUtils.getArrayBits(bits,
				methodInfo, "j"));
		assertEquals(2, PDSAnnotationUtils.getArrayBits(bits,
				methodInfo, "m"));
		assertEquals(4, PDSAnnotationUtils.getArrayBits(bits,
				methodInfo, "p"));
		
		MethodInfo info = clazz.getMethod("<clinit>", "()V").getMethodInfo();
		assertEquals(5, PDSAnnotationUtils.getDefaultArrayBits(info));
	}
	
	public void testGetFieldBit() throws Exception {
		
		List<PDSField> fieldList = clazz.getStaticFieldList();	
		assertEquals(3, fieldList.size());
		for (int i = 0; i < fieldList.size(); i++) {
			assertEquals(i, fieldList.get(i).getBits());
		}
		
		fieldList = clazz.getNonStaticFieldList();
		assertEquals(3, fieldList.size());
		for (int i = 0; i < fieldList.size(); i++) {
			assertEquals(i, fieldList.get(i).getBits());
		}
	}
	
	public void testParameterBits() throws Exception {
		
		MethodInfo info = clazz.getMethod("test", "(II[I)V").getMethodInfo();
		ElementValue[] values = PDSAnnotationUtils.getBits(info);
		assertNotNull(values);
		
		ClassFile cf = info.getClassFile();
		assertEquals(3, PDSAnnotationUtils.getBit(values, cf, "", 0));
		assertEquals(1, PDSAnnotationUtils.getBit(values, cf, "", 1));
		assertEquals(4, PDSAnnotationUtils.getBit(values, cf, "", 2));
		
		PDS pds = new PDS(ParameterBitsTestClass.class, "test", "([IIII)V");
		PDSClass clazz = pds.getRootPDSClass();
		PDSMethod method = clazz.getMethod("test", "([IIII)V");
		info = method.getMethodInfo();
		values = PDSAnnotationUtils.getBits(info);
		assertNotNull(values);
		
		cf = info.getClassFile();
		assertEquals(1, PDSAnnotationUtils.getBit(values, cf, "", 0));
		assertEquals(2, PDSAnnotationUtils.getBit(values, cf, "", 1));
		assertEquals(5, PDSAnnotationUtils.getBit(values, cf, "", 2));
		
		assertEquals(1, method.getPDSLocalVar(1).getBits());
		assertEquals(2, method.getPDSLocalVar(2).getBits());
		assertEquals(5, method.getPDSLocalVar(3).getBits());
		assertEquals(0, method.getPDSLocalVar(4).getBits());
	}
	
	@PDSArrayBits(5)
	public static class ParameterBitsTestClass {
		
		private static int f1 = 2;
		
		@PDSFieldBits(1)
		private static int f2;
		
		@PDSBits({"2"})
		private static int f3;
		
		@PDSFieldBits(7)
		private Object f4;
		
		@PDSBits({"1"})
		private int f5;
		
		@PDSBits({"2"})
		@PDSFieldBits(1)
		private int f6;
		
		@PDSArrayBits(4)
		@PDSBits({"2", "k = 3", "j=2", "j[] 1", "l:=1", "m 3", "m[] 2"})
		@PDSParameterBits({3, 1, 4})
		public static void test(int i, int[] j, int k) {
			
			int l;
			int[][] m;
			int o;
			int[] p;
			
			test(i, k, j);
		}
		
		@PDSParameterBits({3, 1, 4})
		public static void test(int i, int j, int[] k) {
		}
		
		@PDSParameterBits({1, 2, 5})
		public void test(int[] i, int j, int k, int l) {
		}
	}
	
}
