package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;

public class DescriptorParserTest extends PDSTestCase 
{

	public void testEmptyParse()
	{
		Parameter[] params = DescriptorParser.parse("()V");
		assertEquals(0, params.length);
	}
	
	public void testValues()
	{
		Parameter[] params = DescriptorParser.parse("(I)V");
		assertEquals(1, params.length);
		assertTrue(params[0].isValue());
		assertFalse(params[0].isValueArray());
		assertEquals("I", params[0].getType());
		
		params = DescriptorParser.parse("(III)V");
		assertEquals(3, params.length);
		
		params = DescriptorParser.parse("(ZIZ)V");
		assertEquals(3, params.length);
	}
	
	public void testValueArray()
	{
		Parameter[] params = DescriptorParser.parse("([I)V");
		assertEquals(1, params.length);
		assertFalse(params[0].isValue());
		assertTrue(params[0].isValueArray());
		
		params = DescriptorParser.parse("([I[[Z)V");
		assertEquals(2, params.length);
		assertFalse(params[0].isValue());
		assertTrue(params[0].isValueArray());
		assertFalse(params[1].isValue());
		assertTrue(params[1].isValueArray());
		assertEquals("Z", params[1].getType());
	}
	
	public void testReference()
	{
		Parameter[] params = DescriptorParser.parse("(Ljava/lang/Object;)V");
		assertEquals(1, params.length);
		assertTrue(params[0].isReference());
		
		params = DescriptorParser.parse("(Ljava/lang/Object;Ljava/lang/String;)V");
		assertEquals(2, params.length);
		assertTrue(params[0].isReference());
		assertTrue(params[1].isReference());
	}
	
	public void testReferenceArray()
	{
		Parameter[] params = DescriptorParser.parse("([Ljava/lang/Object;)V");
		assertEquals(1, params.length);
		assertTrue(params[0].isReferenceArray());
		
		params = DescriptorParser.parse("([Ljava/lang/Object;[Ljava/lang/String;)V");
		assertEquals(2, params.length);
		assertTrue(params[0].isReferenceArray());
		assertEquals("java/lang/Object", params[0].getType());
		assertTrue(params[1].isReferenceArray());
		assertEquals("java/lang/String", params[1].getType());
	}
	
	public void testMixed()
	{
		Parameter[] params =
			DescriptorParser.parse("(I[Z[Ljava/lang/Object;[ILjava/lang/String;Z)V");
		assertEquals(6, params.length);
		assertTrue(params[0].isValue());
		assertTrue(params[1].isValueArray());
		assertTrue(params[2].isReferenceArray());
		assertTrue(params[3].isValueArray());
		assertTrue(params[4].isReference());
		assertTrue(params[5].isValue());
		assertEquals("Z", params[5].getType());
	}
	
}
