package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.HeapOverflowTest.AllocatorClass;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.TraceGeneratorTest.InputValueClassCoverageListener;

public class JavaHeapOverflowResolverTest extends PDSTestCase 
{

	private int getMissing(Class clazz, int heapSize) throws Exception
	{
		CoverageTester tester = new CoverageTester(clazz);
		InputValueClassCoverageListener listener = new InputValueClassCoverageListener();
		tester.addCoverageListener(listener);
		tester.getInfo().setHeapSize(heapSize);
		
		tester.test();
		assertNotNull(listener.result);
		assertEquals(1, listener.overflowMap.size());
		
		Transition transition = listener.overflowMap.values().iterator().next();
		assertNotNull(transition);
		
		return JavaHeapOverflowResolver.getMissingHeapFields(listener.result, transition);
	}
	
	public void testHeapOverFlowSizeOfNew() throws Exception
	{
		assertEquals(2, getMissing(AllocatorClass.class, 7));
		assertEquals(1, getMissing(AllocatorClass.class, 6));
	}
	
	public void testHeapOverFlowSizeOfNewArray() throws Exception
	{
		assertEquals(1, getMissing(ArrayInstantiationClass.class, 7));
		assertEquals(2, getMissing(ArrayInstantiationClass.class, 6));
		assertEquals(3, getMissing(ArrayInstantiationClass.class, 5));
		assertEquals(4, getMissing(ArrayInstantiationClass.class, 4));
		assertEquals(5, getMissing(ArrayInstantiationClass.class, 3));
		assertEquals(6, getMissing(ArrayInstantiationClass.class, 2));
		assertEquals(7, getMissing(ArrayInstantiationClass.class, 1));
	}
	
	public static class ArrayInstantiationClass
	{
		public static void main(String[] args) 
		{
			int[] array = new int[6];
		}
	}
	
	public void testHeapOverFlowSizeOfMultiNewArray() throws Exception
	{
		assertEquals(3, getMissing(MultiArrayInstantionClass.class, 7));
		assertEquals(4, getMissing(MultiArrayInstantionClass.class, 6));
		assertEquals(8, getMissing(MultiArrayInstantionClass.class, 2));
		assertEquals(9, getMissing(MultiArrayInstantionClass.class, 1));
	}
	
	public static class MultiArrayInstantionClass
	{
		public static void main(String[] args) 
		{
			int[][] array = new int[2][2];
		}
	}
	
	public void testGetMultiArraySize()
	{
		// dimensions are ordered backwards, as they lie on the stack
		assertEquals(9, JavaHeapOverflowResolver.getMultiArraySize(new int[] { 2, 2 }));
		assertEquals(5, JavaHeapOverflowResolver.getMultiArraySize(new int[] { 2, 1 }));
		assertEquals(13, JavaHeapOverflowResolver.getMultiArraySize(new int[] { 3, 2, 1 }));
	}
	
}
