package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.awt.event.ComponentAdapter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.Delayed;

import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.InterfaceImplementerClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.SubSubTestClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.SubTestClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.TestClass;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.PDSInstCoverageTest.InstanceOfClass;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.TraceGeneratorTest.InputValueClassCoverageListener;

// TODO remove println and add more asserts
public class JavaMethodArgumentsResolverTest extends PDSTestCase 
{

	public void testRetrieveStaticIntArguments() throws Exception
	{
		CoverageTester tester = new CoverageTester(StaticIntArgumentsClass.class);
		
		MethodsAndConfig pair = getMethodsAndConfig(tester);
		
		JavaMethodArgumentsResolver resolver = new JavaMethodArgumentsResolver();
		MethodCall[] methodCalls = resolver.resolveMethodCalls(pair.startConfig, 
				getMethod(pair.methods, "test").getMethod(), tester.getPDS());
		
		MethodArgument[] args = methodCalls[methodCalls.length - 1].getArguments();
		assertEquals(3, args.length);
		assertEquals(5, args[0].getVariable().intValue());
		assertEquals(7, args[1].getVariable().intValue());
		assertEquals(3, args[2].getVariable().intValue());
	}
	
	private MethodsAndConfig getMethodsAndConfig(CoverageTester tester)
	{
		InputValueClassCoverageListener listener = new InputValueClassCoverageListener();
		tester.addCoverageListener(listener);
		
		MethodsAndConfig pair = new MethodsAndConfig();
		pair.methods = tester.test();
		
		CoverageResult result = listener.result;
		assertNotNull(result);

		Transition transition = listener.transitions[0];
		if (transition == null) {
			transition = listener.map.values().iterator().next();
		}
		assertNotNull(transition);
		
		IntConfig startConfig = TraceGenerator.getTraceConfig(result, transition);
		assertNotNull(startConfig);
		pair.startConfig = startConfig;
		
		return pair;
	}
	
	public static class StaticIntArgumentsClass
	{
		
		public static void main(String[] args)
		{
			int k = 3;
			test(5, 7, k);
		}

		public static void test(int i, int j, int k) 
		{
			assert(false);
		}
		
	}
	
	public void testStaticArrayArguments() throws Exception
	{
		CoverageTester tester = new CoverageTester(StaticArrayArgumentsClass.class);
		tester.getInfo().setHeapSize(15);
		
		MethodsAndConfig pair = getMethodsAndConfig(tester);
		
		JavaMethodArgumentsResolver resolver = new JavaMethodArgumentsResolver();
		MethodCall[] methodCalls = resolver.resolveMethodCalls(pair.startConfig, 
				getMethod(pair.methods, "test").getMethod(), tester.getPDS());
		
		MethodArgument[] args = methodCalls[methodCalls.length - 1].getArguments();
		assertEquals(3, args.length);
		assertTrue(Arrays.equals(new int[0], args[0].getArray()));
		assertTrue(Arrays.equals(new int[] { 1, 2, 3 }, args[1].getArray()));
		assertTrue(Arrays.equals(new int[] { 1 }, args[2].getArray()));
	}
	
	/**
	 * Throw away test case to find out, we need heapsize of 15. 
	 * @throws Exception
	 */
	public void testReachable() throws Exception
	{
		PostStarReachabilityTester tester = 
			new PostStarReachabilityTester(StaticArrayArgumentsClass.class);
		tester.getInfo().setHeapSize(15);
		
		LabelReachedListener listener = new LabelReachedListener(tester,
				"de_uni_stuttgart_fmi_szs_jmoped_coverage_" +
				"JavaMethodArgumentsResolverTest_StaticArrayArgumentsClass_test_" +
				"AIAIAI_V6");
		
		tester.test(listener.label);
		assertTrue(listener.reachedLabel());
	}
	
	public static class StaticArrayArgumentsClass
	{
		public static void main(String[] args)
		{
			int[] k =  { 1 };
			test(new int[0], new int[] { 1, 2 , 3 } , k);
		}

		public static void test(int[] i, int[] j, int[] k) 
		{
			assert(false);
		}
		
	}
	
	public void testStaticCallTrace() throws Exception
	{
		CoverageTester tester = new CoverageTester(StaticCallTraceClass.class);

		MethodsAndConfig pair = getMethodsAndConfig(tester);
		
		IntConfig config = pair.startConfig;
		JavaMethodArgumentsResolver resolver = new JavaMethodArgumentsResolver();
		MethodCall[] callTrace = resolver.resolveMethodCallTrace(config, tester.getPDS());
		printCallTrace(callTrace);
	}
	
	void printCallTrace(MethodCall[] callTrace)
	{
		for (MethodCall call : callTrace) {
			int stackLength = call.getConfig().getStackLength() - 1;
			char[] indent = new char[stackLength];
			Arrays.fill(indent, '\t');
			System.out.println(new String(indent) + call);
		}
	}
	
	void printCallTraceTree(TraceNode[] roots)
	{
		printCallTraceTree(roots, 0);
	}
	
	void printCallTraceTree(TraceNode[] roots, int indent)
	{
		char[] space = new char[indent];
		Arrays.fill(space, '\t');
		for (TraceNode root : roots) {
			System.out.println(new String(space) + root);
			printCallTraceTree(root.getChildren(), indent + 1);
		}
	}
	
	public static class StaticCallTraceClass
	{
		public static void main(String[] args)
		{
			test(1, 3, 5);
		}
		
		public static void test(int i, int j, int k)
		{
			test2(i, j, k, new int[] { 1, 2 });
		}
		
		public static void test2(int i, int j, int k, int[] l)
		{
			assert(false);
		}
		
	}
	
	public void testBooleans() throws Exception
	{
		CoverageTester tester = new CoverageTester(BooleanMethodArgsClass.class);
		MethodsAndConfig pair = getMethodsAndConfig(tester);
		
		CoverageMethod method = getMethod(pair.methods, "test");
		assertNotNull(method);
		
		JavaMethodArgumentsResolver resolver = new JavaMethodArgumentsResolver();
		
		MethodCall[] calls = 
			resolver.resolveMethodCalls(pair.startConfig, method.getMethod(), 
				tester.getPDS());
		
		MethodArgument[] args = calls[calls.length - 1].getArguments();
		assertEquals(3, args.length);
		assertEquals(true, args[0].getVariable().booleanValue());
		assertEquals(false, args[1].getVariable().booleanValue());
		assertTrue(Arrays.equals(new int[] { 0, 1 }, args[2].getArray()));
	}
	
	public static class BooleanMethodArgsClass
	{
		public static void main(String[] args)
		{
			test(true, false, new boolean[] { false, true });
		}
		
		public static void test(boolean i, boolean j, boolean[] k)
		{
			assert(false);
		}
	}
	
	public void testMultiCallsAndRecursionTrace() throws Exception
	{
		CoverageTester tester = 
			getTester(MultiCallsAndRecursionClass.class, "other", "(I)V");
		MethodsAndConfig pair = getMethodsAndConfig(tester);
		
		JavaMethodArgumentsResolver resolver = new JavaMethodArgumentsResolver();
		MethodCall[] trace = resolver.resolveMethodCallTrace(pair.startConfig, 
				tester.getPDS());
		
		printCallTrace(trace);
	}
	
	public static class MultiCallsAndRecursionClass
	{
		public static void main(String[] args)
		{
			other(Integer.parseInt(""));
		}
		
		public static void other(int i) 
		{
			if (i == 0) {
				other(i + 1);
			}
			else {
				test(i + 1);
				test(i + 2);
			}
		}
		
		public static void test(int i)
		{
			assert(i != 3);
		}
	}
	
	public void testBubbleSortAssertionTrace() throws Exception
	{
		CoverageTester tester = getTester(BubbleSort.class, "sort", "([I)V");
		tester.getInfo().setHeapSize(4);
		MethodsAndConfig pair = getMethodsAndConfig(tester);
		
		JavaMethodArgumentsResolver resolver = new JavaMethodArgumentsResolver();
		MethodCall[] trace = resolver.resolveMethodCallTrace(pair.startConfig, 
				tester.getPDS());
		
		printCallTrace(trace);
	}
	
	public void testBubbleSortConstructor() throws Exception
	{
		CoverageTester tester = getTester(BubbleInstanceSort.class, "sort", "([I)V");
		tester.getInfo().setHeapSize(4);
		MethodsAndConfig pair = getMethodsAndConfig(tester);
		
		JavaMethodArgumentsResolver resolver = new JavaMethodArgumentsResolver();
		TraceNode[] roots = resolver.resolveMethodCallTraceTree(pair.startConfig, 
				tester.getPDS());
		
		TraceNode current = roots[roots.length - 1];
		
		MethodCall sortCall = current.getMethodCall();
		assertEquals("sort", sortCall.getMethod().getName());
		MethodArgument instance = sortCall.getInstance();
		assertNotNull(instance);
		MethodCall constructor = instance.getConstructorCall();
		assertNull(constructor);
	}
	
	public static class BubbleSort
	{
		public static void main(String[] args)
		{
			int [] array  = new int[2];
			for (int i = 0; i < array.length; i++) {
				array[i] = Integer.parseInt(args[0]);
			}
			sort(array);
			assert((new BubbleSort().isSorted(array)));
		}
		
		public static void sort(int[] array)
		{
			for (int i = 0; i < array.length; i++) {
				boolean changed = false;
				for (int j = 0; j < array.length - 1; j++) {
					if (array[j] > array[j+1]) {
						int tmp = array[j];
						array[j] = array[j+1];
						array[j+1] = tmp;
						changed = true;
					}
				}
				if (!changed) {
					break;
				}
			}
			if (array.length >= 2) {
				assert(array[1] <= array[0]);
			}
		}
		
		boolean isSorted(int[] array)
		{
			for (int i = 0; i < array.length - 1; i++) {
				if (array[i] > array[i+1]) {
					return false;
				}
			}
			return true;
		}
	}
	
	public static class BubbleInstanceSort
	{		
		public static void main(String[] args)
		{
			int [] array  = new int[2];
			for (int i = 0; i < array.length; i++) {
				array[i] = Integer.parseInt(args[0]);
			}
			new BubbleInstanceSort().sort(array);
		}
		
		public void sort(int[] array)
		{
			for (int i = 0; i < array.length; i++) {
				boolean changed = false;
				for (int j = 0; j < array.length - 1; j++) {
					if (array[j] > array[j+1]) {
						int tmp = array[j];
						array[j] = array[j+1];
						array[j+1] = tmp;
						changed = true;
					}
				}
				if (!changed) {
					break;
				}
			}
			assert(!isSorted(array));
		}
		
		boolean isSorted(int[] array)
		{
			for (int i = 0; i < array.length - 1; i++) {
				if (array[i] > array[i+1]) {
					return false;
				}
			}
			return true;
		}
	}
	
	
	public void testResolveConstructorCall() throws Exception
	{
		CoverageTester tester = 
			new CoverageTester(ResolveConstructorCallClass.class);
		MethodsAndConfig pair = getMethodsAndConfig(tester);
		
		CoverageMethod method = getMethod(pair.methods, "test");
		assertNotNull(method);
		
		JavaMethodArgumentsResolver resolver = new JavaMethodArgumentsResolver();
		MethodCall[] calls = 
			resolver.resolveMethodCalls(pair.startConfig, method.getMethod(),
					tester.getPDS());
		MethodCall constructor = calls[0].getArguments()[0].getConstructorCall();
		assertNotNull(constructor);
		
		MethodArgument[] args = constructor.getArguments();
		assertEquals(1, args.length);
		assertEquals(3, args[0].getVariable().intValue());
	}
	
	public static class ResolveConstructorCallClass
	{
			private int i;
			
			public ResolveConstructorCallClass(int i)
			{
				this.i = i;
			}
			
			public static void main(String[] args)
			{
				ResolveConstructorCallClass instance =
					new ResolveConstructorCallClass(3);
				test(instance);
			}
			
			public static void test(ResolveConstructorCallClass instance)
			{
				assert(false);
			}
	}
	
	public void testResolveMethodCallTraceTree() throws Exception
	{
		// for main method
		CoverageTester tester = 
			new CoverageTester(MultiCallsAndRecursionClass.class);
		MethodsAndConfig pair = getMethodsAndConfig(tester);
		
		JavaMethodArgumentsResolver resolver = new JavaMethodArgumentsResolver();
		
		TraceNode[] roots = 
			resolver.resolveMethodCallTraceTree(pair.startConfig, tester.getPDS());
		
		assertEquals(2, roots.length);
		
		// for non-deterministic driver
		tester = getTester(MultiCallsAndRecursionClass.class, "other", "(I)V");
		pair = getMethodsAndConfig(tester);
		
		roots = resolver.resolveMethodCallTraceTree(pair.startConfig, tester.getPDS());
		assertEquals(2, roots.length);
	}
	
	public void testConstructorCalls() throws Exception
	{
		CoverageTester tester = new CoverageTester(InstanceOfClass.class);
		MethodsAndConfig pair = getMethodsAndConfig(tester);
		
		JavaMethodArgumentsResolver resolver = new JavaMethodArgumentsResolver();
		
		TraceNode[] roots = 
			resolver.resolveMethodCallTraceTree(pair.startConfig, tester.getPDS());
		
		assertEquals(3, roots.length);
		assertEquals(2, roots[2].getChildren().length);
		assertTrue(roots[2].getChildren()[0].getChildren()[0].hasChildren());
	}
	
	private static class InstanceOfClass 
	{
		public static void main(String[] args) 
		{
			SubSubTestClass instance = new SubSubTestClass();
			if (instance instanceof SubTestClass) {
				if (instance instanceof TestClass) {
					if (!(instance instanceof Comparable)) {
						InterfaceImplementerClass impl = new InterfaceImplementerClass();
						if (impl instanceof Comparable) {
							Delayed delayed = impl;
							if (delayed instanceof Comparable) {
								if (impl instanceof ComponentAdapter) {
									if (!(impl instanceof Iterable)) {
										// assert is only reached if all instanceof tests work correctly
										assert(false);
									}
								}
							}
						}
					}
				}
			}
		}
	}
	
	public void testRecursion() throws Exception
	{
		CoverageTester tester = new CoverageTester(RecursiveClass.class);
		MethodsAndConfig pair = getMethodsAndConfig(tester);
		
		JavaMethodArgumentsResolver resolver = new JavaMethodArgumentsResolver();
		
		TraceNode[] roots = 
			resolver.resolveMethodCallTraceTree(pair.startConfig, tester.getPDS());
		// TODO test depth 
	}
	
	public static class RecursiveClass
	{
		public static void main(String[] args) 
		{
			recurse(1);
		}
		
		public static void recurse(int i)
		{
			assert(i != 4);
			recurse(i+1);
		}
	}
	
	private static class MethodsAndConfig
	{
		public CoverageMethod[] methods;
		public IntConfig startConfig;
	}
	
	public static class LabelReachedListener extends CoverageAdapter
	{
		public String label;
		public List<String> reachedLabels = new ArrayList<String>();
		
		public LabelReachedListener(Coverage coverage, String label)
		{
			this.label = label;
			coverage.addCoverageListener(this);
			coverage.setEnabledNotifyofReachedLabels(true);
		}
		
		@Override
		public void labelReached(String label) 
		{
			reachedLabels.add(label);
		}
		
		public boolean reachedLabel()
		{
			return reachedLabels.contains(label);
		}
	}
}
