package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.TraceGeneratorTest.InputValueClassCoverageListener;

public class LineTest extends PDSTestCase
{
	
	private MethodCall getMethodCall(Transition transition, InputValueClassCoverageListener listener)
	{
		IntConfig config = TraceGenerator.getTraceConfig(listener.result, transition);
		JavaMethodArgumentsResolver resolver = new JavaMethodArgumentsResolver();
		return resolver.resolveMethodCallTraceTree(config, listener.result.getPDS())[1].getMethodCall();
	}

	public void testGetTransitions() throws Exception 
	{
		CoverageTester tester = getTester(TestTransitionsClass.class, "test", "(I)V");
		InputValueClassCoverageListener listener = new InputValueClassCoverageListener();
		tester.addCoverageListener(listener);
		tester.getInfo().setCheckForIndexOutOfBounds(true);
		tester.getInfo().setCheckForNullPointerExceptions(true);
		
		CoverageMethod test= getMethod(tester.test(), "test");
		//assertTrue(test.isFullyCovered());
		Line[] lines = test.getLines();
		Line line = lines[lines.length - 2];
		assertEquals(1, line.getTransitions(State.ASSERTION_FAILED).size());
		assertEquals(1, line.getTransitions(State.INDEX_OUT_OF_BOUNDS).size());
		assertEquals(2, line.getTransitions(State.NPE).size());
//		assertEquals(1, line.getTransitions(State.HEAP_OVERFLOW).size());
		
		MethodCall call = getMethodCall(line.getTransitions(State.ASSERTION_FAILED).get(0), 
				listener);
		assertEquals(3, call.getArguments()[0].getVariable().intValue());
		
		call = getMethodCall(line.getTransitions(State.INDEX_OUT_OF_BOUNDS).get(0), 
				listener);
		assertEquals(0, call.getArguments()[0].getVariable().intValue());
		
		call = getMethodCall(line.getTransitions(State.NPE).get(0), 
				listener);
		// Remy: here was 7
		assertEquals(4, call.getArguments()[0].getVariable().intValue());
		
		call = getMethodCall(line.getTransitions(State.NPE).get(1), 
				listener);
		assertEquals(1, call.getArguments()[0].getVariable().intValue());
		
//		call = getMethodCall(line.getTransitions(State.HEAP_OVERFLOW).get(0), 
//				listener);
//		assertEquals(3, call.getArguments()[0].getVariable().intValue());
	}
	
	public static class TestTransitionsClass
	{
		boolean n;
		
		public TestTransitionsClass(boolean n)
		{
			this.n = n;
		}
		
		public boolean nop()
		{
			return n;
		}
		
		public static void test(int i)
		{
			TestTransitionsClass[] instance = null;
			if (i == 0) {
				instance = new TestTransitionsClass[1];
			}
			else if (i == 1) {
				instance = new TestTransitionsClass[2];
			}
			else if (i == 2) {
				instance = new TestTransitionsClass[2];
				instance[1] = new TestTransitionsClass(true);
			}
			else if (i == 3) {
				instance = new TestTransitionsClass[2];
				instance[1] = new TestTransitionsClass(false);
			}
			assert(instance[1].nop()); 
			instance = new TestTransitionsClass[5];
		}
	}
}
