package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.Arrays;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;

public class MulticastCoverageListenerTest extends PDSTestCase 
{

	private MulticastCoverageListener multicaster;
	
	@Override
	protected void setUp() throws Exception 
	{
		multicaster = new MulticastCoverageListener();
	}
	
	public void testCoverageStarted()
	{
		MulticastTestListener listener1 = new MulticastTestListener();
		multicaster.addCoverageListener(listener1);
		multicaster.coverageStarted(null);
		assertTrue(listener1.coverageStarted);
		
		listener1.coverageStarted = false;
		MulticastTestListener listener2 = new MulticastTestListener();
		multicaster.addCoverageListener(listener2);
		
		multicaster.coverageStarted(null);
		assertTrue(listener1.coverageStarted);
		assertTrue(listener2.coverageStarted);
		
		multicaster.removeCoverageListener(listener1);
		assertFalse(Arrays.asList(multicaster.getListeners()).contains(listener1));
		assertTrue(Arrays.asList(multicaster.getListeners()).contains(listener2));
		multicaster.removeCoverageListener(listener2);
		assertFalse(Arrays.asList(multicaster.getListeners()).contains(listener2));
	}
	
	public void testCoverageFailed()
	{
		MulticastTestListener listener1 = new MulticastTestListener();
		multicaster.addCoverageListener(listener1);
		multicaster.coverageFailed("", null);
		assertTrue(listener1.coverageFailed);
		
		listener1.coverageFailed = false;
		MulticastTestListener listener2 = new MulticastTestListener();
		multicaster.addCoverageListener(listener2);
		
		multicaster.coverageFailed("", null);
		assertTrue(listener1.coverageFailed);
		assertTrue(listener2.coverageFailed);
	}
	
	private static class MulticastTestListener extends CoverageAdapter
	{

		public boolean coverageStarted;
		public boolean coverageFailed;
		
		@Override
		public void coverageStarted(PDS pds) 
		{
			coverageStarted = true;
		}

		public void coveredMethodOffset(CoverageMethod method, int lineOffset, 
				State state) 
		{
		}

		public void assertErrorReachedFrom(CoverageMethod method, int lineOffset) 
		{
					
		}

		public void coverageEnded(CoverageMethod[] methods) 
		{
					
		}

		public void coverageFailed(String message, Throwable t) 
		{
			coverageFailed = true;			
		}

		public void heapOverflow(CoverageMethod method, int lineOffset) {
			// TODO Auto-generated method stub
			
		}

		public void labelReached(String label) {
			// TODO Auto-generated method stub
			
		}
		
	}
	
}
