package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.awt.event.ComponentAdapter;
import java.util.List;
import java.util.concurrent.Delayed;

import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.InterfaceImplementerClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.SubSubTestClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.SubTestClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.TestClass;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.UndefTest.AssertionCountListener;

/**
 * Contains the PDSInst tests that need the coverage package for testing.
 */
public class PDSInstCoverageTest extends PDSTestCase 
{

	public void testMultinewarrayInst() throws Exception
	{
		CoverageTester tester = new CoverageTester(MultiNewArrayClass.class);
		AssertionCountListener listener = new AssertionCountListener();
		tester.addCoverageListener(listener);

		assertTrue(getMethod(tester.test(), "main").isFullyCovered());
		assertEquals(0, listener.asserts.size());
	}
	
	public static class MultiNewArrayClass
	{
		public static void main(String[] args) 
		{
			int[][] array = new int[1][2];
			array[0][0] = 5;
			array[0][1] = 3;
			assert(array[0][0] == 5);
			assert(array[0][1] == 3);
		}
	}
	
	public void testIndexOutOfBounds() throws Exception
	{
		CoverageTester tester = getTester(IndexOutOfBoundsClass.class, "store", "()V");
		IndexOutOfBoundsListener listener = new IndexOutOfBoundsListener();
		tester.addCoverageListener(listener);
		
		tester.getInfo().setCheckForIndexOutOfBounds(false);
		tester.test();
		assertFalse(listener.ioob);
		
		tester.getInfo().setCheckForIndexOutOfBounds(true);
		tester.test();
		assertTrue(listener.ioob);
	}
	
	public static class IndexOutOfBoundsClass
	{
		public static void store()
		{
			int[] array = new int[2];
			array[3] = 0;
		}
		
		public static void load()
		{
			int[] array = new int[2];
			int i = array[3];
		}
	}
	
	public static class IndexOutOfBoundsListener extends CoverageAdapter
	{
		boolean ioob = false;
		
		@Override
		public void coveredMethodOffset(CoverageMethod method, int lineOffset, 
				int pcOffset, State state, Transition transition) 
		{
			if (state == State.INDEX_OUT_OF_BOUNDS) {
				ioob = true;
			}
		}
	}
	
	public void testInstanceofInst() throws Exception
	{
		CoverageTester tester = new CoverageTester(InstanceOfClass.class);
		CoverageMethod main = getMethod(tester.test(), "main");
		assertTrue(main.isFullyCovered());
	}
	
	public static class InstanceOfClass 
	{
		public static void main(String[] args) 
		{
			SubSubTestClass instance = new SubSubTestClass();
			if (instance instanceof SubTestClass) {
				if (instance instanceof TestClass) {
					if (!(instance instanceof Comparable)) {
						InterfaceImplementerClass impl = new InterfaceImplementerClass();
						if (impl instanceof Comparable) {
							Delayed delayed = impl;
							if (delayed instanceof Comparable) {
								if (impl instanceof ComponentAdapter) {
									if (!(impl instanceof Iterable)) {
										// assert is only reached if all instanceof tests work correctly
										//assert(false);
									}
								}
							}
						}
					}
				}
			}
		}
	}
	
	public void testGetSetFieldInst() throws Exception
	{
		CoverageTester tester = new CoverageTester(ConcretePublicFieldClass.class);
		AssertionCountListener listener = new AssertionCountListener();
		tester.addCoverageListener(listener);
		// Remy: this function is not used any more
		tester.getInfo().setAbortAfterFailedAssertion(true);
		CoverageMethod main = getMethod(tester.test(), "main");
		assertTrue(main.isFullyCovered());
		assertTrue(listener.asserts.isEmpty());
	}
	
	public static abstract class AbstractPublicFieldClass
	{
		public int field;
	}
	
	public static class ConcretePublicFieldClass extends AbstractPublicFieldClass
	{
		public int field;
		
		public static void main(String[] args) 
		{
			ConcretePublicFieldClass instance = new ConcretePublicFieldClass();
			instance.field = 1;
			AbstractPublicFieldClass sup = instance;
			sup.field = 2;
			//assert(instance.field == 1);
			assert(sup.field == 2);
		}
	}
	
	public void testIntegerMaxMin() throws Exception
	{
		CoverageTester tester = new CoverageTester(MaxMinClass.class);
		AssertionCountListener listener = new AssertionCountListener();
		tester.addCoverageListener(listener);
		tester.test();
		assertEquals(0, listener.asserts.size());
	}
	
	public static class MaxMinClass
	{
		public static void main(String[] args) 
		{
			int i = 7;
			assert(Integer.MAX_VALUE == i);
			i = 0;
			assert(Integer.MIN_VALUE == i);
		}
	}
	
	/**
	 * In the current implementation, we are content when the assertion error was
	 * reached once. Not all possible errors are determined at once.
	 * @throws Exception
	 */
	public void testAssertionReachedMultipleTimes() throws Exception
	{
		CoverageTester tester = getTester(MultiAssertionClass.class, "test", "(I)V");
		CoverageMethod method = getMethod(tester.test(), "test");
		List<Transition> assertions = method.getLines()[0].getTransitions(State.ASSERTION_FAILED);
		assertEquals(1, assertions.size());
	}
	
	public static class MultiAssertionClass
	{
		public static void test(int i)
		{
			assert(i == 0);
			assert(i == 2);
			assert(i == 3);
			assert( i == 4);
			assert(i == 5);
			assert(i == 6);
			assert(i == 7);
		}
	}
}
