package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.awt.event.ComponentAdapter;
import java.util.concurrent.Delayed;

import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.InterfaceImplementerClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.SubSubTestClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.SubTestClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfoTest.TestClass;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageTesterTest.BubbleSort;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.PDSInstCoverageTest.InstanceOfClass;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.PreStarReachabilityTesterTest.CoverageCountListener;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.UndefTest.AssertionCountListener;

public class PostStarReachabilityTesterTest extends PDSTestCase 
{

	public void testPostStarReachability() throws Exception
	{
		CoverageTester tester = new CoverageTester(BubbleSort.class);
		tester.getInfo().setBits(2);
		CoverageMethod[] methods = tester.test();
		
		PostStarReachabilityTester psTester = new PostStarReachabilityTester(tester.getPDS());
		CoverageCountListener listener = new CoverageCountListener();
		psTester.addCoverageListener(listener);
		
		for (CoverageMethod method : methods) {
			for (AnnotatedInstruction inst : method.getAnnotatedInstructions()) {
				
				if (inst.isCovered()) {
					
					System.out.printf("Testing covered inst: %s%n", inst);
					listener.count = 0;
					psTester.test(inst.getLabel());
					assertTrue(listener.count > 0);
				}
				else {
					
					System.out.printf("Testing uncovered inst: %s%n", inst);
					listener.count = 0;
					psTester.test(inst.getLabel());
					assertEquals(10, listener.count);
				}
			}
		}
	}
	
	public void testInstanceOfInst() throws Exception
	{
		PostStarReachabilityTester tester = 
			new PostStarReachabilityTester(InstanceOfClass.class);
		AssertionCountListener listener = new AssertionCountListener();
		tester.addCoverageListener(listener);
		tester.test("de_uni_stuttgart_fmi_szs_jmoped_PDSInstTest_" +
				"InstanceOfClass_main_ALjava_lang_String_V64");
		
		assertEquals(1, listener.asserts.size());
	}
	
	private static class InstanceOfClass 
	{
		public static void main(String[] args) 
		{
			SubSubTestClass instance = new SubSubTestClass();
			if (instance instanceof SubTestClass) {
				if (instance instanceof TestClass) {
					if (!(instance instanceof Comparable)) {
						InterfaceImplementerClass impl = new InterfaceImplementerClass();
						if (impl instanceof Comparable) {
							Delayed delayed = impl;
							if (delayed instanceof Comparable) {
								if (impl instanceof ComponentAdapter) {
									if (!(impl instanceof Iterable)) {
										// assert is only reached if all instanceof tests work correctly
										assert(false);
									}
								}
							}
						}
					}
				}
			}
		}
	}
	
}
