package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageTesterTest.TestProgressMonitor;
import junit.framework.TestCase;

public class PreStarReachabilityTesterTest extends TestCase
{

	public void testPreStarReachability() throws Exception
	{
		PreStarReachabilityTester tester = new PreStarReachabilityTester
			(PreStarTestClass.class, new String[0]);
		
		String mainMethodLabel = "de_uni_stuttgart_fmi_szs_jmoped_coverage_" +
				"PreStarReachabilityTesterTest_PreStarTestClass_" +
				"main_ALjava_lang_String_V";
		
		CoverageCountListener listener = new CoverageCountListener();
		tester.addCoverageListener(listener);
		tester.test(mainMethodLabel);
		assertEquals(1, listener.count);
		listener.count = 0;
		
		String[] labels = new String[4];
		for (int i = 1; i <= 4; i++) {
			labels[i-1] = mainMethodLabel + i;
		}
		
		tester.test(labels);
		assertEquals(5, listener.count);
		listener.count = 0;
		
		System.out.println("Two-bit testing ..");
		tester.getInfo().setBits(2);
		System.out.println("testing label 7");
		tester.test(mainMethodLabel + 7);
		System.out.println("testing label 10");
		tester.test(mainMethodLabel + 10);
		System.out.println("testing label 11");
		tester.test(mainMethodLabel + 11);
		assertEquals(6, listener.count);
		
		listener.count = 0;
		System.out.println("Four-bit testing ..");
		tester.getInfo().setBits(4);
		System.out.println("testing label 7");
		tester.test(mainMethodLabel + 7);
		System.out.println("testing label 10");
		tester.test(mainMethodLabel + 10);
		System.out.println("testing label 11");
		tester.test(mainMethodLabel + 11);
		assertEquals(24, listener.count);
	}
	
	public static class PreStarTestClass
	{
		public static void main(String[] args)
		{
			int i = 0;
			i = 1;
			if (i < 9) {
				i = 0;
			}
		}
	}
	
	public void testProgressMonitor() throws Exception
	{
		PreStarReachabilityTester tester = new PreStarReachabilityTester
		(PreStarTestClass.class, new String[0]);
		
		String mainMethodLabel = "de_uni_stuttgart_fmi_szs_jmoped_coverage_" +
				"PreStarReachabilityTesterTest_PreStarTestClass_" +
				"main_ALjava_lang_String_V";
		
		CoverageCountListener listener = new CoverageCountListener();
		tester.addCoverageListener(listener);
		
		tester.test(mainMethodLabel);
		assertEquals(1, listener.count);
		listener.count = 0;
		
		TestProgressMonitor monitor = new TestProgressMonitor();
		monitor.setCanceled(true);
		
		tester.test(mainMethodLabel, monitor);
		assertEquals(0, listener.count);
		
		for (int i = 1; i <= 4; i++) {
			tester.test(mainMethodLabel + i, monitor);
		}
		assertEquals(0, listener.count);
	}
	

	public void testTestClass() throws Exception
	{
		PreStarReachabilityTester tester = new PreStarReachabilityTester
		(TestClass.class, new String[0]);
		CoverageCountListener listener = new CoverageCountListener();
		tester.addCoverageListener(listener);
		
		String mainLabel = "de_uni_stuttgart_fmi_szs_jmoped_coverage_" +
				"PreStarReachabilityTesterTest_TestClass_" +
				"main_ALjava_lang_String_V";
		String[] labels = new String[] {
				mainLabel + "56",				
				mainLabel + "57",
				mainLabel + "59",
		};
		
		System.out.println("Default-bit testing ..");
		tester.test(labels);
		assertEquals(1, listener.count);
		
		System.out.println("Four-bit testing ..");
		listener.count = 0;
		tester.getInfo().setBits(4);
		tester.test(labels);
		assertEquals(9, listener.count);
	}
	
	public static class TestClass
	{
		public static void main(String[] args)
		{
			int i = 1;
			assert(i == 1);
			if (args.length == 0) {
				i = 0;
			}
			else {
				i = 1;
			}
			if (true) {
				i = 6;
			}
			assert(i != 6);
			if (i < 9) {
				i = 3;
			}
			if (false) {
				i = 5;
			}
		}
	}

	
	public static class CoverageCountListener extends CoverageAdapter
	{
		public int count = 0;
		
		@Override
		public void coveredMethodOffset(CoverageMethod method, int lineOffset, 
				int pcOffset, State state, Transition transition) 
		{
			System.out.printf("name: %s, pcOffset: %d, lineOffset: %d, state: %s%n", 
					method.getMethod().getName(), pcOffset, lineOffset, state);
			// mimicking the old behavior
			if (state == State.COVERED || state == State.PARTIALLY_COVERED) {
				count++;
			}
		}
	}
}
