package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.Timer;
import java.util.TimerTask;

import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;
import de.uni_stuttgart.fmi.szs.jmoped.annotation.PDSArrayBits;
import de.uni_stuttgart.fmi.szs.jmoped.annotation.PDSParameterBits;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageTesterTest.BubbleSort;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageTesterTest.QuickSort;
import de.uni_stuttgart.fmi.szs.jmoped.data.TestChange;

public class ProgressMonitorTest extends PDSTestCase 
{

	private Timer timer = new Timer();
	
	public void testDummy()
	{
		
	}
	
	public void testPerformance() throws Exception
	{
		CoverageTester tester = getTester(ArrayUtilsClass.class, "verifyReversal", "([I)V");
		tester.getInfo().setHeapSize(9);
		long time = System.currentTimeMillis();
		tester.test();
		System.out.println((System.currentTimeMillis() - time) / 1000);
	}
	
	public void ftestPerformance2() throws Exception
	{
		CoverageTester tester = getTester(QuickSort.class, "sort", "([I)V");
		tester.test();
	}
	
	public static class ArrayUtilsClass
	{

	    @PDSParameterBits(2)
	    @PDSArrayBits(1)
	    public static void verifyReversal(int[] array)
		{
			int[] copy = new int[array.length];
			for (int i = 0; i < copy.length; i++) {
				copy[i] = array[i];
			}
			reverse(array);
			for (int i = 0; i < copy.length; i++) {
				assert(copy[i] == array[copy.length - 1 - i]);
			}
		}
	    
	    /**
	     * <p>Reverses the order of the given array.</p>
	     * 
	     * <p>This method does nothing if <code>null</code> array input.</p>
	     * 
	     * @param array  the array to reverse, may be <code>null</code>
	     */
	    public static void reverse(int[] array) {
	        if (array == null) {
	            return;
	        }
	        int i = 0;
	        //int i = 1;
	        int j = array.length - 1;
	        int tmp;
	        while (j > i) {
	            tmp = array[j];
	            array[j] = array[i];
	            array[i] = tmp;
	            j--;
	            i++;
	        }
	    }
	}
	
	public void ftestCoverageTesterResponsiveness() throws Exception
	{
		CoverageTester tester = new CoverageTester(QuickSort.class);
		testResponsiveness(tester);
		
		tester = new CoverageTester(BubbleSort.class);
		tester.getInfo().setHeapSize(31);
//		tester.getInfo().setBits(7);
//		testResponsiveness(tester, 100);
//		testResponsiveness(tester, 5000);
		
		tester = new CoverageTester(QuickSort.class, "sort", "([III)V");
//		tester.getInfo().setBits(2);
//		tester.getInfo().setHeapSize(31);
//		testResponsiveness(tester, 200);
		
		tester = new CoverageTester(TestChange.class);
		tester.getInfo().setHeapSize(31);
		Coverage.setMaximumMemory(256 * 1024 * 1024);
		testResponsiveness(tester, 200);
	}
	
	private void testResponsiveness(CoverageTester tester) throws Exception
	{
		testResponsiveness(tester, 1000);
	}
	
	private void testResponsiveness(CoverageTester tester, int increment) 
		throws Exception
	{
		long accumulatedResponseTimes = 0;
		int runs = 0;
		ResponsivenessMonitor monitor = new ResponsivenessMonitor();
		// set to true so we can enter the loop
		monitor.canceled = true;
		
		for (int millis = 0; monitor.canceled; millis += increment) {
			monitor.setTimeout(millis);
			timer.schedule(new CancelMonitorTask(monitor), millis);
			tester.test(monitor);
			if (monitor.canceled) {
				accumulatedResponseTimes += monitor.getResponseTimeMillis();
				System.out.println("Response Time " + (float)monitor.getResponseTimeMillis() / 1000);
				++runs;
			}
		}
		monitor.doneTime = System.currentTimeMillis();
		System.out.println("Runs " + runs );
		System.out.println("Average Response Time " 
				+ (double)accumulatedResponseTimes / (1000 * runs));
		System.out.println("Last Not Response Time " 
				+ (double)monitor.getResponseTimeMillis() / 1000);	
	}
	
	public static class CancelMonitorTask extends TimerTask
	{
		
		public ProgressMonitor monitor;
		
		public CancelMonitorTask(ProgressMonitor monitor) 
		{
			this.monitor = monitor;
		}

		@Override
		public void run() 
		{
			monitor.setCanceled(true);
		}
		
	}
	
	public static class ResponsivenessMonitor implements ProgressMonitor
	{

		public long startTime = 0;
		public long timeoutMillis = 0;
		public long canceledStartTime = 0;
		public long doneTime = 0;
		public boolean canceled;
		
		public void setTimeout(long millis)
		{
			timeoutMillis = millis;
			canceledStartTime = 0;
			doneTime = 0;
			startTime = System.currentTimeMillis();
			canceled = false;
		}
		
		public long getResponseTimeMillis()
		{
			return canceled ? doneTime - canceledStartTime : doneTime - startTime + timeoutMillis;
		}
		
		public void beginTask(String description, int workAmount) 
		{

		}

		public void done() 
		{
			doneTime = System.currentTimeMillis();
		}

		public boolean isCanceled() 
		{
			return canceled;
		}

		public void setCanceled(boolean cancel) 
		{
			canceledStartTime = System.currentTimeMillis();
			canceled = cancel;
		}

		public void subTask(String description) {
			// TODO Auto-generated method stub
			
		}

		public void worked(int amount) {
			// TODO Auto-generated method stub
			
		}
		
	}
	
}
