package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.IntVariable.Type;

public class TraceGeneratorTest extends PDSTestCase 
{

	public void testGetInputValuesForInitialMethod() throws Exception
	{
		CoverageTester tester = getTester(InputValueClass.class, "assertTrace", "(I)V");
		InputValueClassCoverageListener listener = new InputValueClassCoverageListener();
		tester.addCoverageListener(listener);
		
		CoverageMethod method = getMethod(tester.test(), "assertTrace");
		assertNotNull(method);
		assertNotNull(listener.result);
				
		for (int i = 0; i < listener.transitions.length; i++) {
			IntConfig prefixConfig = TraceGenerator.getTraceConfig(listener.result,
					listener.transitions[i]);
			IntVariable[] vars = prefixConfig.getConfigAt(method.getMethod().getFormattedName()).getStackVariables();
			IntVariable v1 = getVariable(vars, "v0");
			assertEquals(i, v1.intValue());
		}
	}
	
	public void testGetArrayValuesForInitialMethod() throws Exception
	{
		CoverageTester tester = getTester(InputValueClass.class, "assertArrayTrace", "([I)V");
		InputValueClassCoverageListener listener = new InputValueClassCoverageListener();
		tester.addCoverageListener(listener);
		
		CoverageMethod method = getMethod(tester.test(), "assertArrayTrace");
		assertNotNull(method);
		assertNotNull(listener.result);
		assertNotNull(listener.transitions[0]);
		
		IntConfig prefixConfig = TraceGenerator.getTraceConfig(listener.result,
				listener.transitions[0]);
		
		IntConfig config = prefixConfig.getConfigAt(method.getMethod().getFormattedName());
		assertNotNull(config);
		
		IntVariable[] stackVars = config.getStackVariables();
		IntVariable[] globalVars = config.getGlobalVariables();
		
		print(stackVars);
		print(globalVars);
		
		IntVariable heapOffsetVar= getVariable(stackVars, "v0");
		assertNotNull(heapOffsetVar);
		int heapOffset = heapOffsetVar.intValue();
		IntVariable heapVar = getVariable(globalVars, tester.getInfo().getHeapName());
		assertNotNull(heapVar);
		int[] heap = heapVar.arrayValue();
		
		assertEquals(2, heap[heapOffset]);
		assertEquals(1, heap[heapOffset + 1]);
		assertEquals(3, heap[heapOffset + 2]);
	}
	
	
	
	private void print(IntVariable[] vars)
	{
		for (IntVariable var : vars) {
			if (var.getType() == Type.ARRAY) {
				System.out.println(var.getName() + ": " +  Arrays.toString(var.arrayValue()));
			}
			else if (var.getType() == Type.BOOLEAN) {
				System.out.println(var.getName() + ": " +  var.booleanValue());
			}
			else {
				System.out.println(var.getName() + ": " +  var.intValue());
			}
		}
	}
	
	private IntVariable getVariable(IntVariable[] vars, String name)
	{
		for (IntVariable var : vars) {
			if (var.getName().equals(name)) {
				return var;
			}
		}
		return null;
	}
	
	public static class InputValueClass
	{
		public static void assertTrace(int i)
		{
			assert(i != 0);
			assert(i != 1);
			assert(i != 2);
			assert(i != 3);
		}
		
		public static void assertArrayTrace(int[] i)
		{
			assert(i.length != 2 || i[0] != 1 || i[1] != 3);
		}
		
		public static void test(int kaa, Object o)
		{
			int v;
			Object t;
			
			v = kaa;
			t = o;
		}
		
	}
	
	public static class InputValueClassCoverageListener extends CoverageAdapter
	{
		CoverageResult result;
		
		Transition[] transitions = new Transition[4];
		
		Map<Integer, Transition> map = new HashMap<Integer, Transition>(); 
		
		Map<Integer, Transition> overflowMap = new HashMap<Integer, Transition>();
		
		CoverageMethod overflowMethod = null;
		
		int overFlowOffset;
		
		@Override
		public void coveredMethodOffset(CoverageMethod method, int lineOffset,
				int pcOffset, State state, Transition transition) 
		{
			System.out.println(method + " " + state + " " + lineOffset + " " + pcOffset);
			if (state == State.ASSERTION_FAILED) {
				assertErrorReachedFrom(method, lineOffset, transition);
			}
			else if (state == State.HEAP_OVERFLOW) {
				overflowMap.put(lineOffset, transition);
				overflowMethod = method;
				overFlowOffset = pcOffset;
			}
		}
		
		public void assertErrorReachedFrom(CoverageMethod method, int lineOffset, 
				Transition transition) 
		{
			System.out.println("lineOffset: " + lineOffset);
			if (lineOffset <= 4) {
				transitions[lineOffset] = transition;
			}
			else {
				map.put(lineOffset, transition);
			}
		}
		
		@Override
		public void coverageEnded(CoverageResult result) 
		{
			this.result = result;
		}
	}
	
}
