package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.TraceGeneratorTest.InputValueClassCoverageListener;

public class TraceMinimizerTest extends PDSTestCase 
{
	
	private CoverageResult getResult(Class clazz, String name, String descriptor) throws Exception
	{
		CoverageTester tester = getTester(clazz, name, descriptor);
		InputValueClassCoverageListener listener = new InputValueClassCoverageListener();
		tester.addCoverageListener(listener);
		CoverageMethod[] methods = tester.test();
		return listener.result;
	}
	
	public void testShouldReturnOneTransitions() throws Exception
	{
		CoverageResult result = getResult(SingleExecutionPathClass.class, "test", "()V");
		CoverageMethod method = getMethod(result.getMethods(), "test");
		TraceMinimizer minimizer = new TraceMinimizer(result, method);
		assertTrue(minimizer.isCovered());
		assertEquals(1, minimizer.getTransitions().length);
	}
	
	public static class SingleExecutionPathClass
	{
		public static void test()
		{
			int i = 5;
			int j = i - 3;
			i = 2 * j;
			if (i == 4) {
				j = 5;
			}
			i = 3;
		}
	}
	
	public void testShouldReturnTwoTransitions() throws Exception
	{
		CoverageResult result = getResult(TwoPathsClass.class, "test", "(I)V");
		TraceMinimizer minimizer = new TraceMinimizer(result, getMethod(result.getMethods(), "test"));
		assertTrue(minimizer.isCovered());
		assertEquals(2, minimizer.getTransitions().length);
	}
	
	public void testWithBranchCoverageTester() throws Exception
	{
		BranchCoverageTester tester = new BranchCoverageTester(
				TwoPathsClass.class, "test", "(I)V");
		InputValueClassCoverageListener listener = new InputValueClassCoverageListener();
		tester.addCoverageListener(listener);
		tester.test();
		TraceMinimizer minimizer = new TraceMinimizer(listener.result, 
				getMethod(listener.result.getMethods(), "test"));
		assertEquals(2, minimizer.getTransitions().length);
		// covers as well but cannot ascertained
	}
	
	public static class TwoPathsClass
	{
		public void test(int i)
		{
			int j;
			if (i == 0) {
				j = 3;
			}
			else {
				j = 5;
			}
			i = j;
		}
	}
	
}
