package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.HashSet;

import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;

public class UndefTest extends PDSTestCase
{
	
	public void testUndef() throws Exception
	{
		CoverageTester tester = new CoverageTester(UndefTestClass.class);
		tester.getInfo().setBits(2);
		AssertionCountListener listener = new AssertionCountListener();
		tester.addCoverageListener(listener);
		tester.test();
		assertEquals(4, listener.asserts.size());
	}

	public static class UndefTestClass
	{
		public static void main(String[] args)
		{
			int i = Integer.parseInt("");
			assert(i != 0);
			assert(i != 1);
			assert(i != 2);
			assert(i != 3);
		}
	}
	
	/**
	 * Test that shows that if the heap is not initialized its values are undefined.
	 * 
	 * This test is commented out, since the implementation was fixed.
	 */
	public void ftestHeapUndef() throws Exception
	{
		CoverageTester tester = new CoverageTester(HeapUndefTestClass.class);
		tester.getInfo().setBits(3);
		AssertionCountListener listener = new AssertionCountListener();
		tester.addCoverageListener(listener);
		tester.test();
		assertEquals(4, listener.asserts.size());
	}
	
	public static  class HeapUndefTestClass 
	{
		public static void main(String[] args)
		{
			int[] array = new int[2];
			assert(array[0] != 0);
			assert(array[0] != 1);
			assert(array[0] != 2);
			assert(array[0] != 3);
		}		
	}
	
	public void testArrayValues()
	{
		int[] array = new int[2];
		assertEquals(0, array[0]);
		assertEquals(0, array[1]);
	}

	public static class AssertionCountListener extends CoverageAdapter
	{
		public HashSet<Integer> asserts = new HashSet<Integer>();

		@Override
		public void coveredMethodOffset(CoverageMethod method, int lineOffset, 
				int pcOffset, State state, Transition transition) 
		{
			System.out.printf("method: %s, lineOffset: %d, pcOffset: %d, state: %s, transition: %s%n",
					method, lineOffset, pcOffset, state, transition);
			if (state == State.ASSERTION_FAILED) {
				assertErrorReachedFrom(method, lineOffset);
			}
		}
		
		public void assertErrorReachedFrom(CoverageMethod method, int lineOffset) 
		{
			asserts.add(lineOffset);
		}
	}
	
}
