#include "common.h"
#include "IntConfig.h"

static intConfig* getConfigAtLabel(intConfig *config, const char *label)
{
  intConfig *c;
  for (c = config; c; c = c->next_config) {
    if (strcmp(wIdentString(c->stack->stack), label) == 0) {
      return c;
    }
  }
  return NULL;
}

static jint getLength(intValuation *valuation)
{
  jint length = 0;
  intValuation *current;
  for (current = valuation; current; current = current->next) {
    ++length;
  }
  return length;
}

static jintArray createIntArray(JNIEnv *env, int *array, int size)
{
  jintArray intArray = (*env)->NewIntArray(env, size);
  (*env)->SetIntArrayRegion(env, intArray, 0, size, array);
  return intArray;
}

static int getStackLength(intConfig *config)
{
  int length = 0;
  intPath *stack;
  for (stack = config->stack; stack; stack = stack->rest_path) {
    ++length;
  }
  return length;
}

static void setVariable(JNIEnv *env, intValuation *valuation, 
			jobjectArray array, jsize index, jclass clazz)
{
  intVariable *variable = valuation->var;
  jstring name = (*env)->NewStringUTF(env, wIdentString(variable->name));
  int type = variable->type;
  int dimension = variable->dim;
  if (dimension == 0) {
    int value = valuation->value.integer;
    jmethodID constructor = 
      (*env)->GetMethodID(env, clazz, "<init>", "(Ljava/lang/String;III)V");
    jobject var;

    assert(constructor != NULL);

    var = (*env)->NewObject(env, clazz, constructor, name, type, dimension,
			    value);

    (*env)->SetObjectArrayElement(env, array, index, var);
    (*env)->DeleteLocalRef(env, var);
  }
  else {
    jintArray intArray = createIntArray(env, valuation->value.array, 
				     variable->dim);
    jmethodID constructor = 
      (*env)->GetMethodID(env, clazz, "<init>", "(Ljava/lang/String;II[I)V");
    jobject var;

    assert(constructor != NULL);
    
    var = (*env)->NewObject(env, clazz, constructor, name, type, dimension,
			    intArray);

    (*env)->SetObjectArrayElement(env, array, index, var);
    (*env)->DeleteLocalRef(env, var);
    (*env)->DeleteLocalRef(env, intArray);
  }
  (*env)->DeleteLocalRef(env, name);
}

static jobject createConfig(JNIEnv *env, jobject parent, intConfig *config)
{
  jclass clazz = (*env)->FindClass(env,
				   "de/uni_stuttgart/fmi/szs/jmoped/coverage/"
				   "IntConfig");
  jmethodID constructor = 
    (*env)->GetMethodID(env, clazz, "<init>",
			"(I"
			"Lde/uni_stuttgart/fmi/szs/jmoped/coverage/IntConfig;)"
			"V");
  jobject object;

  assert(config != NULL);
  assert(constructor != NULL);

  object = (*env)->NewObject(env, clazz, constructor, (int)config, parent);

  (*env)->DeleteLocalRef(env, clazz);

  return object;
}

JNIEXPORT jobjectArray JNICALL
Java_de_uni_1stuttgart_fmi_szs_jmoped_coverage_IntConfig_getStackVariablesNative
(JNIEnv *env, jobject jconfig)
{
  intConfig *config = (intConfig*)getHandlePointer(env, jconfig, "handle");
  int length = getLength(config->stack->values);
  intValuation *valuation;
  jclass clazz = (*env)->FindClass(env, 
				   "de/uni_stuttgart/fmi/szs/jmoped/coverage/"
				   "IntVariable");
  jobjectArray variables = (*env)->NewObjectArray(env, length, clazz, NULL);
  jsize index;
  
  collectGarbage(env);

  for (index = 0, valuation = config->stack->values; valuation;
       valuation = valuation->next, ++index) {
    setVariable(env, valuation, variables, index, clazz);
  }

  (*env)->DeleteLocalRef(env, clazz);

  return variables;
}

JNIEXPORT jobjectArray JNICALL 
Java_de_uni_1stuttgart_fmi_szs_jmoped_coverage_IntConfig_getGlobalVariablesNative
(JNIEnv *env, jobject jconfig)
{
  intConfig *config = (intConfig*)getHandlePointer(env, jconfig, "handle");
  int length = getLength(config->values);
  intValuation *valuation;
  jclass clazz = (*env)->FindClass(env, 
				   "de/uni_stuttgart/fmi/szs/jmoped/coverage/"
				   "IntVariable");
  jobjectArray variables = (*env)->NewObjectArray(env, length, clazz, NULL);
  jsize index;
  
  collectGarbage(env);

  for (index = 0, valuation = config->values; valuation;
       valuation = valuation->next, ++index) {
    setVariable(env, valuation, variables, index, clazz);
  }

  (*env)->DeleteLocalRef(env, clazz);

  return variables;
}


JNIEXPORT jobject JNICALL
Java_de_uni_1stuttgart_fmi_szs_jmoped_coverage_IntConfig_getConfigAtNative
(JNIEnv *env, jobject jconfig, jstring jlabel)
{
  intConfig *config = (intConfig*)getHandlePointer(env, jconfig, "handle");
  char *label = getJavaString(env, jlabel);
  intConfig *result;

  assert(config != NULL);

  collectGarbage(env);

  result = getConfigAtLabel(config, label);
  Free(label);

  if (result) {
    return createConfig(env, jconfig, result);
  }
  else {
    return NULL;
  }
}

JNIEXPORT jstring JNICALL
Java_de_uni_1stuttgart_fmi_szs_jmoped_coverage_IntConfig_getStackSymbolNative
(JNIEnv *env, jobject jconfig)
{
  intConfig *config = (intConfig*)getHandlePointer(env, jconfig, "handle");

  collectGarbage(env);

  return (*env)->NewStringUTF(env, wIdentString(config->stack->stack));
}

JNIEXPORT jobject JNICALL 
Java_de_uni_1stuttgart_fmi_szs_jmoped_coverage_IntConfig_getNextConfigNative
(JNIEnv *env, jobject jconfig)
{
  intConfig *config = (intConfig*)getHandlePointer(env, jconfig, "handle");
  intConfig *next = config->next_config;

  collectGarbage(env);

  if (next) {
    return createConfig(env, jconfig, next);
  }
  else {
    return NULL;
  }
}

JNIEXPORT jint JNICALL
Java_de_uni_1stuttgart_fmi_szs_jmoped_coverage_IntConfig_getStackLengthNative
(JNIEnv *env, jobject jconfig)
{
 intConfig *config = (intConfig*)getHandlePointer(env, jconfig, "handle");

 collectGarbage(env);

 return getStackLength(config);
}
