#include "PostStarReachabilityTester.h"
#include "common.h"

/*typedef struct postReachabilityContext
{
  JNIEnv *env;
  jobject tester;
  progressMonitor *monitor;
  wHash* requiredStackElements;
  int unvisitedStackElements;
} postReachabilityContext;

postReachabilityContext* createPostReachabilityContext(JNIEnv *env, 
						       jobject tester,
						       progressMonitor *monitor,
						       wIdentList *labels)
{
  int count = 0;
  stackElement *elem;
  stackElement *found;

  postReachabilityContext *context = Malloc(sizeof(postReachabilityContext));
  context->env = env;
  context->tester = tester;
  context->monitor = monitor;
  context->requiredStackElements = createStackElementHash();
  
  for (; labels; labels = labels->next) {
    elem = Malloc(sizeof(stackElement));
    elem->hash = elem->stack = labels->wId;
    found = wHashInsert(context->requiredStackElements, elem);
    if (found == elem) {
      ++count;
    }
    else {
      Free(elem);
    }
  }
  context->unvisitedStackElements = count;
  return context;
}

void freePostReachabilityContext(postReachabilityContext *context)
{
  freeStackElementHash(context->requiredStackElements);
  Free(context);
}

stackElement *getRequired(wIdent label, wHash *requiredStackElements)
{
  stackElement elem = { .hash = label, .stack = label };
  return wHashFind(requiredStackElements, &elem);
}*/

int isPostStarReachabilitySolved(wTrans *t, bddProblem *problem) {

	callbackContext *callback = (callbackContext*) problem->user;
	coverageContext *context = callback->context;

	if (IS_CANCELED(callback->monitor)) return 1;

	if (t->from == problem->state
			&& !bddEmpty(t->label)
			&& !isMarkedAsReached(context, t->name)) {

		if (matchPrefix(t->name, context->stackPrefix)) {

			WORKED(1, callback->monitor);
			markAsReached(context, t->name);

			JNIEnv *env = callback->env;
			const char *label = wIdentString(t->name);
			fireCovered(env, callback->tester, label,
				t->from, t->name,
				t->to, callback->fireCovered);
		}
	}

	//printf("%s -%s-> %s\n", wIdentString(t->from), wIdentString(t->name), wIdentString(t->to)); fflush(stdout);
	return context->unvisitedStackElements == 0
		|| (t->from == problem->state && t->name == problem->stack && !bddEmpty(t->label));

/*  postReachabilityContext *context = (postReachabilityContext*) problem->user;

  if (IS_CANCELED(context->monitor)) {
    return 1;
  }
  else if (transition->from == problem->state && !bddEmpty(transition->label)) {
    stackElement *elem = getRequired(transition->name, 
				     context->requiredStackElements);
    if (elem) {
      wHashRemove(context->requiredStackElements, elem);
      WORKED(1, context->monitor);
      fireCoveredLabelWithTransition(context->env, context->tester,
				     wIdentString(transition->name),
				     transition->from, transition->name, 
				     transition->to);
      --(context->unvisitedStackElements);
    }
  }
  return context->unvisitedStackElements == 0;*/
}

extern wIdent default_state_name;

/**
 * Creates a callback problem for post* analysis.
 * TODO: labels is a list, but we use only the first element in the list
 */
intProblem* createPostProblem(JNIEnv *env, jobject analyzer, intPds *ipds,
		const char *stackPrefix, wIdentList *labels, 
		progressMonitor *monitor) {

	intProblem *problem = Malloc(sizeof(intProblem));
	problem->type = PROBLEM_REACHABILITY;
	problem->state = default_state_name;
	problem->stack = labels->wId;
	problem->isSolved = &isPostStarReachabilitySolved;
	problem->user = createCallbackContext(env, analyzer, ipds, stackPrefix, monitor);
	problem->queue = NULL; 
	//problem->queue = createPreferringNewTransitionsQueueStrategy();
	//problem->queue = createFIFOPreferringNewTransitionsQueueStrategy();
	//problem->queue = createFIFOQueueStratey();
	//problem->queue = createLIFOQueueStratey();

	return problem;
}

JNIEXPORT jobject JNICALL 
Java_de_uni_1stuttgart_fmi_szs_jmoped_coverage_PostStarReachabilityTester_analyseReachability
(JNIEnv *env, jobject analyzer, jstring fName, jobjectArray jLabels,
 jobject monitor) {

	progressMonitor *javaMonitor = createJavaProgressMonitor(env, monitor);
	BEGIN_TASK("Starting coverage analysis", 10000, javaMonitor);

	progressMonitor *readMonitor = CREATE_SUB_MONITOR(3000, javaMonitor);
	const char *fileName = (*env)->GetStringUTFChars(env, fName, 0);
	intPds *ipds = readPds(fileName, readMonitor);
	(*env)->ReleaseStringUTFChars(env, fName, fileName);
	freeSubProgressMonitor(readMonitor);

	if (!ipds) {
		freeJavaProgressMonitor(javaMonitor);
		return NULL;
	}

	collectGarbage(env);

	progressMonitor *analyseMonitor = CREATE_SUB_MONITOR(7000, javaMonitor);

	wIdentList *labels = extractLabels(env, jLabels);
	intProblem *iproblem = createPostProblem(env, analyzer, ipds, NULL, labels, analyseMonitor);

	jobject result = jniAnalyse(env, analyzer, ipds, iproblem, NULL, analyseMonitor, NULL);
  
	freeIntProblem(iproblem); 
	freeSubProgressMonitor(analyseMonitor);
	freeJavaProgressMonitor(javaMonitor);
	freeWIdentList(labels);

	return result;
}
