#include "PreStarReachabilityTester.h"
#include "common.h"

int isPreStarReachabilitySolved(wTrans *transition, bddProblem *problem) {

	callbackContext *callback = (callbackContext*) problem->user;
	coverageContext *context = callback->context;

	if (IS_CANCELED(callback->monitor)) return 1;

	if (transition->from == problem->state
			&& wIsFinal(transition->to)
			&& !bddEmpty(transition->label)
			&& !isMarkedAsReached(context, transition->name)) {

		if (matchPrefix(transition->name, context->stackPrefix)) {

			WORKED(1, callback->monitor);
			markAsReached(context, transition->name);

			JNIEnv *env = callback->env;
			const char *label = wIdentString(transition->name);
			fireCovered(env, callback->tester, label, 
				transition->from, transition->name,
				transition->to, callback->fireCovered);
		}
	}

	return context->unvisitedStackElements == 0;
  
  // TODO simple removal won't do later on, they will have all the same initial
  // values

/*   if (transition->from == context->state && transition->name == context->stack */
/*       && transition->to == context->finalstate && !bddEmpty(t->label)) { */
/*     fireCoveredLabelWithTransition(context->env, context->tester,  */
/* 				   wIdentString(transition->name),  */
/* 				   transition->from, transition->name, */
/* 				   transition->to); */
/*     WORKED(1, context->monitor); */
/*   } */

/*   if (!bddEmpty(transition->label) && lengthSymList(*(context->labels)) == 1) { */
/*     if (*(context->labels) && in(transition->name, *(context->labels))) { */
/*       removeIdent(transition->name, context->labels); */
/*       fireCoveredLabel(context->env, context->tester, */
/* 		       wIdentString(transition->name)); */
/*     } */
/*   } */
  
  //return IS_CANCELED(callback->monitor);
}

extern wIdent default_state_name;

/**
 * Creates a callback problem for pre* analysis.
 * TODO: labels is a list, but we use only the first element in the list
 */
intProblem* createPreProblem(JNIEnv *env, jobject analyzer, intPds *ipds,
		const char *stackPrefix, wIdentList *labels, 
		progressMonitor *monitor) {

	intProblem *problem = Malloc(sizeof(intProblem));
	problem->type = PROBLEM_REACHABILITY_PRE_2;
	problem->state = default_state_name;
	problem->stack = labels->wId;
	problem->isSolved = &isPreStarReachabilitySolved;
	problem->user = createCallbackContext(env, analyzer, ipds, stackPrefix, monitor);
	problem->queue = NULL; 
	//problem->queue = createPreferringNewTransitionsQueueStrategy();
	//problem->queue = createFIFOPreferringNewTransitionsQueueStrategy();
	//problem->queue = createFIFOQueueStratey();
	//problem->queue = createLIFOQueueStratey();

	return problem;
}

JNIEXPORT jobject JNICALL 
Java_de_uni_1stuttgart_fmi_szs_jmoped_coverage_PreStarReachabilityTester_analyseReachability
(JNIEnv *env, jobject analyzer, jstring fName, jobjectArray jLabels, jobject monitor) {

	progressMonitor *javaMonitor = createJavaProgressMonitor(env, monitor);
	BEGIN_TASK("Starting coverage analysis", 10000, javaMonitor);

	progressMonitor *readMonitor = CREATE_SUB_MONITOR(3000, javaMonitor);
	const char *fileName = (*env)->GetStringUTFChars(env, fName, 0);
	intPds *ipds = readPds(fileName, readMonitor);
	(*env)->ReleaseStringUTFChars(env, fName, fileName);
	freeSubProgressMonitor(readMonitor);

	if (!ipds) {
		freeJavaProgressMonitor(javaMonitor);
		return NULL;
	}

	collectGarbage(env);

	progressMonitor *analyseMonitor = CREATE_SUB_MONITOR(7000, javaMonitor);

	wIdentList *labels = extractLabels(env, jLabels);
	intProblem *iproblem = createPreProblem(env, analyzer, ipds, NULL, labels, analyseMonitor);

	jobject result = jniAnalyse(env, analyzer, ipds, iproblem, NULL, analyseMonitor, NULL);
  
	freeIntProblem(iproblem); 
	freeSubProgressMonitor(analyseMonitor);
	freeJavaProgressMonitor(javaMonitor);
	freeWIdentList(labels);

	return result;

/*  const char *fileName;
  const char *label;
  intPds *ipds;
  intProblem *problem = Malloc(sizeof(intProblem));
  intResult *result;
  wIdentList *labels = NULL;
  //reachabilityContext *context;

  progressMonitor *javaMonitor = createJavaProgressMonitor(env, monitor);
  javaMonitor->beginTask("Starting prestar analysis", 10000, javaMonitor);
  progressMonitor *readMonitor = createSubProgressMonitor(3000, javaMonitor);

  collectGarbage(env);

  // read pds and return if monitor was canceled in the meantime
  fileName = (*env)->GetStringUTFChars(env, fName, 0);
  ipds = readPds(fileName, readMonitor);
  (*env)->ReleaseStringUTFChars(env, fName, fileName);
  freeSubProgressMonitor(readMonitor);

  if (!ipds) {
    freeJavaProgressMonitor(javaMonitor);
    return NULL;
  }

  progressMonitor *analyseMonitor = createSubProgressMonitor(7000, javaMonitor);
  callbackContext *callback = createCallbackContext(env, tester, ipds, NULL, analyseMonitor);

  // create problem
  problem->type = PROBLEM_REACHABILITY_PRE;
  problem->isSolved = &isPreStarReachabilitySolved;
  problem->state = default_state_name;
  problem->user = callback;
  //context = createReachabilityContext(env, tester, analyseMonitor,
//					   &labels);
  
  // set pre star values to context
  //context->state = ipds->pds->start->state;
  //context->stack = *(ipds->pds->start->stack);
  //context->finalstate = bddFinalState();

  //problem.user = context;
  problem->queue = NULL; //createFIFOPreferringNewTransitionsQueueStrategy();

  labels = extractLabels(env, jLabels);

  analyseMonitor->beginTask("Analysing", callback->context->unvisitedStackElements, analyseMonitor);
  //analyseMonitor->beginTask("Analysing", lengthSymList(labels), analyseMonitor);
  //analyseMonitor->subTask("HAHAHA", analyseMonitor);

  // set progress monitor for cudd
  setProgressMonitor(ipds, analyseMonitor);

  problem->stack = labels->wId;

  // ***PRESTAR STARTS HERE***
  result = intAnalyse(ipds, problem);

  // reset progress monitor
  resetProgressMonitor(ipds);

  DONE(analyseMonitor);

  freeSubProgressMonitor(analyseMonitor);
  freeQueueStrategy(problem->queue);

  //freeReachabilityContext((reachabilityContext*)problem.user);
  freeCallbackProblem(problem);
  freeJavaProgressMonitor(javaMonitor);
  // free it again in case we stopped because monitor was cancelled
  freeWIdentList(labels);

  jobject jpds = createIntPds(env, ipds);
  return createCoverageResult(env, jpds, createIntResult(env, result, jpds, 
							 JNI_FALSE));*/
}
