#include "coverage.h"
#include "absRef.h"

#include "bdd.h"
#include "ir.h"
#include "bp.h"

/**
 * Prints out usage of coverage program and exits with return value 1.
 */
static void usage (const char *progname)
{
  printf("usage: %s <file>\n\n"
	 "the file is checked until all states have been reached at least "
	 "once.\n"
	 "<file> must be a file from Remy's new input language.\n",
	 progname);
  exit(1);
}

int main (int argc, char **argv)
{
  intPds *ipds;
  wIdent *excludedStates;
  intProblem *problem;

  if (argc != 2)
  {
    usage(argv[0]);
  }

  wInit();
  bddInit();

  ipds = irReadPds(argv[1], 0);
  excludedStates = Calloc(2, sizeof(wIdent));
  *excludedStates = wIdentCreate("error");

  problem = analyseCoverage(ipds, NULL, NULL);

  wIdentList* unreached = 
    getUnreachableStates(ipds, (coverageContext *)problem->user);

    
  for (; unreached; unreached = unreached->next) {
    printf("%s\n", wIdentString(unreached->wId));
  }
  
  freeCoverageProblem(problem);
  Free(ipds);
  Free(excludedStates);

  return 0;
}
