/*****************************************************************************/

#include "alloc.h"
#include "crc32.h"
#include "int.h"

/***************************************************************************/

int intEhCompare (void *va, void *vb)
{
        intExpr *a = va, *b = vb;

        #define intHashCompare(field) \
                if (a->field != b->field) return (a->field < b->field)? -1 : 1;
        intHashCompare(type);
        intHashCompare(left); intHashCompare(right);
	if (a->q)
	{
		intHashCompare(q->name);
		intHashCompare(q->lower); intHashCompare(q->upper);
	}
	if (a->vref)
	{
		intHashCompare(vref->var);
		intHashCompare(vref->primes);
		intHashCompare(vref->index);
	}
	intHashCompare(value);
        return 0;
}

static intExpr* intExprGenerate (intExpr *expr)
{
	intExpr *nexpr;
	unsigned long hash = 0xffffffff;

	/* hash over intExpr structure */
	#define iegAdd(addr) hash = \
		crcFromSeed(hash,(char*)&(addr),sizeof(addr));

	iegAdd(expr->type);
	iegAdd(expr->left);
	iegAdd(expr->right);
	if (expr->q) iegAdd(expr->q->name);
	if (expr->q) iegAdd(expr->q->lower);
	if (expr->q) iegAdd(expr->q->upper);
	if (expr->vref) iegAdd(expr->vref->var);
	if (expr->vref) iegAdd(expr->vref->primes);
	if (expr->vref) iegAdd(expr->vref->index);
	iegAdd(expr->value);

	expr->hash = hash;

	nexpr = wHashInsert(intExprHash,expr);
	if (nexpr == expr) return expr;

	if (expr->vref) Free(expr->vref);
	Free(expr);

	return nexpr;
}

void intEhDelete (void *a)
{
	intExpr *expr = (intExpr*) a;

	/* Fixme: Cannot deref the BDD node without knowing the manager;
	   however, the manager is not included in the structure (and
	   including it does not make proper sense). */
	//if (expr->bdd) Cudd_RecursiveDeref(...mgr...,expr->bdd);
	if (expr->type == EXPR_EXISTS || expr->type == EXPR_FORALL)
		Free(expr->q);
	if (expr->vref) Free(expr->vref);
	Free(expr);
}

/***************************************************************************/

intExpr* intExprBinary (char type, intExpr *left, intExpr *right)
{
	intExpr *expr = Calloc(1,sizeof(struct intExpr));

	expr->type = type;
	expr->left = left;
	expr->right = right;

	return intExprGenerate(expr);
}

intExpr* intExprQuantified (char type, intQuantifier *q, intExpr *subexpr)
{
	intExpr *expr = Calloc(1,sizeof(struct intExpr));

	expr->type = type;
	expr->q = q;
	expr->left = subexpr;

	return intExprGenerate(expr);
}

intExpr* intExprVarRef (intVariable *var, char primes, intExpr *index)
{
	intExpr *expr = Calloc(1,sizeof(struct intExpr));

	expr->type = EXPR_VAR;
	expr->vref = Malloc(sizeof(struct intVarRef));
	expr->vref->var = var;
	expr->vref->primes = primes;
	expr->vref->index = index;

	return intExprGenerate(expr);
}

intExpr* intExprQvar (intQuantifier *q)
{
	intExpr *expr = Calloc(1,sizeof(struct intExpr));

	expr->type = EXPR_QVAR;
	expr->q = q;

	return intExprGenerate(expr);
}

intExpr* intExprConst (int value)
{
	intExpr *expr = Calloc(1,sizeof(struct intExpr));

	expr->type = EXPR_CONST;
	expr->value = value;

	return intExprGenerate(expr);
}

intQuantifier* intCreateQuantifier (wIdent name, int lower, int upper)
{
	intQuantifier *q = Malloc(sizeof(struct intQuantifier));

	q->name = name;
	q->lower = lower;
	q->upper = upper;

	return q;
}

/***************************************************************************/

static void intExprDeleteVrefs (intVarRef *vref)
{
	while (vref)
	{
		intVarRef *tmp = vref->next;
		Free(vref);
		vref = tmp;
	}
}

void intExprDelete (intExpr *expr)
{	/* deprecated */
	intVarRef *vref;

	if (!expr) return;
	switch (expr->type)
	{
	    case EXPR_OR:   case EXPR_AND:   case EXPR_EQV:  case EXPR_XOR:
	    case EXPR_PLUS: case EXPR_MINUS: case EXPR_MULT:
	    case EXPR_DIV:  case EXPR_SHL:   case EXPR_SHR:
	    case EXPR_BIT_OR: case EXPR_BIT_AND: case EXPR_BIT_XOR:
		intExprDelete(expr->left);
		intExprDelete(expr->right);
		break;
	    case EXPR_LT:   case EXPR_LE:    case EXPR_EQ:
	    case EXPR_NE:   case EXPR_GE:    case EXPR_GT:
		intExprDelete(expr->left);
		intExprDelete(expr->right);
		intExprDeleteVrefs((intVarRef*)(expr->value));
		break;
	    case EXPR_NOT:
		intExprDelete(expr->left);
		break;
	    case EXPR_FORALL:
	    case EXPR_EXISTS:
		intExprDelete(expr->left);
		Free(expr->q);
		break;
	    case EXPR_VAR:
		vref = expr->vref;
		if (vref->index) intExprDelete(vref->index);
		if (vref->var->type == 0 && vref->index)
			intExprDeleteVrefs((intVarRef*)(expr->value));
		if (vref->var->type == 0) Free(vref);
		break;
	    case EXPR_QVAR:
	    case EXPR_CONST:
		break;
	    default:
		fprintf(stderr,"unknown token in intFreeExpr\n");
		exit(1);
	}
	Free(expr);
}

/***************************************************************************/
