#include "test.h"

#include <bdd.h>
#include <coverage.h>
#include <int.h>
#include <ir.h>
#include <absRef.h>

static void testPushingAndPopping()
{
  queueStrategy *queue = createPreferringNewTransitionsQueueStrategy();
  wTrans *trans1 = Malloc(sizeof(wTrans));
  wTrans *trans2 = Malloc(sizeof(wTrans));
  wTrans *trans3 = Malloc(sizeof(wTrans));
  wTrans *trans4 = Malloc(sizeof(wTrans));

  bddPds bpds = { .zero = (struct DdNode*)1 };
  bddDomain bdd = { .bdd = NULL, .pds = &bpds };

  trans1->user = NULL;
  trans2->user = NULL;
  trans3->user = NULL;
  trans4->user = NULL;

  trans1->label = &bdd;
  trans2->label = &bdd;
  trans3->label = &bdd;
  trans4->label = &bdd;
  
  /* adding one transition and popping it */
  pushPreferringNewTransitions(trans1, queue);
  assert(trans1 == popPreferringNewTransitions(queue));
  assert(NULL == popPreferringNewTransitions(queue));

  /* adding one transition multiple times and to the first queue the first
     time */
  trans1->user = NULL;
  pushPreferringNewTransitions(trans1, queue);
  pushPreferringNewTransitions(trans1, queue);
  assert(trans1 == popPreferringNewTransitions(queue));
  assert(NULL == popPreferringNewTransitions(queue));

  /* adding one transition multiple times and always to the second queue */
  pushPreferringNewTransitions(trans1, queue);
  pushPreferringNewTransitions(trans1, queue);
  assert(trans1 == popPreferringNewTransitions(queue));
  assert(NULL == popPreferringNewTransitions(queue));

  /* adding first to old queue and then to new queue and checking correct
     order */
  pushPreferringNewTransitions(trans1, queue);
  pushPreferringNewTransitions(trans2, queue);
  assert(trans2 == popPreferringNewTransitions(queue));
  assert(trans1 == popPreferringNewTransitions(queue));
  assert(NULL == popPreferringNewTransitions(queue));

  /* adding intermingledly and checking order and emptiness */
  pushPreferringNewTransitions(trans3, queue);
  pushPreferringNewTransitions(trans1, queue);
  pushPreferringNewTransitions(trans4, queue);
  pushPreferringNewTransitions(trans2, queue);
  assert(trans4 == popPreferringNewTransitions(queue));
  assert(trans3 == popPreferringNewTransitions(queue));
  pushPreferringNewTransitions(trans4, queue);
  assert(trans4 == popPreferringNewTransitions(queue));
  assert(trans2 == popPreferringNewTransitions(queue));
  assert(trans1 == popPreferringNewTransitions(queue));
  assert(NULL == popPreferringNewTransitions(queue));
  

  freeQueueStrategy(queue);
  assert(NULL == queue->data);

  Free(trans1);
  Free(trans2);
  Free(trans3);
  Free(trans4);
}

static void testFIFOPushingAndPopping()
{
  queueStrategy *queue = createFIFOPreferringNewTransitionsQueueStrategy();
  wTrans *trans1 = Malloc(sizeof(wTrans));
  wTrans *trans2 = Malloc(sizeof(wTrans));
  wTrans *trans3 = Malloc(sizeof(wTrans));
  wTrans *trans4 = Malloc(sizeof(wTrans));

  bddPds bpds = { .zero = (struct DdNode*)1 };
  bddPds nullPds = { .zero = NULL };
  bddDomain nullBdd = { .bdd = NULL, .pds = &nullPds };
  bddDomain bdd = { .bdd = NULL, .pds = &bpds };

  trans1->user = NULL;
  trans2->user = NULL;
  trans3->user = NULL;
  trans4->user = NULL;

  trans1->label = &bdd;
  trans2->label = &bdd;
  trans3->label = &bdd;
  trans4->label = &bdd;
  
  /* adding one transition and popping it */
  queue->pushTransition(trans1, queue);
  assert(trans1 == queue->popTransition(queue));
  assert(NULL == queue->popTransition(queue));

  /* adding one transition multiple times and to the first queue the first
     time */
  trans1->user = NULL;
  queue->pushTransition(trans1, queue);
  queue->pushTransition(trans1, queue);
  assert(trans1 == queue->popTransition(queue));
  assert(NULL == queue->popTransition(queue));

  /* adding one transition multiple times and always to the second queue */
  queue->pushTransition(trans1, queue);
  queue->pushTransition(trans1, queue);
  assert(trans1 == queue->popTransition(queue));
  assert(NULL == queue->popTransition(queue));

  /* adding first to old queue and then to new queue and checking correct
     order */
  queue->pushTransition(trans1, queue);
  queue->pushTransition(trans2, queue);
  assert(trans2 == queue->popTransition(queue));
  assert(trans1 == queue->popTransition(queue));
  assert(NULL == queue->popTransition(queue));

  /* adding intermingledly and checking order and emptiness */
  queue->pushTransition(trans3, queue);
  queue->pushTransition(trans1, queue);
  queue->pushTransition(trans4, queue);
  queue->pushTransition(trans2, queue);
  assert(trans3 == queue->popTransition(queue));
  assert(trans4 == queue->popTransition(queue));
  queue->pushTransition(trans4, queue);
  assert(trans1 == queue->popTransition(queue));
  assert(trans2 == queue->popTransition(queue));
  assert(trans4 == queue->popTransition(queue));
  assert(NULL == queue->popTransition(queue));

  /* transition with empty bdd */
  trans1->user = NULL;
  trans1->label = &nullBdd;
  trans2->user = NULL;

  queue->pushTransition(trans1, queue);
  queue->pushTransition(trans3, queue);
  queue->pushTransition(trans2, queue);
  queue->pushTransition(trans4, queue);

  /* adding intermingledly and having a transition with empty bdd */
  assert(trans1 == queue->popTransition(queue));
  queue->pushTransition(trans1, queue);
  assert(trans2 == queue->popTransition(queue));
  assert(trans1 == queue->popTransition(queue));
  assert(trans3 == queue->popTransition(queue));
  queue->pushTransition(trans2, queue);
  queue->pushTransition(trans1, queue);
  assert(trans1 == queue->popTransition(queue));
  assert(trans4 == queue->popTransition(queue));
  assert(trans2 == queue->popTransition(queue));

  freeQueueStrategy(queue);
  assert(NULL == queue->data);

  Free(trans1);
  Free(trans2);
  Free(trans3);
  Free(trans4);
}


static wIdentList* createBubbleIgnoredList()
{
  wIdentList *excluded = NULL;
  insert(wIdentCreate("AssertError"), &excluded);
  insert(wIdentCreate("HeapOverflow"), &excluded);
  return excluded;
}

static void testStrategy(intPds *pds, wIdentList *excluded, 
			 queueStrategy *queue, const char *method,
			 const char *name)
{
  intProblem *problem = createCoverageProblem(pds, NULL, excluded);
  intResult *result;
  struct tms time;
  clock_t start;

  problem->queue = queue;

  times(&time);
  start = time.tms_utime;
  result = intAnalyse(pds, problem);
  times(&time);
  printf("Ticks for analysing %s - %s %d\n", name, method,
	 time.tms_utime - start);

  {
    wIdentList *unreachedStates = getUnreachableStates
      (pds, (coverageContext*)problem->user);
    wIdentList *state;
    printList(unreachedStates);    
    for (state = unreachedStates; state; state = state->next) {
      assert(isNotReachable(pds, state->wId));
    }
    assertEquals(0, lengthSymList(unreachedStates)); 
    assertEquals(0, ((coverageContext*)problem->user)->unvisitedStackElements);
    freeWIdentList(unreachedStates);
  }

  intDeleteResult(result);
  freeCoverageProblem(problem);
}

static void testStrategyForPds(intPds *pds, wIdentList *excluded,
				const char *name)
{
  queueStrategy *queue = NULL;
 
  testStrategy(pds, excluded, queue, "No strategy", name);
   
  queue = createFIFOQueueStratey();

  testStrategy(pds, excluded, queue, "FIFO", name);

  freeQueueStrategy(queue);

  queue = createLIFOQueueStratey();
  
  testStrategy(pds, excluded, queue, "LIFO", name); 

  freeQueueStrategy(queue);

  queue = createPreferringNewTransitionsQueueStrategy();
  
  testStrategy(pds, excluded, queue, "new transitions first", name); 

  freeQueueStrategy(queue);

  queue = createFIFOPreferringNewTransitionsQueueStrategy();

  testStrategy(pds, excluded, queue, "new transitions FIFO first", name);

  freeQueueStrategy(queue);
}

static wIdentList* createQuickSortIgnoredList()
{
  wIdentList *excluded = NULL;
/*   insert(wIdentCreate("AssertError"), &excluded); */
/*   insert(wIdentCreate("HeapOverflow"), &excluded); */
  insert(wIdentCreate("de_uni_stuttgart_fmi_szs_jmoped_coverage_" 
 		      "CoverageTesterTest_QuickSort_sort_AIII_V179"),
 	 &excluded);
  return excluded;
}

static void testQueueStrategy()
{
  intPds *pds = irReadPds("BubbleSort.rem", 0);
  wIdentList *ignored = createBubbleIgnoredList();
  testStrategyForPds(pds, ignored, "BubbleSort.rem");  

  freeWIdentList(ignored);
  intPdsDelete(pds);

  pds = irReadPds("QuickSort3.rem", 0);
  ignored = createQuickSortIgnoredList();

  testStrategyForPds(pds, ignored, "QuickSort3.rem");

  freeWIdentList(ignored);
  intPdsDelete(pds);
}


int main(int argc, char** argv)
{
  TestTuple tests[] = {
    ADD_TEST(testPushingAndPopping),
    ADD_TEST(testFIFOPushingAndPopping),
    ADD_TEST(testQueueStrategy),
  };
  setUp();
  run_tests(sizeof(tests) / sizeof(TestTuple), tests, argc, argv);
  return 0;
}

/*
 * Local Variables:
 * compile-command: "make -k queuestrategytest"
 * End:
 */
