#ifndef TEST_H
#define TEST_H

#include <assert.h>
#include <stdio.h>

#include <sys/times.h>
#ifndef HZ
# include <sys/param.h>
#endif
#if !defined(HZ) && defined(CLOCKS_PER_SEC)
# define HZ CLOCKS_PER_SEC
#endif
#if !defined(HZ) && defined(CLK_TCK)
# define HZ CLK_TCK
#endif
#ifndef HZ
# define HZ 60
#endif

#include <absRef.h>

#define RUN_TEST(test) \
printf("Running  test %s\n", #test); \
test(); \
printf("Finished test %s\n\n", #test);

#define assertEquals(expected, actual) \
if (expected != actual) { \
printf("Expected %d, but was %d\n", expected, actual); \
} \
assert(expected == actual); \

typedef void (*TestCase)(void);

enum Run { RUN_BY_DEFAULT, NOT_RUN_BY_DEFAULT };

typedef struct TestTuple
{
  const char* name;
  TestCase test;
  enum Run run;
} TestTuple;

#define ADD_TEST(test) \
 { #test, &test, RUN_BY_DEFAULT }

#define ADD_LONG_TEST(test) \
 { #test, &test, NOT_RUN_BY_DEFAULT }

void run_test(TestTuple test)
{
  printf("Running test %s\n", test.name);
  test.test();
  printf("Finished test %s\n\n", test.name);
}

void run_tests(int testc, TestTuple tests[], int argc, char** argv)
{
  if (argc > 1) {
    if (strcmp(argv[1], "--list") == 0) {
      int i;
      for (i = 0; i < testc; i++) {
	if (tests[i].run == NOT_RUN_BY_DEFAULT) {
	  printf("%s (not run by default)\n", tests[i].name);
	}
	else {
	  printf("%s\n", tests[i].name);
	}
      }
    }
    else {
      int i;
      for (i = 1; i < argc; i++) {
	int j;
	for (j = 0; j < testc; j++) {
	  if (strcmp(argv[i], tests[j].name) == 0) {
	    run_test(tests[j]);
	  }
	}
      }
    }
  }
  else {
    int j;
    for (j = 0; j < testc; j++) {
      if (tests[j].run == RUN_BY_DEFAULT) {
	run_test(tests[j]);
      }
    }
  }

}

void printList(wIdentList *list)
{
  wIdentList *iter;
  for (iter = list; iter; iter = iter->next)
  {
    printf("%s\n", wIdentString(iter->wId));
  }
}

extern int irVerbose;

void setUp()
{
  // turn off verbosity
  irVerbose = 0;
  wInit();
  bddInit();
  intInit();
}

#include <int.h>

int isNotReachable(intPds *pds, wIdent stack)
{
  intProblem prob = { PROBLEM_REACHABILITY, wIdentCreate("q"), stack,
		      &bddStateIsReached, NULL, NULL };
  intResult *result = intAnalyse(pds, &prob);
  bddResultType answer = result->yesno;
  intDeleteResult(result);
  return answer == RESULT_NO;
}

#endif
