#include "test.h"

#include <coverage.h>
#include <int.h>
#include <ir.h>
#include <absRef.h>

static void testStartsWith()
{
  // positive cases
  assert(startsWith("", "") != 0);
  assert(startsWith("hello", "hel") != 0);
  assert(startsWith("helo", "hel") != 0);
  assert(startsWith("hello", "hello") != 0);


  // negative cases
  assertEquals(0, startsWith("", "f"));
  assertEquals(0, startsWith("hel", "hello"));
  assertEquals(0, startsWith("hel", "helo"));
}

static void testMatchesPrefix()
{
  wIdent id1 = wIdentCreate("hello_world");
  assert(matchPrefix(id1, "hello_world"));
  assert(matchPrefix(id1, NULL));
  assert(matchPrefix(id1, NULL));
  assert(matchPrefix(id1, "hello_"));
  assert(matchPrefix(id1, "hello_"));
}

static void testIrReadPdsPerformance()
{
  intPds *pds = irReadPds("bubble.rem", 0);
  intPdsDelete(pds);
}

static void testPerformance()
{
  intPds *pds;
  intProblem *problem;
  struct tms time;
  clock_t start;
  

  times(&time);
  start = time.tms_utime;

  irVerbose = 2;
  
  pds = irReadPds("QuickSort_sort.rem", 0);

  times(&time);
  printf("Ticks for reading QuickSort_sort.rem %d\n", time.tms_utime - start);

  times(&time);
  start = time.tms_utime;

  problem = analyseCoverage(pds, NULL, NULL);

  times(&time);
  printf("Ticks for analysing %d\n", time.tms_utime - start);
  
  freeCoverageProblem(problem);
  intPdsDelete(pds);
}

int main(int argc, char** argv)
{
  TestTuple tests[] = {
    ADD_TEST(testStartsWith),
    ADD_TEST(testMatchesPrefix),
    ADD_TEST(testIrReadPdsPerformance),
    ADD_LONG_TEST(testPerformance),
  };
  setUp();
  run_tests(sizeof(tests) / sizeof(TestTuple), tests, argc, argv);
  return 0;
}

/*
 * Local Variables:
 * compile-command: "make -k utilstest"
 * End:
 */
