/*****************************************************************************
   Example taken from:

     S. Schwoon, S. Jha, T. Reps, and S. Stubblebine, On generalized
     authorization problems. To appear in Proc. 16th IEEE Computer
     Security Foundations Workshop, (June 30 - July 2, 2003, Asilomar,
     Pacific Grove, CA).  Tech. Report version: TR-1469, Computer Sciences
     Department, University of Wisconsin, Madison, WI, Jan. 2003.
     http://www.cs.wisc.edu/~reps/#csfw03
 *****************************************************************************/

#include <stdio.h>
#include "wpds.h"
#include "print.h"

#include <assert.h>

void *nullfn (void *a, void *b) { return NULL; }
int   eqfn (void *a, void *b)   { return a==b; }
void *null ()			{ return NULL; }

int main()
{
  wPDS *pds;
  wFA *fa, *fa_pre;
  wPath *p;
  wSemiring sr = { &nullfn,&nullfn,NULL,&null,&null,NULL,&eqfn,NULL,NULL };
		/* extend, combine, diff, one, zero, q-one, eq, ref, deref */
  wSemiring *s = &sr;
  wIdent k_x,k_h,k_aids,k_im,k_alice;
  wIdent y_patient, y_square, y_bsquare;
  wIdent final_state;
  wTrans *found;

  wInit();

  k_x = wIdentCreate("KX");
  k_h = wIdentCreate("KH");
  k_aids = wIdentCreate("KH-AIDS");
  k_im = wIdentCreate("KH-IM");
  k_alice = wIdentCreate("Alice");

  y_patient = wIdentCreate("patient");
  y_square = wIdentCreate("square");
  y_bsquare = wIdentCreate("b_square");

  final_state = wIdentCreate("__accepting_state__");

  pds = wPDSCreate(s);

  wPDSInsert(pds, k_x, y_square, k_h, y_patient, y_bsquare, NULL);
  wPDSInsert(pds, k_h, y_patient, k_im, y_patient, 0, NULL);
  wPDSInsert(pds, k_h, y_patient, k_aids, y_patient, 0, NULL);
  wPDSInsert(pds, k_im, y_patient, k_alice, 0, 0, NULL);
  wPDSInsert(pds, k_aids, y_patient, k_alice, 0, 0, NULL);

  fa = wFACreate(s);
  wFAInsert(fa,k_alice, y_square , final_state, NULL, NULL);
  wFAInsert(fa,k_alice, y_bsquare , final_state, NULL, NULL);

  print_automaton(fa,"before");
  fa_pre = wPrestar(pds,fa,TRACE_YES);
  print_automaton(fa_pre,"after");

  printf("trace example:\n");
  found = wFAFind(fa_pre,k_x,y_square,final_state);
  assert(found != NULL);
  p = wPathCreate(fa_pre, found,NULL);
  print_trace(fa_pre,p);

  wPDSDelete(pds);
  wFADelete(fa);
  wFADelete(fa_pre);
  wFinish();

  return 0;
} /** end of main **/
